/*
 * TAM
 * Copyright (C) 2021 - Delft University of Technology
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package nl.tudelft.tam.controller;

import java.time.LocalDateTime;
import java.time.LocalTime;
import java.util.List;
import java.util.stream.Collectors;

import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.stereotype.Controller;
import org.springframework.ui.Model;
import org.springframework.validation.annotation.Validated;
import org.springframework.web.bind.annotation.*;

import jakarta.validation.Valid;
import lombok.AllArgsConstructor;
import nl.tudelft.labracore.api.EditionControllerApi;
import nl.tudelft.labracore.api.RoleControllerApi;
import nl.tudelft.labracore.api.dto.*;
import nl.tudelft.labracore.lib.security.user.AuthenticatedPerson;
import nl.tudelft.labracore.lib.security.user.Person;
import nl.tudelft.tam.dto.create.TamEditionCreateDTO;
import nl.tudelft.tam.dto.patch.AddMemberDTO;
import nl.tudelft.tam.dto.view.details.EditionDetailsOffersWorkDTO;
import nl.tudelft.tam.service.*;

@Controller
@AllArgsConstructor
@RequestMapping("edition")
public class EditionController {

	private EditionService editionService;
	private RoleService roleService;
	private PersonService personService;
	private JobOfferService jobOfferService;
	private RoleControllerApi roleApi;
	private EditionControllerApi editionApi;
	private CoordinatorDefaultService coordinatorDefaultService;
	private ApplicationService applicationService;

	/**
	 * Creates a new edition, adds students if selected, and adds creator as teacher.
	 *
	 * @param  person       The creator / currently authenticated person
	 * @param  tamCreateDto The new edition data
	 * @return              The manage job offers page
	 */
	@Validated
	@PostMapping("create")
	@PreAuthorize("@authorisationService.canManageCourse(#tamCreateDto.getCourse().getId())")
	public String createEdition(@AuthenticatedPerson Person person,
			@Valid @ModelAttribute("tamEditionCreateDTO") TamEditionCreateDTO tamCreateDto) {
		// Create the edition
		EditionCreateDTO editionCreateDto = new EditionCreateDTO()
				.name(tamCreateDto.getName())
				.course(tamCreateDto.getCourse())
				.cohort(tamCreateDto.getCohort())
				.enrollability(EditionCreateDTO.EnrollabilityEnum.OPEN)
				.startDate(LocalDateTime.of(tamCreateDto.getStartDate(), LocalTime.of(0, 0)))
				.endDate(LocalDateTime.of(tamCreateDto.getEndDate(), LocalTime.of(23, 59)));

		Long newEditionId = editionService.addEdition(editionCreateDto);

		// Set the creating person as a teacher for the course (ie. manager)
		roleService.changeToTeacher(person.getId(), newEditionId);

		return "redirect:/job-offer/manage";
	}

	@Validated
	@GetMapping("{id}")
	@PreAuthorize("@authorisationService.canManageEdition(#id)")
	public String viewEdition(@PathVariable Long id, Model model) {
		List<EditionDetailsOffersWorkDTO> editions = editionService
				.getEditionsWithJobsAndWork(List.of(id));
		List<RolePersonDetailsDTO> TAsWithoutJobOffer = jobOfferService.getTAsWithoutAJobOfferInEdition(id);
		model.addAttribute("edition", editions.get(0));
		model.addAttribute("roles", editionService.getEditionRoles(id));
		model.addAttribute("TAsWithoutJobOffer", TAsWithoutJobOffer);
		model.addAttribute("TAsWithoutJobOfferIdentifiers", TAsWithoutJobOffer.stream()
				.map(x -> x.getPerson().getUsername()).collect(Collectors.joining("\n")));
		coordinatorDefaultService.addDefaultsToEditionsWithJobsAndWork(editions);
		return "edition/view";
	}

	/**
	 * Adds a user to an edition.
	 *
	 * @param  id   The id of the edition
	 * @param  user The user to add
	 * @return      The link to which the user is redirected
	 */
	@Validated
	@PostMapping("{id}/add-staff")
	@PreAuthorize("@authorisationService.canManageEdition(#id)")
	public String addMember(@PathVariable Long id, @ModelAttribute("addMember") AddMemberDTO user) {
		Long personId = personService.getPersonByUsername(user.getUsername()).getId();
		roleApi.addRole(new RoleCreateDTO().edition(new EditionIdDTO().id(id))
				.person(new PersonIdDTO().id(personId)).type(user.getRole())).block();
		return "redirect:/edition/{id}";
	}

	/**
	 * Removes a members from an edition
	 *
	 * @param  id       The id of the edition
	 * @param  username The username of the person
	 * @return          The page to redirect to
	 */
	@Validated
	@PostMapping("{id}/remove-staff")
	@PreAuthorize("@authorisationService.canManageEdition(#id)")
	public String removeMember(@AuthenticatedPerson Person person, @RequestParam String username,
			@PathVariable Long id) {
		Long personId = personService.getPersonByUsername(username).getId();
		applicationService.rejectPersonFromAllOffersInEdition(id, personId, person.getId());
		return "redirect:/edition/{id}";
	}

}
