/*
 * TAM
 * Copyright (C) 2021 - Delft University of Technology
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package nl.tudelft.tam.controller;

import java.util.List;
import java.util.Map;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.stereotype.Controller;
import org.springframework.ui.Model;
import org.springframework.web.bind.annotation.*;
import org.springframework.web.servlet.mvc.support.RedirectAttributes;

import nl.tudelft.labracore.api.dto.PersonSummaryDTO;
import nl.tudelft.tam.model.Profile;
import nl.tudelft.tam.service.ApplicationService;
import nl.tudelft.tam.service.PersonService;
import nl.tudelft.tam.service.ProgramService;

@Controller
@RequestMapping("person")
public class PersonController {

	@Autowired
	PersonService personService;

	@Autowired
	ApplicationService applicationService;

	@Autowired
	ProgramService programService;

	/**
	 * Search for candidate for a job offer based on one or a list of identifiers.
	 *
	 * @param  offerId       The job offer number for which the search is being conducted
	 * @param  identifiers   The identifiers to identify the people searched for
	 * @param  redirectAttrs The model attributes to keep when redirecting
	 * @return               A redirect to the specific job offer page
	 */
	@PostMapping("find-candidate")
	@PreAuthorize("@authorisationService.isTeacherOfAny()")
	public String findCandidates(@RequestParam Long offerId,
			@RequestParam String identifiers,
			RedirectAttributes redirectAttrs) {
		List<PersonSummaryDTO> candidates = personService.searchForPeople(identifiers);

		redirectAttrs.addFlashAttribute("specificOfferCandidates", candidates);
		redirectAttrs.addFlashAttribute("specificOfferQuery", identifiers);

		return "redirect:/job-offer/" + offerId;
	}

	/**
	 * Search for candidate to reject for a job offer based on one or a list of identifiers.
	 *
	 * @param  offerId       The job offer number for which the search is being conducted
	 * @param  identifiers   The identifiers to identify the people searched for
	 * @param  redirectAttrs The model attributes to keep when redirecting
	 * @return               A redirect to the specific job offer page
	 */
	@PostMapping("find-candidate-to-reject")
	@PreAuthorize("@authorisationService.isTeacherOfAny()")
	public String findCandidatesToReject(@RequestParam Long offerId,
			@RequestParam String identifiers,
			RedirectAttributes redirectAttrs) {
		List<PersonSummaryDTO> candidates = personService.searchForPeople(identifiers);

		redirectAttrs.addFlashAttribute("specificRejectCandidates", candidates);
		redirectAttrs.addFlashAttribute("specificRejectQuery", identifiers);

		return "redirect:/job-offer/" + offerId;
	}

	/**
	 * Search for trainees based on one or a list of identifiers.
	 *
	 * @param  programId     The programs from which the trainees must originate
	 * @param  returnPage    The page to be redirected to after the request is completed
	 * @param  trainees      The identifiers to identify the people searched for
	 * @param  datePassed    The date the training was passed
	 * @param  trainingType  The type of training that was passed
	 * @param  redirectAttrs The model attributes to keep when redirecting
	 * @return               A redirect to the specified page
	 */
	@PostMapping("find-trainees/{programId}/")
	@PreAuthorize("@authorisationService.canEditTaTrainingForProgram(#programId)")
	public String findTrainees(@PathVariable Long programId,
			@RequestParam String returnPage,
			@RequestParam String trainees,
			@RequestParam String datePassed,
			@RequestParam String trainingType,
			RedirectAttributes redirectAttrs) {
		Map<PersonSummaryDTO, Profile> people = personService
				.findPersonProfilePairs(trainees);

		redirectAttrs.addFlashAttribute("foundPeople", people);
		redirectAttrs.addFlashAttribute("searchQuery", trainees);
		redirectAttrs.addFlashAttribute("datePassed", datePassed);
		redirectAttrs.addFlashAttribute("trainingType", trainingType);

		return "redirect:" + returnPage;
	}

	/**
	 * Search for people based on one or a list of identifiers.
	 *
	 * @param  returnPage    The page to be redirected to after the request is completed
	 * @param  identifiers   The identifiers to identify the people searched for
	 * @param  newValue      The new value to be set
	 * @param  redirectAttrs The model attributes to keep when redirecting
	 * @return               A redirect to the specified page
	 */
	@PostMapping("find-people")
	@PreAuthorize("@authorisationService.canEditPayScale()")
	public String findTAsToPromote(@RequestParam String returnPage,
			@RequestParam String identifiers,
			@RequestParam String newValue,
			RedirectAttributes redirectAttrs) {
		Map<PersonSummaryDTO, Profile> people = personService
				.findPersonProfilePairs(identifiers);

		redirectAttrs.addFlashAttribute("foundPeople", people);
		redirectAttrs.addFlashAttribute("newValue", newValue);
		redirectAttrs.addFlashAttribute("searchQuery", identifiers);

		return "redirect:" + returnPage;
	}

	/**
	 * Searches for all people.
	 *
	 * @param  q     The search query
	 * @param  model The model to add the data to
	 * @return       The search page
	 */
	@GetMapping("search")
	@PreAuthorize("@authorisationService.isTeacherOfAny()")
	public String searchForPeople(@RequestParam(required = false) String q, Model model) {
		Map<PersonSummaryDTO, Profile> people = personService.findPersonProfilePairs(q);
		model.addAttribute("people", people);
		return "person/search";
	}

}
