/*
 * TAM
 * Copyright (C) 2021 - Delft University of Technology
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package nl.tudelft.tam.controller;

import java.util.List;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.stereotype.Controller;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.ui.Model;
import org.springframework.web.bind.annotation.*;

import nl.tudelft.labracore.api.dto.PersonSummaryDTO;
import nl.tudelft.labracore.api.dto.ProgramDetailsDTO;
import nl.tudelft.labracore.api.dto.RoleEditionDetailsDTO;
import nl.tudelft.labracore.lib.security.user.AuthenticatedPerson;
import nl.tudelft.labracore.lib.security.user.Person;
import nl.tudelft.tam.cache.EditionCacheManager;
import nl.tudelft.tam.dto.patch.NotificationPatchDTO;
import nl.tudelft.tam.dto.patch.ShirtPatchDTO;
import nl.tudelft.tam.dto.view.summary.RaiseRequestSummaryDTO;
import nl.tudelft.tam.model.Profile;
import nl.tudelft.tam.service.*;

@Controller
@RequestMapping("profile")
public class TAMProfileController {

	@Autowired
	private ProfileService profileService;

	@Autowired
	private RaiseRequestService raiseRequestService;

	@Autowired
	private PersonService personService;

	@Autowired
	private RoleService roleService;

	@Autowired
	private EditionCacheManager editionCache;
	@Autowired
	private ProgramService programService;

	/**
	 * Loads the profile page.
	 *
	 * @param  authPerson the person who is currently authenticated
	 * @param  model      the model into which the profile data will be loaded
	 * @return            the profile page
	 */
	@GetMapping
	@Transactional
	public String getProfile(@AuthenticatedPerson Person authPerson, Model model) {
		Profile profile = profileService.findByIdOrThrow(authPerson.getId());
		PersonSummaryDTO person = personService.getOrThrow(authPerson.getId());

		model.addAttribute("profile", profile);
		model.addAttribute("person", person);
		model.addAttribute("teacherView", false);

		List<ProgramDetailsDTO> coordinatingPrograms = programService
				.getCoordinatingPrograms(authPerson.getId());
		List<RoleEditionDetailsDTO> roles = roleService.getStaffRolesForPerson(person.getId());

		model.addAttribute("coordinatingPrograms", coordinatingPrograms);
		model.addAttribute("roles", roles);

		RaiseRequestSummaryDTO openRaiseRequest = raiseRequestService.getOpenForPerson(authPerson);
		model.addAttribute("openRaiseRequest", openRaiseRequest);

		model.addAttribute("poloPatch", new ShirtPatchDTO());

		return "profile/view";
	}

	/**
	 * Loads the profile page of a specific person
	 *
	 * @param  personId The person for whom to load the profile
	 * @param  model    The model int which the profile data will be loaded
	 * @return          The profile page
	 */
	@GetMapping("{personId}")
	@PreAuthorize("@authorisationService.isStaff()")
	public String getPersonProfile(@PathVariable Long personId, Model model) {
		Profile profile = profileService.findByIdOrThrow(personId);
		PersonSummaryDTO person = personService.getOrThrow(personId);

		model.addAttribute("profile", profile);
		model.addAttribute("person", person);
		model.addAttribute("teacherView", true);

		List<ProgramDetailsDTO> coordinatingPrograms = programService.getCoordinatingPrograms(personId);
		List<RoleEditionDetailsDTO> roles = roleService.getStaffRolesForPerson(person.getId());

		model.addAttribute("coordinatingPrograms", coordinatingPrograms);
		model.addAttribute("roles", roles);

		return "profile/view";
	}

	/**
	 * Updates the profile's preferences for a shirt.
	 *
	 * @param  person the authenticated person
	 * @param  patch  the patch which contains the updated information
	 * @return        a redirect to the view page
	 */
	@Transactional
	@PatchMapping("polo")
	public String updatePoloData(@AuthenticatedPerson Person person,
			@ModelAttribute("poloPatch") ShirtPatchDTO patch) {
		profileService.updatePoloData(person.getId(), patch);

		return "redirect:/profile";
	}

	/**
	 * Updates the profile's preferences for notifications.
	 *
	 * @param  person the authenticated person
	 * @param  patch  the patch which contains the updated preferences
	 * @return        a redirect to the view page
	 */
	@Transactional
	@PatchMapping("notifications")
	public String updateNotificationPreferences(@AuthenticatedPerson Person person,
			@ModelAttribute("notificationPatch") NotificationPatchDTO patch) {
		profileService.updateNotificationPreferences(person.getId(), patch);

		return "redirect:/profile";
	}

}
