/*
 * TAM
 * Copyright (C) 2021 - Delft University of Technology
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package nl.tudelft.tam.cron;

import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.scheduling.annotation.Scheduled;
import org.springframework.stereotype.Service;

import nl.tudelft.labracore.api.dto.EditionDetailsDTO;
import nl.tudelft.labracore.api.dto.PersonSummaryDTO;
import nl.tudelft.labracore.api.dto.ProgramSummaryDTO;
import nl.tudelft.labracore.api.dto.RolePersonDetailsDTO;
import nl.tudelft.tam.dto.create.EmailCreateDTO;
import nl.tudelft.tam.dto.view.details.RaiseRequestDetailsDTO;
import nl.tudelft.tam.dto.view.summary.ExtraWorkSummaryDTO;
import nl.tudelft.tam.enums.NotificationFrequency;
import nl.tudelft.tam.model.Coordinator;
import nl.tudelft.tam.model.Profile;
import nl.tudelft.tam.service.*;

@Service
public class CronService {

	@Autowired
	private EmailService emailService;

	@Autowired
	private ExtraWorkService extraWorkService;

	@Autowired
	private DeclarationService declarationService;

	@Autowired
	private RoleService roleService;

	@Autowired
	private EditionService editionService;

	@Autowired
	private ProfileService profileService;

	@Autowired
	private RaiseRequestService raiseRequestService;

	@Autowired
	private ProgramService programService;

	@Autowired
	private CoordinatorService coordinatorService;

	@Autowired
	private PersonService personService;

	/**
	 * Send Accumulated Changes Weekly on a Monday at 00:00. Currently: Teachers - Declarations Aren't Handled
	 * EditPayScale / Coordinators Generally - Raise Requests Awaiting Approval
	 */
	@Scheduled(cron = "0 0 0 * * 1")
	public void sendBulkWeeklyAccumulatedChanges() {
		remindTeachersAboutUnhandledDeclarations(NotificationFrequency.Weekly);
		remindEditPayScalePeopleAboutAwaitingRaiseRequests(NotificationFrequency.Weekly);
	}

	/**
	 * Send Accumulated Changes On Weekdays at 00:00. Currently: Teachers - Declarations Aren't Handled
	 * EditPayScale / Coordinators Generally - Raise Requests Awaiting Approval
	 */
	@Scheduled(cron = "0 0 0 * * 1-5")
	public void sendBulkDailyAccumulatedChanges() {
		remindTeachersAboutUnhandledDeclarations(NotificationFrequency.Daily);
		remindEditPayScalePeopleAboutAwaitingRaiseRequests(NotificationFrequency.Daily);
	}

	private void remindEditPayScalePeopleAboutAwaitingRaiseRequests(NotificationFrequency frequency) {
		List<Long> programIds = programService.getAllPrograms().stream()
				.map(ProgramSummaryDTO::getId)
				.collect(Collectors.toList());

		List<Long> programsWithAwaitingRaiseRequests = raiseRequestService.getAllSubmitted(programIds)
				.stream()
				.map(RaiseRequestDetailsDTO::getRelevantProgramIds)
				.flatMap(List::stream)
				.distinct()
				.toList();

		Map<String, String> coordinatorEmailToDisplayName = new HashMap<>();
		for (Long programId : programsWithAwaitingRaiseRequests) {
			List<Long> coordinatorIds = coordinatorService.getCoordinatorsByProgram(programId).stream()
					.map(Coordinator::getId).toList();
			List<PersonSummaryDTO> coordinators = personService.getPeopleById(coordinatorIds);

			coordinators.forEach(coordinator -> {
				profileService.findById(coordinator.getId()).ifPresent(profile -> {
					if (profile.getEmailNotifications()
							&& profile.getEmailNotificationFrequency().equals(frequency)) {
						coordinatorEmailToDisplayName.put(
								personService.getPersonById(profile.getId()).getEmail(),
								coordinator.getDisplayName());
					}
				});
			});
		}

		// Send Emails
		for (Map.Entry<String, String> entry : coordinatorEmailToDisplayName.entrySet()) {
			Map<String, Object> contextVars = new HashMap<>(Map.of("name", entry.getValue()));

			EmailCreateDTO email = EmailCreateDTO.builder()
					.subject("TAM - Raise Requests Pending")
					.template("html/raise_requests_pending")
					.contextVariables(contextVars)
					.build();

			emailService.sendEmail(List.of(entry.getKey()), email);
		}
	}

	private void remindTeachersAboutUnhandledDeclarations(NotificationFrequency frequency) {
		Map<String, List<String>> teacherEmailToExtraWorkList = new HashMap<>();
		Map<String, String> teacherEmailToDisplayName = new HashMap<>();

		List<ExtraWorkSummaryDTO> openWork = extraWorkService.getAllOpen();
		for (ExtraWorkSummaryDTO work : openWork) {
			float numberOfUnhandledHours = declarationService.getNumberOfUnhandledHours(work.getId());

			if (numberOfUnhandledHours > 0) {
				// Construct the bullet point of the extra work in the email
				EditionDetailsDTO edition = editionService.getEditionById(work.getEditionId());
				String extraWorkBulletPoint = edition.getCourse().getName() + "(" + edition.getName() + ") - "
						+ work.getName() + ": " + numberOfUnhandledHours + " hours";

				// Go through all teachers of the extra work
				List<RolePersonDetailsDTO> teachers = roleService.getTeachersById(work.getEditionId());
				for (RolePersonDetailsDTO teacher : teachers) {
					String email = teacher.getPerson().getEmail();
					String displayName = teacher.getPerson().getDisplayName();

					// Check if notifications are enabled
					Profile profile = profileService.findByIdOrCreate(teacher.getPerson().getId());
					if (profile.getEmailNotifications()
							&& profile.getEmailNotificationFrequency().equals(frequency)) {
						if (!teacherEmailToExtraWorkList.containsKey(email)) {
							teacherEmailToExtraWorkList.put(email, List.of(extraWorkBulletPoint));
						} else {
							List<String> ewList = teacherEmailToExtraWorkList.get(email);
							ewList.add(extraWorkBulletPoint);
							teacherEmailToExtraWorkList.put(email, ewList);
						}
						teacherEmailToDisplayName.put(email, displayName);
					}
				}
			}
		}

		// Send Emails
		for (Map.Entry<String, List<String>> entry : teacherEmailToExtraWorkList.entrySet()) {
			Map<String, Object> contextVars = new HashMap<>(Map.of(
					"name", teacherEmailToDisplayName.get(entry.getKey()),
					"listOfBulletPoints", entry.getValue()));

			EmailCreateDTO email = EmailCreateDTO.builder()
					.subject("TAM - Contract Requests Pending")
					.template("html/contract_requests_pending")
					.contextVariables(contextVars)
					.build();

			emailService.sendEmail(List.of(entry.getKey()), email);
		}
	}

}
