/*
 * TAM
 * Copyright (C) 2021 - Delft University of Technology
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package nl.tudelft.tam.io;

import java.time.LocalDate;
import java.time.LocalDateTime;
import java.util.ArrayList;
import java.util.List;

import org.springframework.stereotype.Component;

import lombok.RequiredArgsConstructor;
import nl.tudelft.labracore.api.dto.*;
import nl.tudelft.librador.dto.DTOConverter;
import nl.tudelft.tam.dto.create.JobOfferCreateDTO;
import nl.tudelft.tam.model.JobOffer;
import nl.tudelft.tam.service.CohortService;
import nl.tudelft.tam.service.CourseService;
import nl.tudelft.tam.service.EditionService;

@Component
@RequiredArgsConstructor
public class JobOfferImportHandler extends ImportHandler<JobOffer> {

	private final CohortService cohortService;
	private final CourseService courseService;
	private final EditionService editionService;

	private final DTOConverter converter;

	private List<EditionDetailsDTO> editions = null;
	private List<CourseSummaryDTO> courses = null;
	private List<CohortDetailsDTO> cohorts = null;

	@Override
	public void preImport() {
		List<EditionSummaryDTO> activeEditions = editionService.getAllActiveEditions(LocalDateTime.now(),
				LocalDateTime.now().plusYears(1));
		editions = editionService.getEditionsById(activeEditions.stream()
				.filter(x -> !x.getIsArchived()).map(EditionSummaryDTO::getId).toList());
		courses = courseService.getAllCourses();
		cohorts = cohortService
				.getCohortsById(cohortService.getAllCohorts().stream().map(CohortSummaryDTO::getId).toList());
	}

	@Override
	protected ImportResult<JobOffer> importLine(String[] line, int lineNumber) {
		JobOfferCreateDTO job = new JobOfferCreateDTO();
		List<ImportWarning> warnings = new ArrayList<>();

		// With or without edition info
		if (line.length != 14 && line.length != 18) {
			return new ImportResult.Error<>(new SimpleImportError(
					"Error on line " + lineNumber
							+ ": Incorrect amount of columns. Did you perhaps use an older version of the template?"));
		}

		// Try to find edition
		String editionName = line[0];
		String courseCode = line[1];
		List<EditionDetailsDTO> matchingEditions = editions
				.stream()
				.filter(e -> e.getCourse().getCode().equals(courseCode)
						&& e.getName().equals(editionName))
				.toList();
		if (matchingEditions.size() >= 2) {
			warnings.add(new SimpleImportWarning("Error on line " + lineNumber
					+ ": Multiple matching editions found, using the first one"));
		}

		// Parse basic job info
		LocalDate startDate = parseDate(line[4]).orElseThrow(() -> new ImportLineException(
				"Error on line " + lineNumber
						+ ": Column 'Start date' contained invalid content. Values allowed are expected in the format: dd-MM-yyyy"));
		LocalDate endDate = parseDate(line[5]).orElseThrow(
				() -> new ImportLineException(
						"Error on line " + lineNumber
								+ ": Column 'End date' contained invalid content. Values allowed are expected in the format: dd-MM-yyyy"));

		int maxHours = parseInt(line[7])
				.orElseThrow(() -> new ImportLineException(
						"Error on line " + lineNumber + ": Max hours is not a whole number"));
		boolean requireApplicationText = parseBoolean(line[9])
				.orElseThrow(() -> new ImportLineException(
						"Error on line " + lineNumber
								+ ": Column 'Require response' contained invalid content. Values allowed are 'yes' or 'no'"));
		boolean isDraft = parseBoolean(line[12])
				.orElseThrow(() -> new ImportLineException(
						"Error on line " + lineNumber
								+ ": Column 'draft' contained invalid content. Values allowed are 'yes' or 'no'"));

		// Set basic info
		job.setName(line[2]);
		job.setContractName(line[3]);
		job.setContractStartDate(startDate);
		job.setContractEndDate(endDate);
		job.setBaanCode(line[6]);
		job.setMaxHours(maxHours);
		job.setDescription(line[8]);
		job.setRequireApplicationText(requireApplicationText);
		job.setHiringMessage(line[10]);
		job.setRejectMessage(line[11]);
		job.setDraft(isDraft);

		// If there is a deadline, parse and set
		if (!line[13].isBlank()) {
			LocalDate deadline = parseDate(line[13]).orElseThrow(() -> new ImportLineException(
					"Error on line " + lineNumber
							+ ": Column 'Deadline' contained invalid content. Values allowed are expected in the format: dd-MM-yyyy"));
			job.setDeadline(deadline);
		}

		// Edition info supplied
		if (line.length == 18) {
			LocalDateTime editionStart = parseDateTime(line[16]).orElseThrow(() -> new ImportLineException(
					"Error on line " + lineNumber
							+ ": Column 'Edition start date' contained invalid content. Values allowed are expected in the format: dd-MM-yyyy HH:mm"));
			LocalDateTime editionEnd = parseDateTime(line[17]).orElseThrow(() -> new ImportLineException(
					"Error on line " + lineNumber
							+ ": Column 'Edition end date' contained invalid content. Values allowed are expected in the format: dd-MM-yyyy HH:mm"));

			// Need to create edition
			if (matchingEditions.isEmpty()) {
				var matchingCourses = courses.stream().filter(c -> c.getCode().equals(courseCode)).toList();
				if (matchingCourses.size() == 0)
					error("Error on line " + lineNumber + ": No courses found matching course code '"
							+ courseCode + "'");
				if (matchingCourses.size() >= 2)
					warnings.add(
							new SimpleImportWarning("Error on line " + lineNumber
									+ ": Multiple matching courses found, using the first one"));
				var matchingCohorts = cohorts.stream().filter(
						c -> c.getProgram().getName().equals(line[15]) && c.getName().equals(line[14]))
						.toList();
				if (matchingCohorts.size() == 0)
					error("Error on line " + lineNumber + ": No cohorts found with name '" + line[14]
							+ "' and programme '" + line[15] + "'");
				if (matchingCohorts.size() >= 2)
					warnings.add(
							new SimpleImportWarning("Error on line " + lineNumber
									+ ": Multiple matching cohorts found, using the first one"));

				EditionCreateDTO edition = new EditionCreateDTO()
						.name(editionName)
						.course(new CourseIdDTO().id(matchingCourses.get(0).getId()))
						.cohort(new CohortIdDTO().id(matchingCohorts.get(0).getId()))
						.startDate(editionStart)
						.endDate(editionEnd)
						.enrollability(EditionCreateDTO.EnrollabilityEnum.OPEN);

				return new ImportResult.Error<>(new EditionDoesNotExistError(
						"Edition '" + editionName + "' for course '" + courseCode + "' does not exist yet",
						job, edition, courseCode));
			}
		}

		// No editions and no creation info provided
		if (matchingEditions.isEmpty()) {
			return new ImportResult.Error<>(
					new SimpleImportError("No matching editions found with course code '" + courseCode
							+ "' and name '" + editionName + "'"));
		}

		// Successful import
		job.setEditionId(matchingEditions.get(0).getId());
		return warnings.isEmpty() ? new ImportResult.Success<>(converter.apply(job))
				: new ImportResult.WithWarnings<>(converter.apply(job), warnings);
	}

	public record EditionDoesNotExistError(String message, JobOfferCreateDTO jobOffer,
			EditionCreateDTO edition, String courseCode) implements ImportError {
	}

}
