/*
 * TAM
 * Copyright (C) 2021 - Delft University of Technology
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package nl.tudelft.tam.model;

import java.time.LocalDate;
import java.util.ArrayList;
import java.util.List;
import java.util.Objects;

import javax.annotation.Nullable;

import jakarta.persistence.*;
import jakarta.validation.constraints.NotNull;
import lombok.*;
import nl.tudelft.tam.enums.NotificationFrequency;
import nl.tudelft.tam.enums.PayScale;
import nl.tudelft.tam.enums.ShirtGender;
import nl.tudelft.tam.enums.ShirtSize;

@Data
@Entity
@Builder
@NoArgsConstructor
@AllArgsConstructor
public class Profile {

	@Id
	@NotNull
	private Long id;

	@Nullable
	private ShirtSize shirtSize;

	@Nullable
	private ShirtGender shirtGender;

	@NotNull
	@Builder.Default
	private PayScale payScale = PayScale.SA2;

	@NotNull
	@Builder.Default
	@ToString.Exclude
	@EqualsAndHashCode.Exclude
	@OneToMany(mappedBy = "person")
	private List<Declaration> declaration = new ArrayList<>();

	@NotNull
	@Builder.Default
	@ToString.Exclude
	@EqualsAndHashCode.Exclude
	@OneToMany(mappedBy = "handler")
	private List<Declaration> handledDeclarations = new ArrayList<>();

	@NotNull
	@Builder.Default
	@ToString.Exclude
	@EqualsAndHashCode.Exclude
	@OneToMany(mappedBy = "handler")
	private List<Application> handledApplications = new ArrayList<>();

	@NotNull
	@Builder.Default
	@ToString.Exclude
	@EqualsAndHashCode.Exclude
	@OneToMany(mappedBy = "requestedBy")
	private List<RaiseRequest> raiseRequests = new ArrayList<>();

	@NotNull
	@Builder.Default
	@ToString.Exclude
	@EqualsAndHashCode.Exclude
	@OneToMany(mappedBy = "approvedBy")
	private List<RaiseRequest> raiseApprovals = new ArrayList<>();

	@NotNull
	@Builder.Default
	@ToString.Exclude
	@EqualsAndHashCode.Exclude
	@OneToMany(mappedBy = "person")
	private List<TrainingApproval> trainingsPassed = new ArrayList<>();

	@NotNull
	@Builder.Default
	@ToString.Exclude
	@EqualsAndHashCode.Exclude
	@OneToMany(mappedBy = "approvedBy")
	private List<TrainingApproval> trainingApprovals = new ArrayList<>();

	@NotNull
	@Builder.Default
	@Column(columnDefinition = "DATE")
	private LocalDate createdDate = LocalDate.now();

	@NotNull
	@Builder.Default
	@Column(columnDefinition = "DATE")
	private LocalDate lastModifiedDate = LocalDate.now();

	@NotNull
	@Builder.Default
	private Boolean emailNotifications = true;

	@NotNull
	@Builder.Default
	private NotificationFrequency emailNotificationFrequency = NotificationFrequency.Weekly;

	@PreUpdate
	public void setLastModifiedDate() {
		this.lastModifiedDate = LocalDate.now();
	}

	@PrePersist
	public void setCreatedDate() {
		this.createdDate = LocalDate.now();
	}

	/**
	 * Gets the application this person handled in an edition.
	 *
	 * @param  editionId The id of the edition
	 * @return           The list of applications handled
	 */
	public List<Application> getHandledApplications(Long editionId) {
		return handledApplications.stream()
				.filter(d -> Objects.equals(d.getJobOffer().getEditionId(), editionId))
				.toList();
	}

	/**
	 * Gets the declaration this person handled in an edition.
	 *
	 * @param  editionId The id of the edition
	 * @return           The list of declarations handled
	 */
	public List<Declaration> getHandledDeclarations(Long editionId) {
		return handledDeclarations.stream()
				.filter(d -> Objects.equals(d.getExtraWork().getEditionId(), editionId))
				.toList();
	}

}
