/*
 * TAM
 * Copyright (C) 2021 - Delft University of Technology
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package nl.tudelft.tam.repository;

import java.util.List;

import org.springframework.data.jpa.repository.JpaRepository;
import org.springframework.data.rest.webmvc.ResourceNotFoundException;

import nl.tudelft.tam.model.Coordinator;

public interface CoordinatorRepository extends JpaRepository<Coordinator, Long> {

	/**
	 * Finds the coordinator by id or throws a ResourceNotFoundException.
	 *
	 * @param  id the id of the coordinator
	 * @return    the Coordinator object
	 */
	default Coordinator findByIdOrThrow(Long id) {
		// Spring will actually wrap this in a JpaObjectRetrievalFailureException
		return findById(id)
				.orElseThrow(() -> new ResourceNotFoundException("Coordinator was not found: " + id));
	}

	/**
	 * Gets all the coordinator roles for a given person.
	 *
	 * @param  personId The id of the person
	 * @return          The person's coordinator roles
	 */
	List<Coordinator> findAllByPersonId(Long personId);

	/**
	 * Gets whether the given person is a coordinator.
	 *
	 * @param  personId The id of the person
	 * @return          True iff the person is a coordinator
	 */
	boolean existsByPersonId(Long personId);

	/**
	 * Gets whether the given person is a coordinator for the given programme.
	 *
	 * @param  personId  The id of the person
	 * @param  programId The id of the programme
	 * @return           True iff the person is a coordinator for the programme
	 */
	boolean existsByPersonIdAndProgramId(Long personId, Long programId);

	/**
	 * Gets all the coordinators for a given programme.
	 *
	 * @param  programId The id of the programme
	 * @return           The programme's coordinators
	 */
	List<Coordinator> findAllByProgramId(Long programId);

}
