/*
 * TAM
 * Copyright (C) 2021 - Delft University of Technology
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package nl.tudelft.tam.security;

import java.time.LocalDate;
import java.util.List;
import java.util.Objects;
import java.util.Set;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.annotation.Lazy;
import org.springframework.security.core.Authentication;
import org.springframework.security.core.context.SecurityContextHolder;
import org.springframework.stereotype.Service;

import nl.tudelft.labracore.api.AuthorizationControllerApi;
import nl.tudelft.labracore.api.dto.PersonIdDTO;
import nl.tudelft.labracore.api.dto.RoleDetailsDTO;
import nl.tudelft.labracore.lib.security.LabradorUserDetails;
import nl.tudelft.labracore.lib.security.user.DefaultRole;
import nl.tudelft.labracore.lib.security.user.Person;
import nl.tudelft.tam.model.JobOffer;
import nl.tudelft.tam.service.*;

@Service
public class AuthorisationService {

	@Autowired
	private AuthorizationControllerApi authApi;

	@Autowired
	private JobOfferService jobOfferService;

	@Autowired
	@Lazy
	private ApplicationService applicationService;

	@Autowired
	private ExtraWorkService extraWorkService;

	@Autowired
	private DeclarationService declarationService;

	@Autowired
	private RaiseRequestService raiseRequestService;

	@Autowired
	private RoleService roleService;

	@Autowired
	private CourseService courseService;

	@Autowired
	private EditionService editionService;

	@Autowired
	private CoordinatorService coordinatorService;

	@Autowired
	private ProgramService programService;

	/**
	 * Returns the authenticated person from the user principal.
	 *
	 * @return the authenticated person object
	 */
	public Person getAuthPerson() {
		return ((LabradorUserDetails) SecurityContextHolder.getContext()
				.getAuthentication().getPrincipal()).getUser();
	}

	/**
	 * Checks whether the user is an admin.
	 *
	 * @return true if the user is an admin
	 */
	public boolean isAdmin() {
		return getAuthPerson().getDefaultRole() == DefaultRole.ADMIN;
	}

	/**
	 * Checks whether the authenticated user is a staff member.
	 *
	 * @return true if the authenticated user is a teacher or admin by default
	 */
	public boolean isStaff() {
		return Set.of(DefaultRole.TEACHER, DefaultRole.ADMIN)
				.contains(getAuthPerson().getDefaultRole());
	}

	/**
	 * Checks whether the user is authenticated.
	 *
	 * @return True if the user is authenticated
	 */
	public boolean isAuthenticated() {
		Authentication authentication = SecurityContextHolder.getContext().getAuthentication();
		return authentication != null && authentication.getPrincipal() instanceof LabradorUserDetails;
	}

	/**
	 * Checks if the person is a coordinator.
	 *
	 * @return true if the authenticated person is a coordinator
	 */
	public boolean isCoordinatorOfAny() {
		return isAdmin() || coordinatorService.existsByPersonId(getAuthPerson().getId());
	}

	/**
	 * Checks if the person is a coordinator for a programme.
	 *
	 * @param  programId The id of the program
	 * @return           true if the authenticated person is a coordinator
	 */
	public boolean isCoordinatorForProgram(Long programId) {
		return isAdmin()
				|| coordinatorService.existsByPersonIdAndProgramId(getAuthPerson().getId(), programId)
				|| programService.get(programId).getCoordinators().stream()
						.anyMatch(c -> Objects.equals(c.getId(), getAuthPerson().getId()));
	}

	/**
	 * Checks if the person is a coordinator for multiple programs
	 *
	 * @param  programIds The ids of the programs
	 * @return            true if the authenticated person is a coordinator
	 */
	public boolean isCoordinatorForPrograms(List<Long> programIds) {
		return isAdmin() || programIds.stream().allMatch(this::isCoordinatorForProgram);
	}

	/**
	 * Checks if the person can edit the coordinators for a programme
	 *
	 * @param  programId the id of the programme
	 * @return           True iff the authenticated person can edit the coordinators for the programme
	 */
	public boolean canEditCoordinatorsForProgram(Long programId) {
		return isAdmin() || programService.get(programId).getCoordinators().stream()
				.anyMatch(c -> Objects.equals(c.getId(), getAuthPerson().getId()));
	}

	/**
	 * Checks if the person is a coordinator for a course.
	 *
	 * @param  courseId The id of the course
	 * @return          True iff the authenticated person is a coordinator
	 */
	public boolean isCoordinatorForCourse(Long courseId) {
		Long programId = courseService.getOrThrow(courseId).getProgram().getId();
		return isAdmin()
				|| coordinatorService.existsByPersonIdAndProgramId(getAuthPerson().getId(), programId);
	}

	/**
	 * Checks if the person is a coordinator for an edition.
	 *
	 * @param  editionId The id of the edition
	 * @return           true if the authenticated person is a coordinator
	 */
	public boolean isCoordinatorForEdition(Long editionId) {
		return isAdmin() || isCoordinatorForCourse(editionService.getOrThrow(editionId).getCourse().getId());
	}

	/**
	 * Checks if the person is a teacher of any course.
	 *
	 * @return true if the authenticated person is a teacher of a course
	 */
	public boolean isTeacherOfAny() {
		return isAdmin() || Boolean.TRUE
				.equals(authApi.hasTeacherRoleInAnyCurrentEdition(getAuthPerson().getId()).block());
	}

	/**
	 * Checks if the person is a manager of any course.
	 *
	 * @return true if the authenticated person is a manager of a course
	 */
	public boolean canManageAny() {
		return isAdmin() || isCoordinatorOfAny() ||
				roleService.getTeacherRolesForPerson(getAuthPerson().getId()).stream()
						.anyMatch(r -> canManageEdition(r.getEdition().getId()));
	}

	/**
	 * Checks if a user can manage job offer.
	 *
	 * @param  offerId The id of the job offer
	 * @return         True iff the user can manage the job offer
	 */
	public boolean canManageJob(Long offerId) {
		Long editionId = jobOfferService.getEditionFromJobOffer(offerId);
		return canManageEdition(editionId);
	}

	/**
	 * Checks if a user can manage an extra work.
	 *
	 * @param  workId The id of the extra work
	 * @return        True iff the user can manage the extra work
	 */
	public boolean canManageWork(Long workId) {
		Long editionId = extraWorkService.getEditionFromExtraWork(workId);
		return canManageEdition(editionId);
	}

	/**
	 * Checks if the user can manage an edition.
	 *
	 * @param  editionId The id of the edition
	 * @return           True iff the teacher is allowed to manage the edition
	 */
	public boolean canManageEdition(Long editionId) {
		if (isAdmin() || isCoordinatorForEdition(editionId)) {
			return true;
		}
		Long programId = courseService.getOrThrow(editionService.getOrThrow(editionId).getCourse().getId())
				.getProgram().getId();
		if (!programService.getTAMProgram(programId).getTeachersHaveAccess()) {
			return false;
		}
		List<RoleDetailsDTO> roles = roleService.getRolesById(List.of(editionId),
				List.of(getAuthPerson().getId()));
		return roles.size() == 1 && roles.get(0).getType().equals(RoleDetailsDTO.TypeEnum.TEACHER);
	}

	/**
	 * Checks if a user can submit an application for a job offer.
	 *
	 * @param  offerId The id of the job offer
	 * @return         True iff the user can submit a job offer
	 */
	public boolean canSubmitApplication(Long offerId) {
		JobOffer jobOffer = jobOfferService.findByIdOrThrow(offerId);
		return !applicationService.appExistsFor(getAuthPerson().getId(), offerId) &&
				!jobOffer.getDraft() &&
				!canManageEdition(jobOffer.getEditionId()) &&
				(jobOffer.getDeadline() == null || !LocalDate.now().isAfter(jobOffer.getDeadline()));
	}

	/**
	 * Checks if an applications exists for that job offer for current authed person
	 *
	 * @param  offerId The job offer for which the application should exist
	 * @return         True if an application exists
	 */
	public boolean applicationExistsFor(Long offerId) {
		return applicationService.appExistsFor(getAuthPerson().getId(), offerId);
	}

	/**
	 * checks if person has status offered for a job offer.
	 *
	 * @param  offerId id of the job offer
	 * @return         true if status is offered
	 */
	public boolean offerExistsFor(Long offerId) {
		return applicationService.hasOffer(getAuthPerson().getId(), offerId);
	}

	/**
	 * Checks if a person owns a declaration by id.
	 *
	 * @param  declarationId id of the declaration
	 * @return               true if person owns the declaration
	 */
	public boolean ownsDeclaration(long declarationId) {
		return declarationService.personOwnsDeclaration(getAuthPerson().getId(), declarationId);
	}

	/**
	 * Checks if a person can manage a declaration.
	 *
	 * @param  declarationId The id of the declaration
	 * @return               True iff person can manage the declaration
	 */
	public boolean canManageDeclaration(long declarationId) {
		return canManageWork(declarationService.getExtraWorkIdFromDeclaration(declarationId));
	}

	/**
	 * Checks whether the user is authorised to approve TA training.
	 *
	 * Note: When changing this, check the controller for ta training, since the actual filtering by programme
	 * is done there, this mainly just determines who can make requests to the endpoints and see the page as a
	 * whole.
	 *
	 * @return true if the user is authorised
	 */
	public boolean canEditTaTraining() {
		return isCoordinatorOfAny();
	}

	/**
	 * Checks whether the user is authorised to approve TA training.
	 *
	 * @return true if the user is authorised
	 */
	public boolean canEditTaTrainingForProgram(Long programId) {
		return isCoordinatorForProgram(programId);
	}

	/**
	 * Checks whether the user is authorised to edit a RaiseRequest claim.
	 *
	 * Note: When changing this, check the controller for raise requests, since the actual filtering by
	 * programme is done there, this mainly just determines who can make requests to the endpoints and see the
	 * page as a whole.
	 *
	 * @return true if the user is authorised
	 */
	public boolean canEditPayScale() {
		return isCoordinatorOfAny();
	}

	/**
	 * Checks whether the user is authorised to edit a RaiseRequest claim.
	 *
	 * @return true if the user is authorised
	 */
	public boolean canEditPayScaleForProgram(Long programId) {
		return isCoordinatorForProgram(programId);
	}

	/**
	 * Checks whether the user is authorised to see a certain raise request by id. The user either owns the
	 * raise request or can edit all raise requests.
	 *
	 * @param  id the id of the raise request
	 * @return    true if the user is authorised
	 */
	public boolean canDownloadProof(Long id) {
		return canEditPayScale()
				|| raiseRequestService.getBy(id).getRequestedBy().getId().equals(getAuthPerson().getId());
	}

	/**
	 * Checks whether the authenticated user can manage a course.
	 *
	 * @param  courseId The course for which to check
	 * @return          True iff the user can manage the course
	 */
	public boolean canManageCourse(Long courseId) {
		if (isAdmin() || isCoordinatorForCourse(courseId)) {
			return true;
		}
		Long programId = courseService.getOrThrow(courseId).getProgram().getId();
		if (!programService.getTAMProgram(programId).getTeachersHaveAccess()) {
			return false;
		}
		return courseService.getManagingCourses(new PersonIdDTO().id(getAuthPerson().getId()))
				.stream().anyMatch(course -> Objects.equals(courseId, course.getId()));
	}

	/**
	 * Checks whether the authenticated user can create an edition of a course.
	 *
	 * @param  courseId The id of the course
	 * @return          True iff the user can create an edition of the course
	 */
	public boolean canCreateEdition(Long courseId) {
		return canManageCourse(courseId);
	}

	/**
	 * Checks whether the authenticated user can declare hours for extra work.
	 *
	 * @param  workId The id of the extra work
	 * @return        True iff the user can declare hours
	 */
	public boolean canDeclareHours(Long workId) {
		Long editionId = extraWorkService.getById(workId).getEdition().getId();
		return isAdmin() || roleService.hasAnyTARole(getAuthPerson().getId(), editionId);
	}

}
