/*
 * TAM
 * Copyright (C) 2021 - Delft University of Technology
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package nl.tudelft.tam.service;

import java.util.ArrayList;
import java.util.List;
import java.util.Objects;
import java.util.stream.Collectors;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import nl.tudelft.labracore.api.CohortControllerApi;
import nl.tudelft.labracore.api.dto.CohortDetailsDTO;
import nl.tudelft.labracore.api.dto.CohortSummaryDTO;
import nl.tudelft.tam.cache.CohortCacheManager;

@Service
public class CohortService {

	@Autowired
	private CohortControllerApi cohortControllerApi;

	@Autowired
	private CohortCacheManager cohortCache;

	/**
	 * Gets all cohorts in the database.
	 *
	 * @return The list of CohortViewDTOs
	 */
	public List<CohortSummaryDTO> getAllCohorts() {
		return cohortControllerApi.getAllCohorts().collectList().block();
	}

	/**
	 * Gets a list of cohorts corresponding to a list of ids.
	 *
	 * @param  ids The cohort ids
	 * @return     The list of cohort details
	 */
	public List<CohortDetailsDTO> getCohortsById(List<Long> ids) {
		return cohortControllerApi.getAllCohortsById(ids).collectList().block();
	}

	/**
	 * Gets cohort details by id.
	 *
	 * @return The CohortDetailsDTO object
	 */
	public CohortDetailsDTO getOrThrowById(Long id) {
		return cohortCache.getOrThrow(id);
	}

	/**
	 * Gets a list of cohorts with provided ids
	 *
	 * @param  ids The list of ids to look for
	 * @return     The list of cohort details objects found
	 */
	public List<CohortDetailsDTO> get(List<Long> ids) {
		ids = ids.stream().distinct().collect(Collectors.toList());
		return cohortCache.get(ids);
	}

	/**
	 * Gets all the cohorts for a list of programmes
	 *
	 * @param  programmeIds The list of programme ids
	 * @return              The list of cohorts
	 */
	public List<CohortSummaryDTO> getCohortsByProgrammeIds(List<Long> programmeIds) {
		List<CohortSummaryDTO> result = new ArrayList<>();
		for (Long id : programmeIds) {
			result.addAll(Objects
					.requireNonNull(cohortControllerApi.getAllCohortsInProgram(id).collectList().block()));
		}
		return result;
	}
}
