/*
 * TAM
 * Copyright (C) 2021 - Delft University of Technology
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package nl.tudelft.tam.service;

import java.util.List;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.rest.webmvc.ResourceNotFoundException;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import nl.tudelft.librador.dto.DTOConverter;
import nl.tudelft.tam.dto.patch.CoordinatorDefaultPatchDTO;
import nl.tudelft.tam.dto.patch.embeddables.ExtraWorkDefaultPatchDTO;
import nl.tudelft.tam.dto.patch.embeddables.JobOfferDefaultPatchDTO;
import nl.tudelft.tam.dto.view.details.EditionDetailsOffersWorkDTO;
import nl.tudelft.tam.model.CoordinatorDefault;
import nl.tudelft.tam.model.TAMProgram;
import nl.tudelft.tam.model.embeddables.ExtraWorkDefault;
import nl.tudelft.tam.model.embeddables.JobOfferDefault;
import nl.tudelft.tam.repository.CoordinatorDefaultRepository;

@Service
public class CoordinatorDefaultService {

	@Autowired
	CoordinatorDefaultRepository repository;

	@Autowired
	CohortService cohortService;

	@Autowired
	ProgramService programService;

	@Autowired
	DTOConverter converter;

	/**
	 * Finds the coordinator default by id
	 *
	 * @param  id The id of the coordinator default to find
	 * @return    The coordinator default object found
	 */
	public CoordinatorDefault findByIdOrThrow(Long id) {
		return repository.findByIdOrThrow(id);
	}

	/**
	 * Either gets the existing defaults from the repo, or creates a new set of blank defaults for the
	 * programme
	 *
	 * @param  id The program id for which to get the defaults
	 * @return    The CoordinatorDefault object
	 */
	public CoordinatorDefault getOrCreate(Long id) {
		try {
			return findByIdOrThrow(id);
		} catch (ResourceNotFoundException e) {
			JobOfferDefault jobOfferDefault = JobOfferDefault.builder().contractName("").contractBaanCode("")
					.build();
			ExtraWorkDefault extraWorkDefault = ExtraWorkDefault.builder().contractName("")
					.contractBaanCode("").build();
			CoordinatorDefault coordinatorDefault = CoordinatorDefault.builder().programId(id)
					.jobOfferDefault(jobOfferDefault).extraWorkDefault(extraWorkDefault)
					.coordinatorContractEmail("").coordinatorContractName("").build();
			return repository.save(coordinatorDefault);
		}
	}

	/**
	 * Adds the coordinator defaults to each edition depending on the program id
	 *
	 * @param editions The list of editions for which to add the defaults
	 */
	public void addDefaultsToEditionsWithJobsAndWork(List<EditionDetailsOffersWorkDTO> editions) {
		for (EditionDetailsOffersWorkDTO edition : editions) {
			Long programId = edition.getProgram().getId();
			CoordinatorDefault defaults = getOrCreate(programId);
			edition.setDefaults(defaults);
		}
	}

	/**
	 * Patches the coordinator defaults for a program
	 *
	 * @param id  of the program to change the defaults for
	 * @param dto the changed defaults
	 */
	@Transactional
	public void patchCoordinatorDefault(Long id, CoordinatorDefaultPatchDTO dto) {
		converter.apply(dto, repository.findByIdOrThrow(id));
		TAMProgram program = programService.getTAMProgram(id);
		program.setTeachersHaveAccess(dto.getTeachersHaveAccess());
	}

	/**
	 * Create a patch DTO from a filled coordinator default
	 *
	 * @param  cd The coordinator default object to use
	 * @return    The patch DTO
	 */
	public CoordinatorDefaultPatchDTO createPatchFromCoordinatorDefault(CoordinatorDefault cd) {
		TAMProgram program = programService.getTAMProgram(cd.getProgramId());
		return CoordinatorDefaultPatchDTO.builder()
				.jobOfferDefaultPatchDTO(JobOfferDefaultPatchDTO.builder()
						.contractName(cd.getJobOfferDefault().getContractName())
						.contractBaanCode(cd.getJobOfferDefault().getContractBaanCode())
						.contractStartDate(cd.getJobOfferDefault().getContractStartDate())
						.contractEndDate(cd.getJobOfferDefault().getContractEndDate())
						.applicationDeadline(cd.getJobOfferDefault().getApplicationDeadline())
						.description(cd.getJobOfferDefault().getDescription())
						.hiringMessage(cd.getJobOfferDefault().getHiringMessage())
						.rejectMessage(cd.getJobOfferDefault().getRejectMessage())
						.build())
				.extraWorkDefaultPatchDTO(ExtraWorkDefaultPatchDTO.builder()
						.contractName(cd.getExtraWorkDefault().getContractName())
						.contractBaanCode(cd.getExtraWorkDefault().getContractBaanCode())
						.declarationDeadline(cd.getExtraWorkDefault().getDeclarationDeadline())
						.build())
				.coordinatorContractName(cd.getCoordinatorContractName())
				.coordinatorContractEmail(cd.getCoordinatorContractEmail())
				.teachersHaveAccess(program.getTeachersHaveAccess())
				.build();
	}
}
