/*
 * TAM
 * Copyright (C) 2021 - Delft University of Technology
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package nl.tudelft.tam.service;

import java.util.List;
import java.util.Objects;
import java.util.Set;
import java.util.stream.Collectors;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import nl.tudelft.labracore.api.PersonControllerApi;
import nl.tudelft.labracore.api.dto.PersonSummaryDTO;
import nl.tudelft.tam.model.Coordinator;
import nl.tudelft.tam.repository.CoordinatorRepository;

@Service
public class CoordinatorService {

	@Autowired
	CoordinatorRepository coordinatorRepository;

	@Autowired
	EditionService editionService;

	@Autowired
	CourseService courseService;

	@Autowired
	PersonControllerApi personApi;

	/**
	 * Finds all coordinator objects by person id
	 *
	 * @param  personId The person id to search by
	 * @return          The list of coordinator objects
	 */
	public List<Coordinator> findAllByPersonId(Long personId) {
		return coordinatorRepository.findAllByPersonId(personId);
	}

	/**
	 * Checks if a coordinator object with provided person id exists
	 *
	 * @param  personId The person id to check for
	 * @return          True, if the person coordinates any course
	 */
	public boolean existsByPersonId(Long personId) {
		return coordinatorRepository.existsByPersonId(personId);
	}

	/**
	 * Checks if a coordinator object exists with provided person and program ids
	 *
	 * @param  personId  The person id to check for
	 * @param  programId The program id to check for
	 * @return           True, if the person is a coordinator for that program
	 */
	public boolean existsByPersonIdAndProgramId(Long personId, Long programId) {
		return coordinatorRepository.existsByPersonIdAndProgramId(personId, programId);
	}

	/**
	 * Finds all program ids by person id
	 *
	 * @param  personId The person id to search by
	 * @return          The list of program ids
	 */
	public Set<Long> getProgramIdsByPersonId(Long personId) {
		return coordinatorRepository.findAllByPersonId(personId).stream()
				.map(Coordinator::getProgramId).collect(Collectors.toSet());
	}

	/**
	 * Finds all coordinators by program id
	 *
	 * @param  programId The program id to search by
	 * @return           The list of coordinators
	 */
	public List<Coordinator> getCoordinatorsByProgram(Long programId) {
		return coordinatorRepository.findAllByProgramId(programId);
	}

	/**
	 * Finds all coordinators by edition id
	 *
	 * @param  editionId The edition id to search by
	 * @return           The list of coordinators
	 */
	public List<Coordinator> getCoordinatorsByEditionId(Long editionId) {
		Long courseId = editionService.getEditionById(editionId).getCourse().getId();
		Long programId = courseService.getOrThrow(courseId).getProgram().getId();
		return coordinatorRepository.findAllByProgramId(programId);
	}

	/**
	 * Sets the coordinators for the given programme to the people corresponding to the specified usernames.
	 *
	 * @param programId The id of the programme
	 * @param usernames The usernames of the new coordinators
	 */
	@Transactional
	public void setCoordinatorsForProgram(Long programId, List<String> usernames) {
		Set<PersonSummaryDTO> coordinators = personApi.searchForPeople(usernames).block().getPeople();
		for (Coordinator coordinator : coordinatorRepository.findAllByProgramId(programId)) {
			if (coordinators.stream().noneMatch(c -> Objects.equals(coordinator.getId(), c.getId()))) {
				coordinatorRepository.delete(coordinator);
			}
		}
		for (PersonSummaryDTO coordinator : coordinators) {
			if (!coordinatorRepository.existsByPersonIdAndProgramId(coordinator.getId(), programId)) {
				coordinatorRepository.save(
						Coordinator.builder().programId(programId).personId(coordinator.getId()).build());
			}
		}
	}

}
