/*
 * TAM
 * Copyright (C) 2021 - Delft University of Technology
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package nl.tudelft.tam.service;

import java.util.Collections;
import java.util.List;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import jakarta.validation.constraints.NotNull;
import nl.tudelft.labracore.api.CourseControllerApi;
import nl.tudelft.labracore.api.dto.CourseDetailsDTO;
import nl.tudelft.labracore.api.dto.CourseSummaryDTO;
import nl.tudelft.labracore.api.dto.PersonIdDTO;
import nl.tudelft.tam.cache.CourseCacheManager;

@Service
public class CourseService {

	@Autowired
	private CourseControllerApi courseApi;

	@Autowired
	private CourseCacheManager courseCache;

	/**
	 * Gets all the courses a person is manager of.
	 *
	 * @param  personId The id of the person
	 * @return          The list of CourseViewDTOs the person manages
	 */
	public List<CourseSummaryDTO> getManagingCourses(PersonIdDTO personId) {
		return courseApi.getAllCoursesByManager(personId.getId()).collectList().block();
	}

	/**
	 * Gets a specific course or throws an exception
	 *
	 * @param  id The id of the course
	 * @return    The course detail of the course
	 */
	@NotNull
	public CourseDetailsDTO getOrThrow(Long id) {
		return courseCache.getOrThrow(id);
	}

	/**
	 * Gets all courses corresponding to a list of ids.
	 *
	 * @param  ids The list of course ids
	 * @return     The list of course details
	 */
	public List<CourseDetailsDTO> getCoursesById(List<Long> ids) {
		if (ids.isEmpty())
			return Collections.emptyList();
		return courseApi.getAllCoursesById(ids).collectList().block();
	}

	/**
	 * Gets a course by id.
	 *
	 * @param  id the course id
	 * @return    The course details
	 */
	public CourseDetailsDTO getCourseById(Long id) {
		return courseApi.getCourseById(id).block();
	}

	/**
	 * Gets all courses.
	 *
	 * @return The list of course summaries
	 */
	public List<CourseSummaryDTO> getAllCourses() {
		return courseApi.getAllCourses().collectList().block();
	}

	/**
	 * Gets all courses in a program.
	 *
	 * @param  programId The id of the program
	 * @return           The list of course summaries
	 */
	public List<CourseSummaryDTO> getAllCoursesInProgram(Long programId) {
		return courseApi.getAllCoursesByProgram(programId).collectList().block();
	}
}
