/*
 * TAM
 * Copyright (C) 2021 - Delft University of Technology
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package nl.tudelft.tam.service;

import java.util.List;
import java.util.Locale;
import java.util.concurrent.CompletableFuture;
import java.util.stream.Collectors;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.mail.javamail.JavaMailSender;
import org.springframework.mail.javamail.MimeMessageHelper;
import org.springframework.scheduling.annotation.Async;
import org.springframework.security.core.context.SecurityContextHolder;
import org.springframework.stereotype.Service;
import org.thymeleaf.TemplateEngine;
import org.thymeleaf.context.Context;

import jakarta.mail.MessagingException;
import jakarta.mail.internet.MimeMessage;
import nl.tudelft.labracore.api.dto.PersonSummaryDTO;
import nl.tudelft.labracore.lib.security.LabradorUserDetails;
import nl.tudelft.tam.dto.create.EmailCreateDTO;

@Service
public class EmailService {

	@Autowired
	private JavaMailSender emailSender;

	@Autowired
	private TemplateEngine emailTemplateEngine;

	@Value("${spring.mail.from-address}")
	String sender;

	/**
	 * Sends an email to a list of recipients.
	 *
	 * @param  recipients    The email addresses of the recipients
	 * @param  messageCreate The message to send
	 * @return               A future when all emails are sent
	 */
	@Async
	public CompletableFuture<Void> sendEmail(List<String> recipients, EmailCreateDTO messageCreate) {
		final Context ctx = new Context(Locale.ENGLISH); // TODO this could be a user preference
		ctx.setVariables(messageCreate.getContextVariables());

		try {
			final MimeMessage mimeMessage = this.emailSender.createMimeMessage();
			final MimeMessageHelper message = new MimeMessageHelper(mimeMessage, true, "UTF-8");

			message.setSubject(messageCreate.getSubject());

			final String htmlContent = this.emailTemplateEngine.process(messageCreate.getTemplate(), ctx);
			message.setText(htmlContent, true);

			for (String recipient : recipients) {
				message.setFrom(sender);
				message.setTo(recipient);
				this.emailSender.send(mimeMessage);
			}

		} catch (MessagingException e) {
			e.printStackTrace();
		}

		return CompletableFuture.completedFuture(null);
	}

	/**
	 * Sends an email to the currently authenticated person.
	 *
	 * @param  messageCreate The message to send
	 * @return               A future when the email is sent
	 */
	@Async
	public CompletableFuture<Void> sendEmailToAuthPerson(EmailCreateDTO messageCreate) {
		return sendEmail(List.of(((LabradorUserDetails) SecurityContextHolder.getContext()
				.getAuthentication().getPrincipal()).getUser().getEmail()), messageCreate);
	}

	/**
	 * Sends an email to a list of recipients.
	 *
	 * @param  people        The recipients
	 * @param  messageCreate The message to send
	 * @return               A future when all emails are sent
	 */
	@Async
	public CompletableFuture<Void> sendEmailToPeople(List<PersonSummaryDTO> people,
			EmailCreateDTO messageCreate) {
		return sendEmail(people.stream().map(PersonSummaryDTO::getEmail).collect(Collectors.toList()),
				messageCreate);
	}

}
