/*
 * TAM
 * Copyright (C) 2021 - Delft University of Technology
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package nl.tudelft.tam.service;

import java.io.IOException;
import java.nio.file.Path;
import java.time.LocalDate;
import java.time.LocalDateTime;
import java.util.*;
import java.util.stream.Collectors;

import org.modelmapper.ModelMapper;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.context.annotation.Lazy;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.web.multipart.MultipartFile;

import com.opencsv.exceptions.CsvValidationException;

import jakarta.validation.constraints.NotNull;
import nl.tudelft.labracore.api.dto.*;
import nl.tudelft.librador.dto.DTOConverter;
import nl.tudelft.tam.dto.create.ExtraWorkCreateDTO;
import nl.tudelft.tam.dto.patch.ExtraWorkPatchDTO;
import nl.tudelft.tam.dto.view.details.ExtraWorkDetailsDTO;
import nl.tudelft.tam.dto.view.summary.ExtraWorkSummaryDTO;
import nl.tudelft.tam.exception.ConstraintViolationException;
import nl.tudelft.tam.exception.WrongDataCallException;
import nl.tudelft.tam.model.Declaration;
import nl.tudelft.tam.model.ExtraWork;
import nl.tudelft.tam.repository.ExtraWorkRepository;

@Service
public class ExtraWorkService {

	// region class attributes

	@Autowired
	ModelMapper mapper;

	@Autowired
	ExtraWorkRepository extraWorkRepository;

	@Lazy
	@Autowired
	DeclarationService declarationService;

	@Autowired
	EditionService editionService;

	@Autowired
	CourseService courseService;

	@Autowired
	CohortService cohortService;

	@Autowired
	CSVService csvService;

	@Autowired
	DTOConverter converter;

	@Value("${tam.filesys.imports-dir}")
	private String importDirectory;

	// endregion

	/**
	 * Returns one detailed view of ExtraWork by its id.
	 *
	 * @param  id of the job offer
	 * @return    detailed view of job offer.
	 */
	public ExtraWorkDetailsDTO getById(Long id) {
		ExtraWorkDetailsDTO dto = mapper.map(extraWorkRepository.findByIdOrThrow(id),
				ExtraWorkDetailsDTO.class);
		dto.setEdition(editionService.getEditionById(dto.getEditionId()));
		return dto;
	}

	/**
	 * Finds a job offer by id and returns the id of related edition.
	 *
	 * @param  id of the job offer
	 * @return    id of the related edition
	 */
	@NotNull
	public Long getEditionFromExtraWork(Long id) {
		return extraWorkRepository.findByIdOrThrow(id).getEditionId();
	}

	/**
	 * Find all open job offers and return them in summary DTO.
	 *
	 * @return list of open job offers
	 */
	public List<ExtraWorkSummaryDTO> getAllOpen() {
		return extraWorkRepository.findAllOpen().stream()
				.map(o -> mapper.map(o, ExtraWorkSummaryDTO.class))
				.collect(Collectors.toList());
	}

	/**
	 * Find all closed job offers and return them in summary DTO.
	 *
	 * @return list of closed job offers
	 */
	public List<ExtraWorkSummaryDTO> getAllClosed() {
		return extraWorkRepository.findAllClosed().stream()
				.map(o -> mapper.map(o, ExtraWorkSummaryDTO.class))
				.collect(Collectors.toList());
	}

	private void validateExtraWork(ExtraWork extraWork) {
		Set<String> errorMessages = new HashSet<>();

		LocalDate deadline = extraWork.getDeadline();
		if (deadline != null && LocalDate.now().isAfter(deadline)) {
			errorMessages.add("Deadline cannot be in the past.");
		}

		if (extraWork.getBaanCode().replaceAll("\\s+", "").length() != 9) {
			errorMessages.add("Baan code must be 9 characters (excluding spaces).");
		}

		if (!errorMessages.isEmpty()) {
			throw new ConstraintViolationException(errorMessages);
		}
	}

	/**
	 * Saves a new Job Offer entry to the database and returns its id.
	 *
	 * @param  dto the new job offer to be saved
	 * @return     the id of the new job offer
	 */
	@Transactional
	public Long newExtraWork(ExtraWorkCreateDTO dto) {
		validateExtraWork(converter.apply(dto));
		return extraWorkRepository.save(converter.apply(dto)).getId();
	}

	/**
	 * Patches a job offer.
	 *
	 * @param id  of the job offer to be patched.
	 * @param dto the changed job offer.
	 */
	@Transactional
	public void patchExtraWork(Long id, ExtraWorkPatchDTO dto) {
		ExtraWork extraWork = extraWorkRepository.findByIdOrThrow(id);
		converter.apply(dto, extraWork);
		validateExtraWork(extraWork);
	}

	/**
	 * Closes a job offer by setting the deadline to now.
	 *
	 * @param id of the offer to be closed
	 */
	@Transactional
	public void closeExtraWork(Long id) {
		ExtraWork extraWork = extraWorkRepository.findByIdOrThrow(id);
		extraWork.setDeadline(LocalDate.now().minusDays(1));
	}

	/**
	 * Returns the list of open job offers filtered by some query.
	 *
	 * @param  filter the query
	 * @return        the filtered list of open job offers.
	 */
	public List<ExtraWorkDetailsDTO> getFilteredOpen(String filter) {
		return getFilteredFor(extraWorkRepository.findAllOpen(), filter);
	}

	/**
	 * Returns the list of closed job offers filtered by some query.
	 *
	 * @param  filter the query
	 * @return        the filtered list of closed job offers
	 */
	public List<ExtraWorkDetailsDTO> getFilteredClosed(String filter) {
		return getFilteredFor(extraWorkRepository.findAllClosed(), filter);
	}

	/**
	 * Retrieves data about the linked edition for a list of job offers from external database, links them
	 * together and filters the results.
	 *
	 * @param  offers the list of job offers
	 * @param  filter the filter
	 * @return        the filtered list of job offers details with editions
	 */
	public List<ExtraWorkDetailsDTO> getFilteredFor(List<ExtraWork> offers, String filter) {
		if (offers.isEmpty()) {
			return new ArrayList<>();
		}

		Map<Long, EditionDetailsDTO> editions = editionService
				.getEditionsById(offers.stream().map(ExtraWork::getEditionId).collect(Collectors.toList()))
				.stream().collect(Collectors.toMap(EditionDetailsDTO::getId, e -> e));

		String f = filter == null ? "" : filter.toLowerCase();

		ArrayList<ExtraWorkDetailsDTO> detailedOffers = new ArrayList<>(offers.size());
		for (ExtraWork o : offers) {
			EditionDetailsDTO e = editions.get(o.getEditionId());
			if (e == null)
				throw new WrongDataCallException(
						"There appears to be job offer linked to edition that does not exist in labracore.");

			if (!(f.isBlank() || o.getName().toLowerCase().contains(f)
					|| e.getName().toLowerCase().contains(f)
					|| e.getCourse().getName().toLowerCase().contains(f)
					|| e.getCourse().getCode().toLowerCase().contains(f)
					|| e.getCohort().getName().toLowerCase().contains(f)))
				continue;
			ExtraWorkDetailsDTO detail = mapper.map(o, ExtraWorkDetailsDTO.class);
			detail.setEdition(e);
			detailedOffers.add(detail);
		}

		return detailedOffers;
	}

	/**
	 * Retrieves a list of all job offers in edition.
	 *
	 * @param  editionIds the id of the edition
	 * @return            list of job offers
	 */
	public List<ExtraWorkSummaryDTO> getAllInEditions(List<Long> editionIds) {
		return extraWorkRepository.findAllByEditions(editionIds).stream()
				.map(o -> mapper.map(o, ExtraWorkSummaryDTO.class))
				.collect(Collectors.toList());
	}

	/**
	 * Retrieves a list of all job offers in edition, filtered by a query.
	 *
	 * @param  editionIds the id of the edition
	 * @param  filter     the query
	 * @return            filtered list of job offers
	 */
	public List<ExtraWorkDetailsDTO> getFilteredAllInEditions(List<Long> editionIds, String filter) {
		return getFilteredFor(extraWorkRepository.findAllByEditions(editionIds), filter);
	}

	/**
	 * Exports all declarations of some extra work.
	 *
	 * @param  id The id of the extra work
	 * @return    The Path to the export file
	 */
	public Path exportExtraWork(Long id) throws IOException {
		ExtraWork work = extraWorkRepository.findByIdOrThrow(id);

		List<Declaration> declarations = declarationService.findAllByWorkId(id);

		return declarationService.exportDeclarations(work.getName(), declarations, true, true);
	}

	/**
	 * Finds the extra work by id
	 *
	 * @param  id The id of the extra work to find
	 * @return    The extra work object found
	 */
	@NotNull
	public ExtraWork findByIdOrThrow(Long id) {
		return extraWorkRepository.findByIdOrThrow(id);
	}

	/**
	 * Imports a list of extra work from a CSV file.
	 *
	 * @param  file                   The csv file
	 * @throws IOException            When the file is invalid/corrupted
	 * @throws CsvValidationException When the format of the file is invalid
	 * @return                        The list of created extra work objects
	 */
	public List<ExtraWork> importExtraWork(MultipartFile file) throws IOException, CsvValidationException {
		Path directory = Path.of(importDirectory);

		if (!directory.toFile().exists()) {
			directory.toFile().mkdirs();
		}

		file.transferTo(Path.of(importDirectory + "/" + file.getOriginalFilename()));

		List<String[]> lines = csvService
				.readCSV(Path.of(importDirectory + "/" + file.getOriginalFilename()), false, 1);

		List<EditionSummaryDTO> activeEditions = editionService.getAllActiveEditions(LocalDateTime.now());
		List<EditionDetailsDTO> editions = editionService.getEditionsById(activeEditions.stream()
				.filter(x -> !x.getIsArchived()).map(EditionSummaryDTO::getId).toList());

		List<ExtraWork> result = new ArrayList<>();

		for (String[] line : lines) {
			ExtraWorkCreateDTO work = new ExtraWorkCreateDTO();

			String editionName = line[0];
			String courseCode = line[1];

			EditionDetailsDTO edition = editions
					.stream()
					.filter(e -> e.getCourse().getCode().equals(courseCode)
							&& e.getName().equals(editionName))
					.findAny()
					.get();

			work.setEditionId(edition.getId());
			work.setName(line[2]);
			work.setContractName(line[3]);
			work.setBaanCode(line[4]);
			work.setMaxHours(Integer.parseInt(line[5]));
			work.setDeadline(LocalDate.parse(line[6]));

			result.add(extraWorkRepository.findByIdOrThrow(newExtraWork(work)));
		}
		return result;
	}
}
