/*
 * TAM
 * Copyright (C) 2021 - Delft University of Technology
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package nl.tudelft.tam.service;

import java.io.*;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.nio.file.StandardCopyOption;

import org.springframework.beans.factory.annotation.Value;
import org.springframework.core.io.Resource;
import org.springframework.core.io.UrlResource;
import org.springframework.stereotype.Service;
import org.springframework.util.StringUtils;
import org.springframework.web.multipart.MultipartFile;

import com.google.common.hash.Hashing;

@Service
public class FileService {

	@Value("${tam.filesys.block-size}")
	private Integer SIZE_OF_BLOCKS;

	@Value("${tam.filesys.level-depth}")
	private Integer NUM_0F_LEVELS;

	/**
	 * Generates a path for a given Submission identifier
	 *
	 * @param  id identifier of the Submission object
	 * @return    the path to where the corresponding Submission file is stored
	 */
	public String getDirectory(Long id) {
		String code = hash(id);

		StringBuilder path = new StringBuilder();

		for (int i = 0; i < NUM_0F_LEVELS; i++) {
			path.append(code, i * SIZE_OF_BLOCKS, (i + 1) * SIZE_OF_BLOCKS);
			path.append(File.separator);
		}

		return path.toString() + id + File.separator;
	}

	/**
	 * Generates a hash for the given ID.
	 *
	 * @param  id the identifier to hash
	 * @return    hash String
	 */
	public String hash(Long id) {
		return Hashing.sha256().hashLong(id).toString();
	}

	/**
	 * Uploads a file to the local file system.
	 *
	 * @param  file        the file to be uploaded
	 * @param  entityId    the identifier of the entity
	 * @param  directory   the name of the directory where file should be stored at the generated path
	 * @throws IOException if file could not be saved
	 */
	public void uploadFile(MultipartFile file, Long entityId, String directory) throws IOException {
		Path path = Paths
				.get(directory + File.separator
						+ StringUtils.cleanPath(getDirectory(entityId) + file.getOriginalFilename()));
		File f = new File(String.valueOf(path));
		f.mkdirs();
		Files.copy(file.getInputStream(), path, StandardCopyOption.REPLACE_EXISTING);
	}

	/**
	 * Downloads a submission file from the local file system.
	 *
	 * @param  entityId  the identifier of the entity
	 * @param  directory the name of the directory where file is stored
	 * @return           the ResponseEntity containing the submission-related file
	 */
	public Resource downloadFile(Long entityId, String directory) throws IOException {
		File dir = new File(directory + File.separator + StringUtils.cleanPath(getDirectory(entityId)));
		File[] files = dir.listFiles();

		if (files.length < 1)
			throw new IOException("No file found.");
		if (files.length > 1)
			throw new IOException("Only 1 file expected, while " + files.length + " were found.");

		return new UrlResource("file:" + files[0].getPath());
	}

	/**
	 * Deletes a directory and all subdirectories.
	 *
	 * @param directory The directory
	 */
	public void deleteDirectory(File directory) {
		if (!directory.exists())
			return;
		File[] files = directory.listFiles();
		if (files != null) {
			for (File file : files) {
				deleteDirectory(file);
			}
		}
		directory.delete();
	}

	/**
	 * Converts a string to a usable file name.
	 *
	 * @param  name The name to convert
	 * @return      The name usable as a file name
	 */
	public String toFileName(String name) {
		return StringUtils
				.cleanPath(name.toLowerCase().replaceAll("\\s+", "_").replaceAll("[^A-Za-z0-9_\\-]+", ""));
	}

}
