/*
 * TAM
 * Copyright (C) 2021 - Delft University of Technology
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package nl.tudelft.tam.service;

import java.io.IOException;
import java.nio.file.Path;
import java.time.LocalDate;
import java.util.*;
import java.util.function.Function;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import org.modelmapper.ModelMapper;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.context.annotation.Lazy;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.web.multipart.MultipartFile;

import com.opencsv.exceptions.CsvValidationException;

import jakarta.validation.constraints.NotNull;
import nl.tudelft.labracore.api.AcademicPeriodControllerApi;
import nl.tudelft.labracore.api.dto.*;
import nl.tudelft.librador.dto.DTOConverter;
import nl.tudelft.librador.dto.patch.Patch;
import nl.tudelft.tam.dto.create.JobOfferCreateDTO;
import nl.tudelft.tam.dto.create.MissingEditionsJobOfferImportDTO;
import nl.tudelft.tam.dto.view.details.JobOfferDetailsDTO;
import nl.tudelft.tam.dto.view.summary.JobOfferSummaryDTO;
import nl.tudelft.tam.enums.Status;
import nl.tudelft.tam.exception.ConstraintViolationException;
import nl.tudelft.tam.exception.WrongDataCallException;
import nl.tudelft.tam.io.ImportHandler;
import nl.tudelft.tam.io.JobOfferImportHandler;
import nl.tudelft.tam.model.Application;
import nl.tudelft.tam.model.JobOffer;
import nl.tudelft.tam.model.TrainingType;
import nl.tudelft.tam.repository.JobOfferRepository;
import nl.tudelft.tam.security.AuthorisationService;

@Service
public class JobOfferService {

	// region class attributes

	@Value("${tam.filesys.exports-dir}")
	String exportDir;

	@Autowired
	ModelMapper mapper;

	@Autowired
	JobOfferRepository jobOfferRepository;

	@Lazy
	@Autowired
	ApplicationService applicationService;

	@Autowired
	EditionService editionService;

	@Autowired
	CourseService courseService;

	@Lazy
	@Autowired
	ProgramService programService;

	@Autowired
	AcademicPeriodControllerApi academicPeriodApi;

	@Autowired
	PersonService personService;

	@Autowired
	CSVService csvService;

	@Autowired
	FileService fileService;

	@Lazy
	@Autowired
	RoleService roleService;

	@Autowired
	ProfileService profileService;

	@Autowired
	CohortService cohortService;

	@Autowired
	CoordinatorService coordinatorService;

	@Autowired
	CoordinatorDefaultService coordinatorDefaultService;

	@Autowired
	JobOfferImportHandler importHandler;

	@Autowired
	private TrainingTypeService trainingTypeService;

	@Autowired
	private DTOConverter converter;

	@Value("${tam.filesys.imports-dir}")
	private String importDirectory;
	@Autowired
	private AuthorisationService authorisationService;

	// endregion

	/**
	 * Returns one detailed view of JobOffer by its id.
	 *
	 * @param  id of the job offer
	 * @return    detailed view of job offer.
	 */
	public JobOfferDetailsDTO getById(Long id) {
		JobOfferDetailsDTO dto = converter.convert(jobOfferRepository.findByIdOrThrow(id),
				JobOfferDetailsDTO.class);
		dto.setEdition(editionService.getEditionById(dto.getEditionId()));
		return dto;
	}

	/**
	 * Finds a job offer by id and returns the id of related edition.
	 *
	 * @param  id of the job offer
	 * @return    id of the related edition
	 */
	@NotNull
	public Long getEditionFromJobOffer(Long id) {
		return jobOfferRepository.findByIdOrThrow(id).getEditionId();
	}

	/**
	 * Find all open job offers and return them in summary DTO.
	 *
	 * @return list of open job offers
	 */
	public List<JobOfferSummaryDTO> getAllOpen() {
		return jobOfferRepository.findAllOpen().stream()
				.map(o -> mapper.map(o, JobOfferSummaryDTO.class))
				.collect(Collectors.toList());
	}

	/**
	 * Find all closed job offers and return them in summary DTO.
	 *
	 * @return list of closed job offers
	 */
	public List<JobOfferSummaryDTO> getAllClosed() {
		return jobOfferRepository.findAllClosed().stream()
				.map(o -> mapper.map(o, JobOfferSummaryDTO.class))
				.collect(Collectors.toList());
	}

	public void validateJobOffer(JobOffer jobOffer) {
		Set<String> errorMessages = new HashSet<>();
		LocalDate deadline = jobOffer.getDeadline();
		if (deadline != null) {
			if (jobOffer.getCreatedDate().isAfter(deadline)) {
				errorMessages.add("Deadline must be after the job offer creation date.");
			}
			if (deadline.isAfter(jobOffer.getContractStartDate())) {
				errorMessages.add("Deadline for applying must be before the contract start date.");
			}
		}

		if (jobOffer.getContractEndDate().isBefore(jobOffer.getContractStartDate())) {
			errorMessages.add("Contract end date must be after the contract start date.");
		}

		if (jobOffer.getBaanCode().replaceAll("\\s+", "").length() != 9) {
			errorMessages.add("Baan code must be 9 characters (excluding spaces).");
		}

		if (!errorMessages.isEmpty()) {
			throw new ConstraintViolationException(errorMessages);
		}
	}

	/**
	 * Saves a new Job Offer entry to the database and returns its id.
	 *
	 * @param  dto the new job offer to be saved
	 * @return     the id of the new job offer
	 */
	@Transactional
	public Long newJobOffer(JobOfferCreateDTO dto) {
		validateJobOffer(converter.apply(dto));
		JobOffer offer = jobOfferRepository.save(converter.apply(dto));
		setDefaultTrainings(offer);
		return offer.getId();
	}

	/**
	 * Patches a job offer.
	 *
	 * @param id  of the job offer to be patched.
	 * @param dto the changed job offer.
	 */
	@Transactional
	public void patchJobOffer(Long id, Patch<JobOffer> dto) {
		JobOffer jobOffer = jobOfferRepository.findByIdOrThrow(id);
		converter.apply(dto, jobOffer);
		validateJobOffer(jobOffer);
	}

	/**
	 * Closes a job offer by setting the deadline to now.
	 *
	 * @param id of the offer to be closed
	 */
	@Transactional
	public void closeJobOffer(Long id) {
		JobOffer jobOffer = jobOfferRepository.findByIdOrThrow(id);
		jobOffer.setDeadline(LocalDate.now().minusDays(1));
	}

	/**
	 * Returns the list of open job offers filtered by some query.
	 *
	 * @param  filter the query
	 * @return        the filtered list of open job offers.
	 */
	public List<JobOfferDetailsDTO> getFilteredOpen(String filter) {
		return getFilteredFor(jobOfferRepository.findAllOpen(), filter);
	}

	/**
	 * Returns the list of closed job offers filtered by some query.
	 *
	 * @param  filter the query
	 * @return        the filtered list of closed job offers
	 */
	public List<JobOfferDetailsDTO> getFilteredClosed(String filter) {
		return getFilteredFor(jobOfferRepository.findAllClosed(), filter);
	}

	/**
	 * Retrieves data about the linked edition for a list of job offers from external database, links them
	 * together and filters the results.
	 *
	 * @param  offers the list of job offers
	 * @param  filter the filter
	 * @return        the filtered list of job offers details with editions
	 */
	public List<JobOfferDetailsDTO> getFilteredFor(List<JobOffer> offers, String filter) {
		if (offers.isEmpty()) {
			return new ArrayList<>();
		}

		Map<Long, EditionDetailsDTO> editions = editionService
				.getEditionsById(offers.stream().map(JobOffer::getEditionId).collect(Collectors.toList()))
				.stream().collect(Collectors.toMap(EditionDetailsDTO::getId, e -> e));

		String f = filter == null ? "" : filter.toLowerCase();

		ArrayList<JobOfferDetailsDTO> detailedOffers = new ArrayList<>(offers.size());
		for (JobOffer o : offers) {
			EditionDetailsDTO e = editions.get(o.getEditionId());
			if (e == null)
				throw new WrongDataCallException(
						"There appears to be job offer linked to edition that does not exist in labracore.");

			if (!(f.isBlank() || o.getName().toLowerCase().contains(f)
					|| e.getName().toLowerCase().contains(f)
					|| e.getCourse().getName().toLowerCase().contains(f)
					|| e.getCourse().getCode().toLowerCase().contains(f)
					|| e.getCohort().getName().toLowerCase().contains(f)))
				continue;
			JobOfferDetailsDTO detail = converter.convert(o, JobOfferDetailsDTO.class);
			detail.setEdition(e);
			detailedOffers.add(detail);
		}

		return detailedOffers;
	}

	/**
	 * Retrieves a list of all job offers in edition.
	 *
	 * @param  editionIds the id of the edition
	 * @return            list of job offers
	 */
	public List<JobOfferSummaryDTO> getAllInEditions(List<Long> editionIds) {
		return jobOfferRepository.findAllByEditions(editionIds).stream()
				.map(o -> converter.convert(o, JobOfferSummaryDTO.class))
				.collect(Collectors.toList());
	}

	/**
	 * Retrieves a list of all job offers in edition, filtered by a query.
	 *
	 * @param  editionIds the id of the edition
	 * @param  filter     the query
	 * @return            filtered list of job offers
	 */
	public List<JobOfferSummaryDTO> getFilteredAllInEditions(List<Long> editionIds, String filter) {
		return getFilteredFor(jobOfferRepository.findAllByEditions(editionIds), filter).stream()
				.map(o -> (JobOfferSummaryDTO) o).collect(Collectors.toList());
	}

	/**
	 * Gets the most common approve/disapprove action-maker on TAM since that person will also approve in
	 * FlexDelft
	 *
	 * @param  editionId the id of the edition for which to find the approver
	 * @return           The person who is the most common approver/disapprover on TAM for that edition
	 */
	public PersonSummaryDTO getApprover(Long editionId) {
		Stream<PersonSummaryDTO> coordinators = coordinatorService.getCoordinatorsByEditionId(editionId)
				.stream()
				.map(coord -> personService.getOrThrow(coord.getPersonId()));
		Stream<PersonSummaryDTO> teachers = roleService.getTeachersById(editionId)
				.stream()
				.map(RolePersonDetailsDTO::getPerson);
		Stream<PersonSummaryDTO> both = Stream.concat(teachers, coordinators).distinct();

		// Return the teacher with the most TAM actions
		return both.collect(Collectors.toMap(
				Function.identity(),
				p -> profileService.findById(p.getId())
						.map(profile -> profile.getHandledDeclarations(editionId).size()
								+ profile.getHandledApplications(editionId).size())
						.orElse(0),
				Integer::sum))
				.entrySet()
				.stream()
				.max(Map.Entry.comparingByValue())
				.map(Map.Entry::getKey)
				.orElse(null);
	}

	/**
	 * Sets the default trainings for a job offer to the programme default.
	 *
	 * @param offer The offer to configure
	 */
	private void setDefaultTrainings(JobOffer offer) {
		Long programId = courseService
				.getCourseById(editionService.getEditionById(offer.getEditionId()).getCourse().getId())
				.getProgram().getId();
		offer.setRequiredTrainings(trainingTypeService.getAllTrainingTypesForPrograms(List.of(programId))
				.stream().filter(TrainingType::getIsDefault).collect(Collectors.toSet()));
	}

	/**
	 * Find a job offer by id
	 *
	 * @param  id The job offer to find
	 * @return    The job offer object found
	 */
	@NotNull
	public JobOffer findByIdOrThrow(Long id) {
		return jobOfferRepository.findByIdOrThrow(id);
	}

	/**
	 * Exports all applications of a job offer.
	 *
	 * @param  id The id of the job offer
	 * @return    The Path to the export file
	 */
	public Path exportJobOffer(Long id) throws IOException {
		JobOffer offer = jobOfferRepository.findByIdOrThrow(id);

		List<Application> applications = applicationService.findAllByOfferId(id).stream()
				.filter(app -> !app.getStatus().equals(Status.REJECTED_BY_STUDENT))
				.sorted(Comparator.comparing(Application::getCreatedDate))
				.collect(Collectors.toList());

		return applicationService.exportApplications(offer.getName(), applications, true);
	}

	/**
	 * Imports a list of job offers from a CSV file. Format:<code>
	 * [0] edition name, [1] course code, [2] job name (TAM), [3] contract name (Flex)
	 * [4] contract start date, [5] contract end date,
	 * [6] baancode, [7] maximum hours, [8] hiring message,
	 * [9] reject message, [10] deadline
	 * (, [11] cohort name, [12] program name, [13] edition start date, [14] edition end date)
	 * </code>
	 *
	 * @param  file                   The csv file containing the job offers.
	 * @throws IOException            If the file is invalid/corrupted.
	 * @throws CsvValidationException If the file format is invalid.
	 */
	@Transactional
	public ImportHandler.Result<JobOffer> importJobOffers(MultipartFile file)
			throws IOException, CsvValidationException {
		List<String[]> lines = csvService.storeAndReadCSV(file, importDirectory);

		var result = importHandler.importFile(lines);
		result.result().forEach(this::setDefaultTrainings);
		jobOfferRepository.saveAll(result.result());
		return result;
	}

	/**
	 * Creates the job offers for editions that did not exist when importing a list of job offers.
	 *
	 * @param dto The DTO with to be created editions and job offers
	 */
	@Transactional
	public void createJobOffersWithMissingEditions(MissingEditionsJobOfferImportDTO dto) {
		dto.getIncludedEditions().forEach((edition, offers) -> {
			Long editionId = editionService.addEdition(edition);
			jobOfferRepository.saveAll(offers.stream().map(offer -> {
				offer.setEditionId(editionId);
				JobOffer created = converter.apply(offer);
				setDefaultTrainings(created);
				return created;
			}).toList());
		});
	}

	/**
	 * Toggles the draft state of a job offer.
	 *
	 * @param id The id of the job offer.
	 */
	@Transactional
	public void toggleHide(Long id) {
		JobOffer offer = jobOfferRepository.findByIdOrThrow(id);
		offer.setDraft(!offer.getDraft());
	}

	/**
	 * Adds a required training to a job offer.
	 *
	 * @param id     The id of the job offer
	 * @param typeId The id of the training type
	 */
	@Transactional
	public void addRequiredTraining(Long id, Long typeId) {
		JobOffer offer = jobOfferRepository.findByIdOrThrow(id);
		offer.getRequiredTrainings().add(trainingTypeService.getTrainingTypeById(typeId));
		jobOfferRepository.save(offer);
	}

	/**
	 * Removes a required training to a job offer.
	 *
	 * @param id     The id of the job offer
	 * @param typeId The id of the training type
	 */
	@Transactional
	public void removeRequiredTraining(Long id, Long typeId) {
		JobOffer offer = jobOfferRepository.findByIdOrThrow(id);
		offer.getRequiredTrainings().remove(trainingTypeService.getTrainingTypeById(typeId));
		jobOfferRepository.save(offer);
	}

	/**
	 * Gets TAs that do not have a job offer in an edition
	 *
	 * @param id The id of the edition
	 */
	public List<RolePersonDetailsDTO> getTAsWithoutAJobOfferInEdition(Long id) {
		return roleService.getTAsAndHeadTAsById(id).stream()
				.filter(x -> !applicationService.hasApplicationInEdition(x.getPerson().getId(), id)).toList();
	}

	/**
	 * Get all academic periods from CORE
	 *
	 * @return all academic periods
	 */
	public List<AcademicPeriodDetailsDTO> getAllAcademicPeriods() {
		return academicPeriodApi.getAll().collectList().block();
	}
}
