/*
 * TAM
 * Copyright (C) 2021 - Delft University of Technology
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package nl.tudelft.tam.service;

import static java.util.function.Function.identity;

import java.util.*;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import org.modelmapper.ModelMapper;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.core.ParameterizedTypeReference;
import org.springframework.data.rest.webmvc.ResourceNotFoundException;
import org.springframework.http.HttpMethod;
import org.springframework.stereotype.Service;
import org.springframework.web.client.RestTemplate;

import lombok.NonNull;
import nl.tudelft.labracore.api.PersonControllerApi;
import nl.tudelft.labracore.api.dto.PersonDetailsDTO;
import nl.tudelft.labracore.api.dto.PersonSummaryDTO;
import nl.tudelft.labracore.api.dto.RoleEditionDetailsDTO;
import nl.tudelft.labracore.app.service.AppConnectionService;
import nl.tudelft.tam.cache.PersonCacheManager;
import nl.tudelft.tam.model.Profile;

@Service
public class PersonService {

	@Autowired
	ProfileService profileService;

	@Autowired
	PersonControllerApi personControllerApi;

	@Autowired
	PersonCacheManager personCache;

	@Autowired
	AppConnectionService appConnectionService;

	@Autowired
	ModelMapper mapper;

	/**
	 * Gets the roles for a person
	 *
	 * @param  personId The person to get the roles for
	 * @return          The list of role details
	 */
	public List<RoleEditionDetailsDTO> getRolesForPerson(Long personId) {
		return personControllerApi.getRolesForPerson(personId)
				.collectList()
				.block();
	}

	/**
	 * Search for people by display name
	 *
	 * @param  displayName The name to search by
	 * @return             The list of people objects matching that display name
	 */
	public List<PersonSummaryDTO> searchPeopleByDisplayNameContaining(String displayName) {
		return personControllerApi
				.searchPeopleByDisplayNameContaining(displayName).collectList().block();
	}

	/**
	 * Search for people by username
	 *
	 * @param  username The username to search by
	 * @return          The list of people objects matching that username
	 */
	public List<PersonSummaryDTO> searchPeopleByUsernameContaining(String username) {
		return personControllerApi
				.searchPeopleByUsernameContaining(username).collectList().block();
	}

	/**
	 * Search for people by student number
	 *
	 * @param  studentNumber The student number to search by
	 * @return               The list of people objects matching that student number
	 */
	public List<PersonSummaryDTO> searchPeopleByNumberContaining(Integer studentNumber) {
		return personControllerApi
				.searchPeopleByNumberContaining(studentNumber).collectList().block();
	}

	/**
	 * Finds a list of people based on their identifiers (net id, name, student number)
	 *
	 * @param  identifier Identifiers as one string
	 * @return            The list of people objects matching the identifiers
	 */
	public List<PersonSummaryDTO> searchForPeople(String identifier) {
		if (identifier == null || identifier.isBlank())
			return Collections.emptyList();
		List<String> identifiers = Arrays.stream(identifier.split("[\n\r]+")).filter(s -> !s.isBlank())
				.toList();
		return Stream.of(
				personControllerApi.searchForPeople(identifiers).block().getPeople().stream(),
				identifiers.stream().flatMap(id -> searchPeopleByDisplayNameContaining(id).stream()))
				.flatMap(identity()).distinct().toList();
	}

	/**
	 * Get pairs of PersonSummaries and Profiles from a list of identifiers. This will also create profiles
	 * for people that haven't used TAM previously (ie. not logged in)
	 *
	 * @param  identifiers The list of identifiers to find the profiles and person summaries for
	 * @return             The pairs of person summaries and profiles
	 */
	public Map<PersonSummaryDTO, Profile> findPersonProfilePairs(
			String identifiers) {
		return searchForPeople(identifiers)
				.stream()
				.map(x -> {
					try {
						return new AbstractMap.SimpleEntry<>(x, profileService.findByIdOrThrow(x.getId()));
					} catch (ResourceNotFoundException e) {
						return new AbstractMap.SimpleEntry<>(x, profileService.createProfile(x.getId()));
					}
				})
				.collect(Collectors.toMap(Map.Entry::getKey, Map.Entry::getValue));
	}

	/**
	 * Get people by ids
	 *
	 * @param  ids The ids to search for
	 * @return     The list of people found
	 */
	public List<PersonSummaryDTO> getPeopleById(List<Long> ids) {
		ids = ids.stream().distinct().collect(Collectors.toList());
		return personCache.get(ids);
	}

	/**
	 * Fetch average star rating for a list of people from Queue
	 *
	 * @param  ids The ids of the persons
	 * @return     The average star ratings
	 */
	public Optional<Map<Long, Double>> getQueueFeedbacks(List<Long> ids) {
		RestTemplate restTemplate = appConnectionService.connect("queue");
		if (restTemplate == null) {
			return Optional.empty();
		}
		return Optional.ofNullable(restTemplate.exchange("/api/average-star-rating?ids={ids}",
				HttpMethod.GET,
				null,
				new ParameterizedTypeReference<Map<Long, Double>>() {
				},
				ids.toArray()).getBody());
	}

	/**
	 * Get a person by id
	 *
	 * @param  id The id of the person
	 * @return    The object of the person
	 */
	@NonNull
	public PersonSummaryDTO getOrThrow(Long id) {
		return personCache.getOrThrow(id);
	}

	/**
	 * Get a person by username
	 *
	 * @param  username The username of the person
	 * @return          The object of the person
	 */
	public PersonDetailsDTO getPersonByUsername(String username) {
		return mapper.map(personCache.get(username), PersonDetailsDTO.class);
	}

	/**
	 * Get a person by id
	 *
	 * @param  id The id of the person
	 * @return    The object of the person
	 */
	public PersonDetailsDTO getPersonById(Long id) {
		return mapper.map(personCache.get(id).get(), PersonDetailsDTO.class);
	}

	/**
	 * Get people by a stream of ids
	 *
	 * @param  ids The stream of ids
	 * @return     The list of people
	 */
	public List<PersonSummaryDTO> get(Stream<Long> ids) {
		return personCache.get(ids.distinct());
	}

}
