/*
 * TAM
 * Copyright (C) 2021 - Delft University of Technology
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package nl.tudelft.tam.service;

import static nl.tudelft.labracore.api.dto.RoleEditionDetailsDTO.TypeEnum.*;

import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import nl.tudelft.labracore.api.EditionControllerApi;
import nl.tudelft.labracore.api.PersonControllerApi;
import nl.tudelft.labracore.api.RoleControllerApi;
import nl.tudelft.labracore.api.dto.*;
import nl.tudelft.tam.cache.EditionRolesCacheManager;

@Service
public class RoleService {

	@Autowired
	JobOfferService jobOfferService;

	@Autowired
	PersonService personService;

	@Autowired
	EditionService editionService;

	@Autowired
	CourseService courseService;

	@Autowired
	RoleControllerApi roleControllerApi;

	@Autowired
	EditionControllerApi editionControllerApi;

	@Autowired
	PersonControllerApi personApi;

	@Autowired
	EditionRolesCacheManager editionRolesCache;

	/**
	 * Get all the TA and Head TA roles for a person.
	 *
	 * @param  personId The id of the person
	 * @return          The list of roles
	 */
	public List<RoleEditionDetailsDTO> getTARolesForPerson(Long personId) {
		return personApi.getRolesForPerson(personId).collectList().block().stream()
				.filter(r -> Set.of(TA, HEAD_TA).contains(r.getType())).toList();
	}

	/**
	 * Gets the teacher roles for a person
	 *
	 * @param  personId the person id
	 * @return          the teacher roles the person has had
	 */
	public List<RoleEditionDetailsDTO> getTeacherRolesForPerson(Long personId) {
		return personApi.getRolesForPerson(personId).collectList().block().stream()
				.filter(r -> r.getType() == TEACHER).toList();
	}

	/**
	 * Gets the staff roles for a person (e.g. previous work experience)
	 *
	 * @param  personId the person id
	 * @return          the staff roles the person has had
	 */
	public List<RoleEditionDetailsDTO> getStaffRolesForPerson(Long personId) {
		return personApi.getRolesForPerson(personId).collectList().block().stream()
				.filter(r -> Set.of(TA, HEAD_TA, TEACHER).contains(r.getType())).toList();
	}

	/**
	 * Checks if a person has a role in an edition
	 *
	 * @param  personId  The person for whom to check
	 * @param  editionId The edition to check in
	 * @return           True if the person has a role in that edition
	 */
	public boolean hasAnyRoleInEdition(Long personId, Long editionId) {
		return !editionRolesCache.getRolesInEdition(personId, editionId).getRoles().isEmpty();
	}

	/**
	 * Checks if a person has any role of those listed in a course
	 *
	 * @param  personId the person to check for
	 * @param  courseId the course to check in
	 * @param  roles    the roles to check for
	 * @return          true if any of the roles match, false otherwise
	 */
	public boolean hasRoleInCourse(Long personId, Long courseId, List<RolePersonDetailsDTO.TypeEnum> roles) {
		return editionService.getEditionsForCourse(courseId).stream()
				.anyMatch(editionId -> hasRoleInEdition(personId, editionId, roles));
	}

	/**
	 * Checks if a person has any role of those listed in an edition
	 *
	 * @param  personId  the person to check for
	 * @param  editionId the edition to check in
	 * @param  roles     the roles to check for
	 * @return           true if any of the roles match, false otherwise
	 */
	public boolean hasRoleInEdition(Long personId, Long editionId,
			List<RolePersonDetailsDTO.TypeEnum> roles) {
		return editionRolesCache.getRolesInEdition(personId, editionId).getRoles().stream()
				.map(RolePersonDetailsDTO::getType)
				.anyMatch(roles::contains);
	}

	/**
	 * Checks if a person has a TA or head TA role in an edition
	 *
	 * @param  personId  The person for whom to check
	 * @param  editionId The edition to check in
	 * @return           True if the person has a TA or head TA role in that edition
	 */
	public boolean hasAnyTARole(Long personId, Long editionId) {
		return hasRoleInEdition(personId, editionId,
				List.of(RolePersonDetailsDTO.TypeEnum.TA, RolePersonDetailsDTO.TypeEnum.HEAD_TA));
	}

	/**
	 * Checks if a person has a head TA role in an edition
	 *
	 * @param  personId  The person for whom to check
	 * @param  editionId The edition to check in
	 * @return           True if the person has a head TA role in that edition
	 */
	public boolean hasHeadTARole(Long personId, Long editionId) {
		return hasRoleInEdition(personId, editionId, List.of(RolePersonDetailsDTO.TypeEnum.HEAD_TA));
	}

	/**
	 * Remove the role of a person in the edition of the provided job offer
	 *
	 * @param personId The person whose role to remove
	 * @param offerId  The job offer whose edition to remove the person from
	 */
	public void removeRole(Long personId, Long offerId) {
		Long editionId = jobOfferService.getEditionFromJobOffer(offerId);

		if (hasAnyRoleInEdition(personId, editionId)) {
			editionService.removePersonFromEdition(editionId, personId);
		}
	}

	/**
	 * Change the role of a person in the edition of the provided job offer to TA
	 *
	 * @param personId The person whose role to change
	 * @param offerId  The job offer whose edition to change the role in
	 * @param demote   Whether to demote the person from Head TA to TA
	 */
	public void changeToTa(Long personId, Long offerId, boolean demote) {
		Long editionId = jobOfferService.getEditionFromJobOffer(offerId);

		if (hasAnyRoleInEdition(personId, editionId)) {
			// ensure student isn't demoted from HeadTA to TA on accepting.
			if (hasHeadTARole(personId, editionId) && !demote) {
				return;
			}

			roleControllerApi
					.patchRole(new RolePatchDTO().type(RolePatchDTO.TypeEnum.TA), personId, editionId)
					.block();
		} else {
			roleControllerApi.addRole(new RoleCreateDTO()
					.edition(new EditionIdDTO().id(editionId))
					.person(new PersonIdDTO().id(personId))
					.type(RoleCreateDTO.TypeEnum.TA)).block();
		}
	}

	/**
	 * Change the role of a person in the edition of the provided job offer to Head TA
	 *
	 * @param personId The person whose role to change
	 * @param offerId  The job offer whose edition to change the role in
	 */
	public void changeToHeadTa(Long personId, Long offerId) {
		Long editionId = jobOfferService.getEditionFromJobOffer(offerId);

		if (hasAnyRoleInEdition(personId, editionId)) {
			roleControllerApi
					.patchRole(new RolePatchDTO().type(RolePatchDTO.TypeEnum.HEAD_TA), personId, editionId)
					.block();
		} else {
			roleControllerApi.addRole(new RoleCreateDTO()
					.edition(new EditionIdDTO().id(editionId))
					.person(new PersonIdDTO().id(personId))
					.type(RoleCreateDTO.TypeEnum.HEAD_TA)).block();
		}
	}

	/**
	 * Change the role of a person in the edition of the provided job offer to Teacher
	 *
	 * @param personId  The person whose role to change
	 * @param editionId The edition to change the role in
	 */
	public void changeToTeacher(Long personId, Long editionId) {
		if (hasAnyRoleInEdition(personId, editionId)) {
			roleControllerApi
					.patchRole(new RolePatchDTO().type(RolePatchDTO.TypeEnum.TEACHER), personId, editionId)
					.block();
		} else {
			roleControllerApi.addRole(new RoleCreateDTO()
					.edition(new EditionIdDTO().id(editionId))
					.person(new PersonIdDTO().id(personId))
					.type(RoleCreateDTO.TypeEnum.TEACHER)).block();
		}
	}

	/**
	 * Get a list of roles by editions and people
	 *
	 * @param  editionIds The editions in which the roles must be
	 * @param  peopleIds  The people who the roles must belong to
	 * @return            The list of role details
	 */
	public List<RoleDetailsDTO> getRolesById(List<Long> editionIds, List<Long> peopleIds) {
		return roleControllerApi.getRolesById(new HashSet<>(editionIds), new HashSet<>(peopleIds))
				.collectList().block();
	}

	/**
	 * Get a list of people with the TEACHER role in an edition
	 *
	 * @param  editionId The edition for which to get the teachers
	 * @return           The list of teachers
	 */
	public List<RolePersonDetailsDTO> getTeachersById(Long editionId) {
		return roleControllerApi.getRolesByEditions(Set.of(editionId))
				.filter(role -> role.getType().equals(RolePersonDetailsDTO.TypeEnum.TEACHER))
				.collectList().block();
	}

	/**
	 * Get a list of people with the TA or HEAD_TA role in an edition
	 *
	 * @param  editionId The edition for which to get the tas
	 * @return           The list of tas
	 */
	public List<RolePersonDetailsDTO> getTAsAndHeadTAsById(Long editionId) {
		return roleControllerApi.getRolesByEditions(Set.of(editionId))
				.filter(x -> Set.of(RolePersonDetailsDTO.TypeEnum.TA,
						RolePersonDetailsDTO.TypeEnum.HEAD_TA).contains(x.getType()))
				.collectList().block();
	}

	/**
	 * Checks if a role is staff
	 *
	 * @param  role the role to check
	 * @return      if the role is staff
	 */
	public boolean isStaffRole(PersonSummaryDTO.DefaultRoleEnum role) {
		return Set.of(PersonSummaryDTO.DefaultRoleEnum.TEACHER, PersonSummaryDTO.DefaultRoleEnum.ADMIN)
				.contains(role);
	}

	/**
	 * Checks if the teacher is the only teacher in the given edition (Modified from Queue).
	 *
	 * @param  personId  the id of person whose role is checked to be teacher
	 * @param  editionId the if of the edition to check the role from
	 * @return           whether the person is the only teacher in the edition
	 */
	public boolean isTheOnlyTeacherInEdition(Long personId, Long editionId) {
		List<Long> teacherIds = getTeachersById(editionId).stream().map(x -> x.getPerson().getId()).toList();
		return teacherIds.contains(personId) && teacherIds.size() == 1;
	}

}
