/*
 * TAM
 * Copyright (C) 2021 - Delft University of Technology
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package nl.tudelft.tam.service;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

import org.modelmapper.ModelMapper;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.annotation.Lazy;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import nl.tudelft.labracore.api.dto.ProgramDetailsDTO;
import nl.tudelft.labracore.api.dto.ProgramSummaryDTO;
import nl.tudelft.librador.dto.DTOConverter;
import nl.tudelft.tam.dto.create.TrainingTypeCreateDTO;
import nl.tudelft.tam.dto.patch.TrainingTypePatchDTO;
import nl.tudelft.tam.dto.view.details.TrainingTypeDetailsDTO;
import nl.tudelft.tam.model.TrainingType;
import nl.tudelft.tam.repository.TrainingTypeRepository;

@Service
public class TrainingTypeService {

	@Autowired
	TrainingTypeRepository repository;

	@Lazy
	@Autowired
	TrainingApprovalService trainingApprovalService;

	@Autowired
	ProgramService programService;

	@Autowired
	ModelMapper mapper;

	@Autowired
	DTOConverter converter;

	/**
	 * Get all training types for programs
	 *
	 * @param  programIds The programs for which to get the training types
	 * @return            The list of training types
	 */
	public List<TrainingType> getAllTrainingTypesForPrograms(List<Long> programIds) {
		List<TrainingType> result = new ArrayList<>();
		for (Long programId : programIds) {
			result.addAll(repository.findByProgramId(programId));
		}

		return result.stream().distinct().collect(Collectors.toList());
	}

	/**
	 * Get the training type by id
	 *
	 * @param  id The id of the type to get
	 * @return    The training type
	 */
	public TrainingType getTrainingTypeById(Long id) {
		return repository.findByIdOrThrow(id);
	}

	/**
	 * Finds all associated training types by name whilst ignoring the case.
	 *
	 * @param  name      The name of the training type
	 * @param  programId The ID of the program
	 *
	 * @return           A list of all training types that correspond to the specified constraints.
	 */
	public List<TrainingType> getTrainingTypesByNameAndProgramId(String name, Long programId) {
		return repository.findByNameIgnoreCaseAndProgramId(name, programId);
	}

	/**
	 * Create a new training type based on a createDTO
	 *
	 * @param createDTO The dto containing the information about the new training type
	 */
	@Transactional
	public void createNewTrainingType(TrainingTypeCreateDTO createDTO) {
		repository.save(converter.apply(createDTO));
	}

	/**
	 * Update an existing training type
	 *
	 * @param patchDTO       The data to update
	 * @param trainingTypeId The id of the training type to update
	 */
	@Transactional
	public void updateTrainingType(TrainingTypePatchDTO patchDTO, Long trainingTypeId) {
		repository.save(converter.apply(patchDTO, repository.findByIdOrThrow(trainingTypeId)));
	}

	/**
	 * Delete an existing training type. This will also delete all training approvals for this training type.
	 *
	 * @param trainingTypeId The training type to delete
	 */
	@Transactional
	public void deleteTrainingType(Long trainingTypeId) {
		trainingApprovalService.deleteAllTrainingApprovalsForTrainingType(trainingTypeId);
		repository.deleteById(trainingTypeId);
	}

	/**
	 * Get all training types with their program names
	 *
	 * @return A hashmap with training types as keys and the respective program names as strings
	 */
	public Map<TrainingType, String> getAllWithProgramNames() {
		List<TrainingType> trainingTypes = repository.findAll();
		Map<TrainingType, String> result = new HashMap<>();

		for (TrainingType trainingType : trainingTypes) {
			result.put(trainingType,
					programService.getProgramsById(List.of(trainingType.getProgramId())).get(0).getName());
		}

		return result;
	}

	/**
	 * Get a list of training type details.
	 *
	 * @param  ids The ids to retrieve
	 * @return     The list of training type details
	 */
	public List<TrainingTypeDetailsDTO> getDetails(List<Long> ids) {
		List<TrainingType> trainingTypes = repository.findAllById(ids);
		Map<Long, ProgramSummaryDTO> programmes = programService
				.getProgramsById(trainingTypes.stream().map(TrainingType::getProgramId).distinct().toList())
				.stream().collect(Collectors.toMap(ProgramDetailsDTO::getId,
						p -> mapper.map(p, ProgramSummaryDTO.class)));
		return trainingTypes.stream().map(t -> {
			TrainingTypeDetailsDTO dto = mapper.map(t, TrainingTypeDetailsDTO.class);
			dto.setProgram(programmes.get(t.getProgramId()));
			return dto;
		}).toList();
	}
}
