/*
 * TAM
 * Copyright (C) 2021 - Delft University of Technology
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package nl.tudelft.tam.controller;

import static org.mockito.Mockito.*;
import static org.springframework.security.test.web.servlet.request.SecurityMockMvcRequestPostProcessors.csrf;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.*;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.*;

import java.time.LocalDate;
import java.util.Collections;
import java.util.List;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.MethodSource;
import org.modelmapper.ModelMapper;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.autoconfigure.web.servlet.AutoConfigureMockMvc;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.boot.test.mock.mockito.MockBean;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.PageImpl;
import org.springframework.data.domain.PageRequest;
import org.springframework.security.test.context.support.WithUserDetails;
import org.springframework.test.web.servlet.MockMvc;
import org.springframework.test.web.servlet.request.MockHttpServletRequestBuilder;
import org.springframework.transaction.annotation.Transactional;

import application.test.TestTAMApplication;
import nl.tudelft.labracore.api.ProgramControllerApi;
import nl.tudelft.labracore.api.dto.*;
import nl.tudelft.labracore.lib.security.user.DefaultRole;
import nl.tudelft.labracore.lib.security.user.Person;
import nl.tudelft.tam.cache.EditionCacheManager;
import nl.tudelft.tam.cache.PersonCacheManager;
import nl.tudelft.tam.dto.create.ApplicationCreateDTO;
import nl.tudelft.tam.dto.view.details.JobOfferDetailsDTO;
import nl.tudelft.tam.dto.view.summary.ApplicationSummaryDTO;
import nl.tudelft.tam.enums.Status;
import nl.tudelft.tam.model.Application;
import nl.tudelft.tam.model.Coordinator;
import nl.tudelft.tam.model.JobOffer;
import nl.tudelft.tam.repository.CoordinatorRepository;
import nl.tudelft.tam.security.AuthorisationService;
import nl.tudelft.tam.service.ApplicationService;
import nl.tudelft.tam.service.JobOfferService;
import nl.tudelft.tam.service.PersonService;
import nl.tudelft.tam.test.TestUserDetailsService;
import reactor.core.publisher.Flux;
import reactor.core.publisher.Mono;

@AutoConfigureMockMvc
@SpringBootTest(classes = TestTAMApplication.class)
@Transactional
public class ApplicationControllerTest {

	@Autowired
	MockMvc mvc;

	@MockBean
	AuthorisationService authService;

	@MockBean
	JobOfferService offerService;

	@MockBean
	ApplicationService applicationService;

	@MockBean
	PersonService personService;

	@MockBean
	EditionCacheManager editionCacheManager;

	@MockBean
	PersonCacheManager personCacheManager;

	@Autowired
	CoordinatorRepository coordinatorRepository;

	@Autowired
	ProgramControllerApi programApi;

	@Autowired
	ModelMapper mapper;

	EditionDetailsDTO edition, oldEdition;
	JobOfferDetailsDTO offer1, offer2, offer3;
	JobOffer offer1Job, offer2Job;
	ApplicationSummaryDTO app1a, app1b, app2a, app3a;

	final long EDITION_ID = 7000;
	final long OLD_EDITION_ID = 7001;
	final long JOFFER_ID_1 = 1001;
	final long JOFFER_ID_2 = 1002;
	final long JOFFER_ID_3 = 1003;
	final long PERSON_ID_1 = 5001;
	final long PERSON_ID_2 = 5002;
	final long ADMIN_ID = 329476L;

	@BeforeEach
	void setUp() {
		edition = new EditionDetailsDTO().id(EDITION_ID)
				.course(new CourseSummaryDTO().name("course-name").code("course-code"))
				.cohort(new CohortSummaryDTO().name("cohort-name")).name("edition-name")
				.isArchived(false);
		oldEdition = new EditionDetailsDTO().id(OLD_EDITION_ID)
				.course(new CourseSummaryDTO().name("course-name").code("course-code"))
				.cohort(new CohortSummaryDTO().name("cohort-name")).name("edition-name")
				.isArchived(true);

		offer1 = JobOfferDetailsDTO.builder().name("offer-1").id(JOFFER_ID_1).editionId(EDITION_ID)
				.edition(edition)
				.deadline(LocalDate.now().plusDays(1)).build();
		offer2 = JobOfferDetailsDTO.builder().name("offer-2").id(JOFFER_ID_2).editionId(EDITION_ID)
				.edition(edition)
				.deadline(LocalDate.now().plusDays(2)).build();
		offer3 = JobOfferDetailsDTO.builder().name("offer-3").id(JOFFER_ID_3).editionId(OLD_EDITION_ID)
				.edition(edition)
				.deadline(LocalDate.now().plusDays(2)).build();
		offer1Job = JobOffer.builder().name("offer-1").id(JOFFER_ID_1).editionId(EDITION_ID)
				.deadline(LocalDate.now().plusDays(1)).build();
		offer2Job = JobOffer.builder().name("offer-2").id(JOFFER_ID_2).editionId(EDITION_ID)
				.deadline(LocalDate.now().plusDays(2)).build();

		app1a = ApplicationSummaryDTO.builder().personId(PERSON_ID_1).jobOfferId(JOFFER_ID_1)
				.status(Status.SUBMITTED).build();
		app1b = ApplicationSummaryDTO.builder().personId(PERSON_ID_2).jobOfferId(JOFFER_ID_1)
				.status(Status.OFFERED).build();
		app2a = ApplicationSummaryDTO.builder().personId(PERSON_ID_1).jobOfferId(JOFFER_ID_2)
				.status(Status.REJECTED_BY_STUDENT).build();
		app3a = ApplicationSummaryDTO.builder().personId(PERSON_ID_1).jobOfferId(JOFFER_ID_3)
				.status(Status.REJECTED_BY_STUDENT).build();
	}

	@Test
	@WithUserDetails("student")
	void myApplications() throws Exception {
		mvc.perform(get("/application/my"))
				.andExpect(model().attribute("applications", Collections.emptyList()))
				.andExpect(status().isOk());
	}

	@Test
	@WithUserDetails("teacher")
	void getAllApplicationsTest() throws Exception {
		when(authService.isCoordinatorOfAny()).thenReturn(true);
		when(authService.getAuthPerson())
				.thenReturn(Person.builder().defaultRole(DefaultRole.TEACHER).build());

		Application app = mapper.map(app1a, Application.class);
		app.setId(new Application.AppId(app1a.getPersonId(), app1a.getJobOfferId()));
		app.setJobOffer(JobOffer.builder().id(app1a.getJobOfferId()).build());
		app.getJobOffer().setEditionId(EDITION_ID);
		List<Application> appList = List.of(app);

		when(applicationService.getFilteredCoordinatingApplications(anyLong(), any()))
				.thenReturn(appList);
		when(editionCacheManager.getOrThrow(eq(EDITION_ID))).thenReturn(edition);

		PersonSummaryDTO person = new PersonSummaryDTO();
		person.setDisplayName("Test");

		coordinatorRepository
				.save(Coordinator.builder().personId(TestUserDetailsService.id).programId(1L).build());
		when(programApi.getProgramById(anyLong())).thenReturn(Mono.just(new ProgramDetailsDTO().id(1L)));

		when(personCacheManager.getOrThrow(anyLong())).thenReturn(person);
		when(programApi.getAllProgramsById(anyList())).thenReturn(Flux.just(new ProgramDetailsDTO().id(1L)));
		when(personService.getPersonById(anyLong())).thenReturn(new PersonDetailsDTO().id(person.getId()).defaultRole(PersonDetailsDTO.DefaultRoleEnum.TEACHER));

		Page<Application> expectedPage = new PageImpl<>(
				appList,
				PageRequest.of(0, 25),
				0);

		mvc.perform(get("/application/all").with(csrf()))
				.andExpect(status().isOk())
				.andExpect(view().name("application/all"))
				.andExpect(model().attribute("applications", expectedPage));

		verify(applicationService).getFilteredCoordinatingApplications(anyLong(), any());
	}

	@Test
	@WithUserDetails("username")
	void addAcceptedApplicationsTest() throws Exception {
		when(authService.isCoordinatorOfAny()).thenReturn(true);
		doNothing().when(applicationService).addAcceptedApplications(anyLong(), anyLong(), anyString());

		mvc.perform(post("/application/add-accepted-apps")
						.param("offerId", Long.toString(JOFFER_ID_1))
						.param("identifiersAll", "csestudent1").with(csrf()))
				.andExpect(status().is3xxRedirection())
				.andExpect(redirectedUrl("/job-offer/" + JOFFER_ID_1));

		verify(applicationService).addAcceptedApplications(anyLong(), anyLong(), anyString());
	}

	@Test
	@WithUserDetails("username")
	void submitApplicationTest() throws Exception {
		when(authService.canSubmitApplication(anyLong())).thenReturn(true);
		when(applicationService.submit(anyLong(), any(ApplicationCreateDTO.class))).thenReturn(null);

		String page = "job-offers";

		mvc.perform(post("/application/submit")
				.param("jobOffer.id", Long.toString(JOFFER_ID_1))
				.param("page", page).with(csrf()))
				.andExpect(status().is3xxRedirection())
				.andExpect(redirectedUrl("/job-offer/open"));

		verify(applicationService).submit(anyLong(), any(ApplicationCreateDTO.class));
	}

	@Test
	@WithUserDetails("username")
	void submitApplicationTestEmptyPage() throws Exception {
		when(authService.canSubmitApplication(anyLong())).thenReturn(true);
		when(applicationService.submit(anyLong(), any(ApplicationCreateDTO.class))).thenReturn(null);

		mvc.perform(post("/application/submit")
				.param("jobOffer.id", Long.toString(JOFFER_ID_1))
				.with(csrf()))
				.andExpect(status().is3xxRedirection())
				.andExpect(redirectedUrl("/job-offer/open"));

		verify(applicationService).submit(anyLong(), any(ApplicationCreateDTO.class));
	}

	@Test
	@WithUserDetails("username")
	void retractApplicationTest() throws Exception {
		when(authService.applicationExistsFor(anyLong())).thenReturn(true);
		doNothing().when(applicationService).retract(anyLong(), anyLong());

		String page = "applications";
		String query = "test-query";

		mvc.perform(post("/application/retract/{id}", JOFFER_ID_1)
				.param("page", page)
				.param("q", query).with(csrf()))
				.andExpect(status().is3xxRedirection())
				.andExpect(redirectedUrl("/application/my?q=" + query));

		verify(authService).applicationExistsFor(JOFFER_ID_1);
		verify(applicationService).retract(anyLong(), eq(JOFFER_ID_1));
	}

	@Test
	@WithUserDetails("username")
	void retractApplicationTestEmptyPage() throws Exception {
		when(authService.applicationExistsFor(anyLong())).thenReturn(true);
		doNothing().when(applicationService).retract(anyLong(), anyLong());

		mvc.perform(post("/application/retract/{id}", JOFFER_ID_1)
				.with(csrf()))
				.andExpect(status().is3xxRedirection())
				.andExpect(redirectedUrl("/job-offer/open"));

		verify(authService).applicationExistsFor(JOFFER_ID_1);
		verify(applicationService).retract(anyLong(), eq(JOFFER_ID_1));
	}

	@Test
	@WithUserDetails("admin")
	void rejectApplicationTest() throws Exception {
		when(authService.canManageJob(anyLong())).thenReturn(true);
		doNothing().when(applicationService).reject(anyLong(), anyLong(), anyLong());

		mvc.perform(post("/application/reject/{personId}/{offerId}", PERSON_ID_1, JOFFER_ID_1).with(csrf()))
				.andExpect(status().is3xxRedirection())
				.andExpect(redirectedUrl("/job-offer/" + JOFFER_ID_1));

		verify(authService).canManageJob(JOFFER_ID_1);
		verify(applicationService).reject(PERSON_ID_1, JOFFER_ID_1, ADMIN_ID);
	}

	@Test
	@WithUserDetails("admin")
	void promoteTest() throws Exception {
		when(authService.canManageJob(anyLong())).thenReturn(true);
		doNothing().when(applicationService).promote(anyLong(), anyLong());

		mvc.perform(post("/application/promote/{personId}/{offerId}", PERSON_ID_1, JOFFER_ID_1).with(csrf()))
				.andExpect(status().is3xxRedirection())
				.andExpect(redirectedUrl("/job-offer/" + JOFFER_ID_1));

		verify(authService).canManageJob(anyLong());
		verify(applicationService).promote(PERSON_ID_1, JOFFER_ID_1);
	}

	@Test
	@WithUserDetails("admin")
	void demoteTest() throws Exception {
		when(authService.canManageJob(anyLong())).thenReturn(true);
		doNothing().when(applicationService).promote(anyLong(), anyLong());

		mvc.perform(post("/application/demote/{personId}/{offerId}", PERSON_ID_1, JOFFER_ID_1).with(csrf()))
				.andExpect(status().is3xxRedirection())
				.andExpect(redirectedUrl("/job-offer/" + JOFFER_ID_1));

		verify(authService).canManageJob(anyLong());
		verify(applicationService).demote(PERSON_ID_1, JOFFER_ID_1);
	}

	@Test
	@WithUserDetails("admin")
	void rejectAllOpenJobOfferTest() throws Exception {
		when(authService.canManageJob(anyLong())).thenReturn(true);

		doNothing().when(applicationService).rejectAllOpenInJobOffer(anyLong(), anyLong());

		mvc.perform(post("/application/reject-all-open/{id}", JOFFER_ID_1).with(csrf()))
				.andExpect(status().is3xxRedirection())
				.andExpect(redirectedUrl("/job-offer/" + JOFFER_ID_1));

		verify(authService).canManageJob(JOFFER_ID_1);
		verify(applicationService).rejectAllOpenInJobOffer(JOFFER_ID_1, ADMIN_ID);
	}

	@Test
	@WithUserDetails("admin")
	void offerApplicationTest() throws Exception {
		when(authService.canManageJob(anyLong())).thenReturn(true);
		doNothing().when(applicationService).offer(anyLong(), anyLong(), anyLong());

		mvc.perform(post("/application/offer/{personId}/{offerId}", PERSON_ID_1, JOFFER_ID_1).with(csrf()))
				.andExpect(status().is3xxRedirection())
				.andExpect(redirectedUrl("/job-offer/" + JOFFER_ID_1));

		verify(authService).canManageJob(anyLong());
		verify(applicationService).offer(PERSON_ID_1, JOFFER_ID_1, ADMIN_ID);
	}

	@Test
	@WithUserDetails("admin")
	void bulkOfferApplicationTest() throws Exception {
		PersonSummaryDTO p1 = new PersonSummaryDTO().id(PERSON_ID_1);
		PersonSummaryDTO p2 = new PersonSummaryDTO().id(PERSON_ID_2);
		when(authService.canManageJob(JOFFER_ID_1)).thenReturn(true);
		when(personService.searchForPeople(anyString())).thenReturn(List.of(p1, p2));
		doNothing().when(applicationService).bulkOffer(anyLong(), anyString(), anyLong());

		String body = "test";
		String id = "test";
		mvc.perform(post("/application/offer/{offerId}", JOFFER_ID_1)
				.param("identifiersAll", body).with(csrf()))
				.andExpect(status().is3xxRedirection())
				.andExpect(redirectedUrl("/job-offer/" + JOFFER_ID_1));

		verify(authService).canManageJob(JOFFER_ID_1);
		verify(applicationService).bulkOffer(JOFFER_ID_1, id, ADMIN_ID);
	}

	@Test
	@WithUserDetails("admin")
	void bulkRejectApplicationTest() throws Exception {
		PersonSummaryDTO p1 = new PersonSummaryDTO().id(PERSON_ID_1);
		PersonSummaryDTO p2 = new PersonSummaryDTO().id(PERSON_ID_2);
		when(authService.canManageJob(JOFFER_ID_1)).thenReturn(true);
		when(personService.searchForPeople(anyString())).thenReturn(List.of(p1, p2));
		doNothing().when(applicationService).bulkReject(anyLong(), anyString(), anyLong());

		String body = "test";
		String id = "test";
		mvc.perform(post("/application/reject/{offerId}", JOFFER_ID_1)
				.param("identifiersAll", body).with(csrf()))
				.andExpect(status().is3xxRedirection())
				.andExpect(redirectedUrl("/job-offer/" + JOFFER_ID_1));

		verify(authService).canManageJob(JOFFER_ID_1);
		verify(applicationService).bulkReject(JOFFER_ID_1, id, ADMIN_ID);
	}

	@Test
	@WithUserDetails("username")
	void acceptApplicationTest() throws Exception {
		when(authService.offerExistsFor(anyLong())).thenReturn(true);
		doNothing().when(applicationService).accept(anyLong(), anyLong());

		mvc.perform(post("/application/accept/{offerId}", JOFFER_ID_1).with(csrf()))
				.andExpect(status().is3xxRedirection())
				.andExpect(redirectedUrl("/job-offer/open"));

		verify(authService).offerExistsFor(JOFFER_ID_1);
		verify(applicationService).accept(anyLong(), eq(JOFFER_ID_1));

		mvc.perform(post("/application/accept/{offerId}", JOFFER_ID_1)
				.queryParam("q", "test")
				.with(csrf()))
				.andExpect(status().is3xxRedirection())
				.andExpect(redirectedUrl("/job-offer/open?q=test"));
	}

	@ParameterizedTest
	@MethodSource(value = "protectedEndpoints")
	void testWithoutUserDetailsIsForbidden(MockHttpServletRequestBuilder request) throws Exception {
		mvc.perform(request.with(csrf()))
				.andExpect(status().is3xxRedirection())
				.andExpect(redirectedUrl("http://localhost/login"));
	}

	private static List<MockHttpServletRequestBuilder> protectedEndpoints() {
		return List.of(
				get("/application/my"),
				post("/application/submit/123"),
				post("/application/retract/123"),
				post("/application/reject/123/456"),
				post("/application/offer/123/456"),
				post("/application/accept/123"),
				post("/application/new-offer/123"));
	}

}
