/*
 * TAM
 * Copyright (C) 2021 - Delft University of Technology
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package nl.tudelft.tam.controller;

import static org.mockito.Mockito.*;
import static org.springframework.security.test.web.servlet.request.SecurityMockMvcRequestPostProcessors.csrf;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.*;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.*;

import java.time.LocalDate;
import java.util.List;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.CsvSource;
import org.junit.jupiter.params.provider.MethodSource;
import org.modelmapper.ModelMapper;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.autoconfigure.web.servlet.AutoConfigureMockMvc;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.boot.test.mock.mockito.MockBean;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.PageImpl;
import org.springframework.data.domain.PageRequest;
import org.springframework.http.MediaType;
import org.springframework.security.test.context.support.WithUserDetails;
import org.springframework.test.web.servlet.MockMvc;
import org.springframework.test.web.servlet.request.MockHttpServletRequestBuilder;
import org.springframework.transaction.annotation.Transactional;

import application.test.TestTAMApplication;
import nl.tudelft.labracore.api.dto.*;
import nl.tudelft.labracore.lib.security.user.DefaultRole;
import nl.tudelft.labracore.lib.security.user.Person;
import nl.tudelft.tam.cache.EditionCacheManager;
import nl.tudelft.tam.cache.PersonCacheManager;
import nl.tudelft.tam.dto.create.DeclarationCreateDTO;
import nl.tudelft.tam.dto.view.details.ExtraWorkDetailsDTO;
import nl.tudelft.tam.dto.view.summary.DeclarationSummaryDTO;
import nl.tudelft.tam.enums.Status;
import nl.tudelft.tam.model.Declaration;
import nl.tudelft.tam.model.ExtraWork;
import nl.tudelft.tam.model.Profile;
import nl.tudelft.tam.security.AuthorisationService;
import nl.tudelft.tam.service.DeclarationService;
import nl.tudelft.tam.service.ProgramService;

@AutoConfigureMockMvc
@SpringBootTest(classes = TestTAMApplication.class)
@Transactional
class DeclarationControllerTest {

	@Autowired
	MockMvc mvc;

	@MockBean
	AuthorisationService authService;

	@MockBean
	ProgramService programService;

	@MockBean
	DeclarationService declarationService;

	@MockBean
	PersonCacheManager personCacheManager;

	@MockBean
	EditionCacheManager editionCacheManager;

	@Autowired
	ModelMapper mapper;

	EditionDetailsDTO edition, oldEdition;
	ExtraWorkDetailsDTO work1, work2, work3;
	DeclarationSummaryDTO dec1a, dec1b, dec2a, dec3a;

	final long EDITION_ID = 7000;
	final long OLD_EDITION_ID = 7001;
	final long WORK_ID_1 = 1001;
	final long WORK_ID_2 = 1002;
	final long WORK_ID_3 = 1003;
	final long PERSON_ID_1 = 5001;
	final long PERSON_ID_2 = 5002;
	final long ADMIN_ID = 329476L;

	@BeforeEach
	void setUp() {
		edition = new EditionDetailsDTO().id(EDITION_ID)
				.course(new CourseSummaryDTO().name("course-name").code("course-code"))
				.cohort(new CohortSummaryDTO().name("cohort-name")).name("edition-name")
				.isArchived(false);
		oldEdition = new EditionDetailsDTO().id(OLD_EDITION_ID)
				.course(new CourseSummaryDTO().name("course-name").code("course-code"))
				.cohort(new CohortSummaryDTO().name("cohort-name")).name("edition-name")
				.isArchived(true);

		work1 = ExtraWorkDetailsDTO.builder().name("exwork-1").id(WORK_ID_1).editionId(EDITION_ID)
				.edition(edition)
				.deadline(LocalDate.now().plusDays(1)).build();
		work2 = ExtraWorkDetailsDTO.builder().name("exwork-2").id(WORK_ID_2).editionId(EDITION_ID)
				.edition(edition)
				.deadline(LocalDate.now().plusDays(2)).build();
		work3 = ExtraWorkDetailsDTO.builder().name("exwork-3").id(WORK_ID_3).editionId(OLD_EDITION_ID)
				.edition(oldEdition)
				.deadline(LocalDate.now().plusDays(2)).build();

		dec1a = DeclarationSummaryDTO.builder().personId(PERSON_ID_1).workId(WORK_ID_1).amount(10)
				.status(Status.SUBMITTED).build();
		dec1b = DeclarationSummaryDTO.builder().personId(PERSON_ID_2).workId(WORK_ID_1).amount(20)
				.status(Status.ACCEPTED).build();
		dec2a = DeclarationSummaryDTO.builder().personId(PERSON_ID_1).workId(WORK_ID_2).amount(30)
				.status(Status.REJECTED_BY_STUDENT).build();
		dec3a = DeclarationSummaryDTO.builder().personId(PERSON_ID_1).workId(WORK_ID_3).amount(40)
				.status(Status.REJECTED_BY_STUDENT).build();
	}

	@Test
	@WithUserDetails("teacher")
	void getAllDeclarationsTest() throws Exception {
		when(authService.getAuthPerson()).thenReturn(Person.builder().defaultRole(DefaultRole.TEACHER).build());
		when(authService.isCoordinatorOfAny()).thenReturn(true);

		List<Declaration> decListStart = List.of(mapper.map(dec1a, Declaration.class),
				mapper.map(dec3a, Declaration.class));
		decListStart.get(0).setExtraWork(ExtraWork.builder().id(WORK_ID_2).editionId(EDITION_ID).build());
		decListStart.get(1).setExtraWork(ExtraWork.builder().id(WORK_ID_3).editionId(OLD_EDITION_ID).build());
		decListStart.forEach(d -> d.setPerson(Profile.builder().id(PERSON_ID_1).build()));
		List<Declaration> decListEnd = List.of(decListStart.get(0));

		when(declarationService.getCoordinatingDeclarationsForPrograms(anyLong(), anyList())).thenReturn(decListStart);
		when(editionCacheManager.getOrThrow(eq(EDITION_ID))).thenReturn(edition);
		when(editionCacheManager.getOrThrow(eq(OLD_EDITION_ID))).thenReturn(oldEdition);
		when(programService.getCoordinatingPrograms(anyLong())).thenReturn(List.of(new ProgramDetailsDTO().id(1L)));

		PersonSummaryDTO person = new PersonSummaryDTO();
		person.setDisplayName("Test");

		when(personCacheManager.getOrThrow(anyLong())).thenReturn(person);

		Page<Declaration> expectedPage = new PageImpl<>(
				decListEnd,
				PageRequest.of(0, 25),
				0);

		mvc.perform(get("/declaration/all").with(csrf()))
				.andExpect(status().isOk())
				.andExpect(view().name("declaration/all"))
				.andExpect(model().attribute("declarations", expectedPage));

		verify(declarationService).getCoordinatingDeclarationsForPrograms(anyLong(), anyList());
	}

	@Test
	@WithUserDetails("admin")
	void getDeclarationsTest() throws Exception {
		mvc.perform(get("/declaration/my"))
				.andExpect(status().is3xxRedirection())
				.andExpect(redirectedUrl("/extra-work/open?tab=declarations"));
	}

	@Test
	@WithUserDetails("admin")
	void getDeclarationsTestWithQ() throws Exception {
		mvc.perform(get("/declaration/my?q=abc123DEF"))
				.andExpect(status().is3xxRedirection())
				.andExpect(redirectedUrl("/extra-work/open?tab=declarations&q=abc123DEF"));
	}

	@Test
	@WithUserDetails("admin")
	void submitDeclarationTest() throws Exception {
		doNothing().when(declarationService).submit(any());
		when(authService.canDeclareHours(anyLong())).thenReturn(true);

		DeclarationCreateDTO createDto = DeclarationCreateDTO.builder()
				.personId(PERSON_ID_1)
				.workId(WORK_ID_1)
				.amount(404)
				.status(Status.SUBMITTED)
				.build();

		mvc.perform(post("/declaration/submit?workId=1").with(csrf())
				.contentType(MediaType.APPLICATION_FORM_URLENCODED)
				.flashAttr("createDto", createDto))
				.andExpect(status().is3xxRedirection())
				.andExpect(redirectedUrl("/declaration/my"));

		verify(declarationService).submit(any(DeclarationCreateDTO.class));
	}

	@Test
	@WithUserDetails("admin")
	void submitDeclarationTestWithPage() throws Exception {
		doNothing().when(declarationService).submit(any());
		when(authService.canDeclareHours(anyLong())).thenReturn(true);

		DeclarationCreateDTO createDto = DeclarationCreateDTO.builder()
				.personId(PERSON_ID_1)
				.workId(WORK_ID_1)
				.amount(404)
				.status(Status.SUBMITTED)
				.build();

		mvc.perform(post("/declaration/submit?page=abc&workId=1").with(csrf())
				.contentType(MediaType.APPLICATION_FORM_URLENCODED)
				.flashAttr("createDto", createDto))
				.andExpect(status().is3xxRedirection())
				.andExpect(redirectedUrl("/abc"));

		verify(declarationService).submit(any(DeclarationCreateDTO.class));
	}

	@ParameterizedTest
	@CsvSource({
			"      ,    , /declaration/my",
			"myPage,    , /myPage",
			"      , myQ, /declaration/my?q=myQ",
			"myPage, myQ, /myPage?q=myQ",
	})
	@WithUserDetails("admin")
	void retractDeclarationTest(String page, String q, String res) throws Exception {
		long id = 687754L;

		when(authService.ownsDeclaration(anyLong())).thenReturn(true);
		doNothing().when(declarationService).retract(anyLong());

		mvc.perform(post("/declaration/retract/{id}", id).param("page", page).param("q", q)
				.with(csrf()))
				.andExpect(status().is3xxRedirection())
				.andExpect(redirectedUrl(res));

		verify(authService).ownsDeclaration(id);
		verify(declarationService).retract(id);
	}

	@Test
	@WithUserDetails("admin")
	void rejectDeclarationTest() throws Exception {
		long decId = 3156845L;
		long exwId = 1354681L;

		when(authService.canManageDeclaration(anyLong())).thenReturn(true);

		doNothing().when(declarationService).reject(anyLong(), anyLong());
		when(declarationService.getExtraWorkIdFromDeclaration(anyLong())).thenReturn(exwId);

		mvc.perform(post("/declaration/reject/{id}", decId).with(csrf()))
				.andExpect(status().is3xxRedirection())
				.andExpect(redirectedUrl("/extra-work/" + exwId));

		verify(authService).canManageDeclaration(decId);
		verify(declarationService).reject(decId, ADMIN_ID);
		verify(declarationService).getExtraWorkIdFromDeclaration(decId);
	}

	@Test
	@WithUserDetails("admin")
	void rejectDeclarationTestWithQ() throws Exception {
		long decId = 3156845L;
		long exwId = 1354681L;
		String q = "abc123DEF";

		when(authService.canManageDeclaration(anyLong())).thenReturn(true);

		doNothing().when(declarationService).reject(anyLong(), anyLong());
		when(declarationService.getExtraWorkIdFromDeclaration(anyLong())).thenReturn(exwId);

		mvc.perform(post("/declaration/reject/{id}", decId).param("q", q).with(csrf()))
				.andExpect(status().is3xxRedirection())
				.andExpect(redirectedUrl("/extra-work/" + exwId + "?q=" + q));

		verify(authService).canManageDeclaration(decId);
		verify(declarationService).reject(decId, ADMIN_ID);
		verify(declarationService).getExtraWorkIdFromDeclaration(decId);
	}

	@Test
	@WithUserDetails("admin")
	void rejectAllOpenInExtraWorkTest() throws Exception {
		when(authService.canManageWork(anyLong())).thenReturn(true);

		doNothing().when(declarationService).rejectAllOpenInExtraWork(anyLong(), anyLong());

		mvc.perform(post("/declaration/reject-all-open/{id}", WORK_ID_1).with(csrf()))
				.andExpect(status().is3xxRedirection())
				.andExpect(redirectedUrl("/extra-work/" + WORK_ID_1));

		verify(authService).canManageWork(WORK_ID_1);
		verify(declarationService).rejectAllOpenInExtraWork(WORK_ID_1, ADMIN_ID);
	}

	@Test
	@WithUserDetails("admin")
	void approveDeclarationTest() throws Exception {
		long decId = 3156845L;
		long exwId = 1354681L;
		String q = "abc123DEF";

		when(authService.canManageDeclaration(anyLong())).thenReturn(true);

		doNothing().when(declarationService).approve(anyLong(), anyLong());
		when(declarationService.getExtraWorkIdFromDeclaration(anyLong())).thenReturn(exwId);

		mvc.perform(post("/declaration/approve/{id}", decId).param("q", q).with(csrf()))
				.andExpect(status().is3xxRedirection())
				.andExpect(redirectedUrl("/extra-work/" + exwId + "?q=" + q));

		verify(authService).canManageDeclaration(decId);
		verify(declarationService).approve(decId, ADMIN_ID);
		verify(declarationService).getExtraWorkIdFromDeclaration(decId);
	}

	@Test
	@WithUserDetails("admin")
	void approveDeclarationTestWithQ() throws Exception {
		long decId = 3156845L;
		long exwId = 1354681L;

		when(authService.canManageDeclaration(anyLong())).thenReturn(true);

		doNothing().when(declarationService).approve(anyLong(), anyLong());
		when(declarationService.getExtraWorkIdFromDeclaration(anyLong())).thenReturn(exwId);

		mvc.perform(post("/declaration/approve/{id}", decId).with(csrf()))
				.andExpect(status().is3xxRedirection())
				.andExpect(redirectedUrl("/extra-work/" + exwId));

		verify(authService).canManageDeclaration(decId);
		verify(declarationService).approve(decId, ADMIN_ID);
		verify(declarationService).getExtraWorkIdFromDeclaration(decId);
	}

	@ParameterizedTest
	@MethodSource(value = "protectedEndpoints")
	void testWithoutUserDetailsIsForbidden(MockHttpServletRequestBuilder request) throws Exception {
		mvc.perform(request.with(csrf()))
				.andExpect(status().is3xxRedirection())
				.andExpect(redirectedUrl("http://localhost/login"));
	}

	private static List<MockHttpServletRequestBuilder> protectedEndpoints() {
		return List.of(
				get("/declaration/my"),
				post("/declaration/submit"),
				post("/declaration/retract/123"),
				post("/declaration/reject/123/"),
				post("/declaration/approve/123"));
	}
}
