/*
 * TAM
 * Copyright (C) 2021 - Delft University of Technology
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package nl.tudelft.tam.controller;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.mockito.Mockito.*;

import java.time.LocalDate;
import java.time.LocalDateTime;
import java.util.ArrayList;
import java.util.List;
import java.util.Objects;

import org.junit.jupiter.api.Test;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.ui.ExtendedModelMap;
import org.springframework.ui.Model;

import nl.tudelft.labracore.api.EditionControllerApi;
import nl.tudelft.labracore.api.RoleControllerApi;
import nl.tudelft.labracore.api.dto.*;
import nl.tudelft.labracore.lib.security.user.Person;
import nl.tudelft.tam.dto.create.TamEditionCreateDTO;
import nl.tudelft.tam.dto.patch.AddMemberDTO;
import nl.tudelft.tam.dto.view.details.EditionDetailsOffersWorkDTO;
import nl.tudelft.tam.service.*;
import reactor.core.publisher.Mono;

@Transactional
public class EditionControllerTest extends ControllerTest {

	private final EditionController editionController;

	private final EditionService editionService;
	private final RoleService roleService;
	private final PersonService personService;
	private final JobOfferService jobOfferService;
	private final RoleControllerApi roleApi;
	private final EditionControllerApi editionApi;
	private final CoordinatorDefaultService coordinatorDefaultService;
	private final ApplicationService applicationService;

	public EditionControllerTest() {
		this.editionService = mock(EditionService.class);
		this.roleService = mock(RoleService.class);
		this.personService = mock(PersonService.class);
		this.jobOfferService = mock(JobOfferService.class);
		this.roleApi = mock(RoleControllerApi.class);
		this.editionApi = mock(EditionControllerApi.class);
		this.coordinatorDefaultService = mock(CoordinatorDefaultService.class);
		this.applicationService = mock(ApplicationService.class);

		this.editionController = new EditionController(editionService, roleService, personService,
				jobOfferService, roleApi, editionApi, coordinatorDefaultService, applicationService);
	}

	@Test
	void createEditionTest() {
		Long personId = randomId();
		Long editionId = randomId();
		Long courseId = randomId();
		Long cohortId = randomId();

		when(editionService.addEdition(any())).thenReturn(editionId);

		editionController.createEdition(Person.builder().id(personId).build(),
				TamEditionCreateDTO.builder()
						.name("Test Edition")
						.course(new CourseIdDTO().id(courseId))
						.cohort(new CohortIdDTO().id(cohortId))
						.startDate(LocalDate.of(2023, 1, 1))
						.endDate(LocalDate.of(2024, 1, 1))
						.build());

		verify(editionService).addEdition(new EditionCreateDTO()
				.name("Test Edition")
				.course(new CourseIdDTO().id(courseId))
				.cohort(new CohortIdDTO().id(cohortId))
				.enrollability(EditionCreateDTO.EnrollabilityEnum.OPEN)
				.startDate(LocalDateTime.of(2023, 1, 1, 0, 0))
				.endDate(LocalDateTime.of(2024, 1, 1, 23, 59)));
		verify(roleService).changeToTeacher(personId, editionId);
	}

	@Test
	void viewEditionTest() {
		Long editionId = 1L;
		EditionDetailsOffersWorkDTO editionDetails = new EditionDetailsOffersWorkDTO();
		RolePersonDetailsDTO rolePersonDetails = new RolePersonDetailsDTO();
		rolePersonDetails.setPerson(new PersonSummaryDTO().username("testUser"));

		when(editionService.getEditionsWithJobsAndWork(List.of(editionId)))
				.thenReturn(List.of(editionDetails));
		when(jobOfferService.getTAsWithoutAJobOfferInEdition(editionId))
				.thenReturn(List.of(rolePersonDetails));
		when(editionService.getEditionRoles(editionId))
				.thenReturn(new ArrayList<>());
		doNothing().when(coordinatorDefaultService).addDefaultsToEditionsWithJobsAndWork(any());

		Model model = new ExtendedModelMap();
		String viewName = editionController.viewEdition(editionId, model);

		assertEquals("edition/view", viewName);
		assertEquals(editionDetails, model.getAttribute("edition"));
		assertEquals(new ArrayList<>(), model.getAttribute("roles"));
		assertEquals(List.of(rolePersonDetails), model.getAttribute("TAsWithoutJobOffer"));
		assertEquals("testUser", model.getAttribute("TAsWithoutJobOfferIdentifiers"));
	}

	@Test
	void addMemberTest() {
		Long editionId = 1L;
		Long personId = 2L;
		AddMemberDTO user = new AddMemberDTO();
		user.setUsername("testUser");
		user.setRole(RoleCreateDTO.TypeEnum.TA);

		when(personService.getPersonByUsername("testUser")).thenReturn(new PersonDetailsDTO().id(personId));
		when(roleApi.addRole(any(RoleCreateDTO.class))).thenReturn(Mono.empty());

		String redirectUrl = editionController.addMember(editionId, user);

		assertEquals("redirect:/edition/{id}", redirectUrl);
		verify(personService).getPersonByUsername("testUser");
		verify(roleApi).addRole(
				argThat(roleCreateDTO -> Objects.equals(roleCreateDTO.getEdition().getId(), editionId) &&
						Objects.equals(roleCreateDTO.getPerson().getId(), personId) &&
						roleCreateDTO.getType().equals(RoleCreateDTO.TypeEnum.TA)));
	}

	@Test
	void removeMemberTest() {
		Long editionId = 1L;
		Long personId = 2L;
		String username = "testUser2";

		doNothing().when(applicationService).rejectPersonFromAllOffersInEdition(anyLong(), anyLong(),
				anyLong());
		when(personService.getPersonByUsername(username)).thenReturn(new PersonDetailsDTO().id(personId));

		String redirectUrl = editionController.removeMember(Person.builder().id(4L).build(), username,
				editionId);

		assertEquals("redirect:/edition/{id}", redirectUrl);
		verify(personService).getPersonByUsername(username);
		verify(applicationService).rejectPersonFromAllOffersInEdition(editionId, personId, 4L);
	}

}
