/*
 * TAM
 * Copyright (C) 2021 - Delft University of Technology
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package nl.tudelft.tam.controller;

import static org.mockito.Mockito.*;
import static org.springframework.security.test.web.servlet.request.SecurityMockMvcRequestPostProcessors.csrf;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.*;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.*;

import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.time.LocalDate;
import java.util.ArrayList;
import java.util.List;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.MethodSource;
import org.modelmapper.ModelMapper;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.boot.test.autoconfigure.web.servlet.AutoConfigureMockMvc;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.boot.test.mock.mockito.MockBean;
import org.springframework.http.MediaType;
import org.springframework.security.test.context.support.WithUserDetails;
import org.springframework.test.web.servlet.MockMvc;
import org.springframework.test.web.servlet.request.MockHttpServletRequestBuilder;
import org.springframework.transaction.annotation.Transactional;

import application.test.TestTAMApplication;
import nl.tudelft.labracore.api.dto.*;
import nl.tudelft.tam.dto.create.ExtraWorkCreateDTO;
import nl.tudelft.tam.dto.patch.ExtraWorkPatchDTO;
import nl.tudelft.tam.dto.view.details.*;
import nl.tudelft.tam.dto.view.summary.DeclarationSummaryDTO;
import nl.tudelft.tam.dto.view.summary.ExtraWorkSummaryDTO;
import nl.tudelft.tam.enums.Status;
import nl.tudelft.tam.model.Application;
import nl.tudelft.tam.model.JobOffer;
import nl.tudelft.tam.security.AuthorisationService;
import nl.tudelft.tam.service.*;

@AutoConfigureMockMvc
@SpringBootTest(classes = TestTAMApplication.class)
@Transactional
class ExtraWorkControllerTest {

	@Value("${tam.filesys.exports-dir}")
	String exportDir;

	@Autowired
	MockMvc mvc;

	@MockBean
	AuthorisationService authService;

	@MockBean
	ExtraWorkService workService;

	@MockBean
	DeclarationService declarationService;

	@MockBean
	EditionService editionService;

	@MockBean
	ApplicationService applicationService;

	@MockBean
	PersonService personService;

	@Autowired
	ModelMapper mapper;

	EditionDetailsDTO edition;
	ExtraWorkDetailsDTO work1, work2;
	DeclarationSummaryDTO dec1a, dec1b, dec2a;
	Application app1a;
	JobOffer jobOffer1a;
	RoleEditionDetailsDTO roleEditionDTO1a;

	final long EDITION_ID = 7000;
	final long WORK_ID_1 = 1001;
	final long WORK_ID_2 = 1002;
	final long PERSON_ID_1 = 5001;
	final long PERSON_ID_2 = 5002;
	final long ADMIN_ID = 329476L;

	@BeforeEach
	void setUp() {
		edition = new EditionDetailsDTO().id(EDITION_ID)
				.course(new CourseSummaryDTO().name("course-name").code("course-code"))
				.cohort(new CohortSummaryDTO().name("cohort-name")).name("edition-name");

		work1 = ExtraWorkDetailsDTO.builder().name("exwork-1").id(WORK_ID_1).editionId(EDITION_ID)
				.edition(edition)
				.deadline(LocalDate.now().plusDays(1)).build();
		work2 = ExtraWorkDetailsDTO.builder().name("exwork-2").id(WORK_ID_2).editionId(EDITION_ID)
				.edition(edition)
				.deadline(LocalDate.now().plusDays(2)).build();

		dec1a = DeclarationSummaryDTO.builder().personId(PERSON_ID_1).workId(WORK_ID_1).amount(10)
				.status(Status.SUBMITTED).build();
		dec1b = DeclarationSummaryDTO.builder().personId(PERSON_ID_2).workId(WORK_ID_1).amount(20)
				.status(Status.ACCEPTED).build();
		dec2a = DeclarationSummaryDTO.builder().personId(PERSON_ID_1).workId(WORK_ID_2).amount(30)
				.status(Status.REJECTED_BY_STUDENT).build();

		jobOffer1a = JobOffer.builder().editionId(EDITION_ID).build();

		app1a = Application.builder().status(Status.ACCEPTED).jobOffer(jobOffer1a).build();

		EditionSummaryDTO eSummary = mapper.map(edition, EditionSummaryDTO.class);

		roleEditionDTO1a = new RoleEditionDetailsDTO().type(RoleEditionDetailsDTO.TypeEnum.TA)
				.edition(eSummary);
	}

	@Test
	@WithUserDetails("username")
	void getOpenTest() throws Exception {
		DeclarationDetailsExtraWorkDTO dec1 = mapper.map(dec1a, DeclarationDetailsExtraWorkDTO.class);
		dec1.setExtraWork(work1);

		when(declarationService.getFilteredDeclarationsForPerson(anyLong(), any()))
				.thenReturn(List.of(dec1));
		when(personService.getRolesForPerson(anyLong()))
				.thenReturn(List.of(roleEditionDTO1a));

		ExtraWorkDetailsDTO work1Detail = mapper.map(work1, ExtraWorkDetailsDTO.class);
		ExtraWorkDetailsDTO work2Detail = mapper.map(work2, ExtraWorkDetailsDTO.class);

		List<ExtraWorkDetailsDTO> workDetailed = List.of(work1Detail, work2Detail);

		when(workService.getFilteredOpen(eq(null))).thenReturn(workDetailed);

		List<ExtraWorkSummaryDTO> hiredWork = List.of(work1, work2);

		List<EditionDetailsOffersWorkDTO> editions = List
				.of(mapper.map(edition, EditionDetailsOffersWorkDTO.class));
		editions.get(0).setOffers(new ArrayList<>());
		editions.get(0).setWork(hiredWork);

		when(editionService.combineOffersAndWorkWithEditions(eq(new ArrayList<>()), eq(hiredWork)))
				.thenReturn(editions);

		mvc.perform(get("/extra-work/open").with(csrf()))
				.andExpect(status().isOk())
				.andExpect(view().name("extra_work/view_many"))
				.andExpect(model().attribute("editions", editions))
				.andExpect(model().attribute("declarations", List.of(dec1)));

		verify(declarationService).getFilteredDeclarationsForPerson(anyLong(), eq(null));
		verify(personService).getRolesForPerson(anyLong());
		verify(workService).getFilteredOpen(eq(null));
		verify(editionService).combineOffersAndWorkWithEditions(eq(new ArrayList<>()), eq(hiredWork));
	}

	@Test
	@WithUserDetails("username")
	void getOpenWithQueryTest() throws Exception {
		String query = "test-query";

		DeclarationDetailsExtraWorkDTO dec1 = mapper.map(dec1a, DeclarationDetailsExtraWorkDTO.class);
		dec1.setExtraWork(work1);

		when(declarationService.getFilteredDeclarationsForPerson(anyLong(), eq(query)))
				.thenReturn(List.of(dec1));
		when(personService.getRolesForPerson(anyLong()))
				.thenReturn(List.of(roleEditionDTO1a));

		ExtraWorkDetailsDTO work1Detail = mapper.map(work1, ExtraWorkDetailsDTO.class);
		ExtraWorkDetailsDTO work2Detail = mapper.map(work2, ExtraWorkDetailsDTO.class);

		List<ExtraWorkDetailsDTO> workDetailed = List.of(work1Detail, work2Detail);

		when(workService.getFilteredOpen(eq(query))).thenReturn(workDetailed);

		List<ExtraWorkSummaryDTO> hiredWork = List.of(work1, work2);

		List<EditionDetailsOffersWorkDTO> editions = List
				.of(mapper.map(edition, EditionDetailsOffersWorkDTO.class));
		editions.get(0).setOffers(new ArrayList<>());
		editions.get(0).setWork(hiredWork);

		when(editionService.combineOffersAndWorkWithEditions(eq(new ArrayList<>()), eq(hiredWork)))
				.thenReturn(editions);

		mvc.perform(get("/extra-work/open")
				.queryParam("q", query)
				.queryParam("tab", "declarations")
				.with(csrf()))
				.andExpect(status().isOk())
				.andExpect(view().name("extra_work/view_many"))
				.andExpect(model().attribute("editions", editions))
				.andExpect(model().attribute("declarations", List.of(dec1)));

		verify(declarationService).getFilteredDeclarationsForPerson(anyLong(), eq(query));
		verify(personService).getRolesForPerson(anyLong());
		verify(workService).getFilteredOpen(eq(query));
		verify(editionService).combineOffersAndWorkWithEditions(eq(new ArrayList<>()), eq(hiredWork));
	}

	@Test
	@WithUserDetails("admin")
	void getExtraWorkDetailsTest() throws Exception {
		when(authService.canManageWork(anyLong())).thenReturn(true);

		DeclarationDetailsPersonDTO ap1 = mapper.map(dec1a, DeclarationDetailsPersonDTO.class);
		ap1.setPerson(new PersonSummaryDTO().id(PERSON_ID_1).username("user1").displayName("name1")
				.externalId("exid1").number(4201).email("1@poo"));
		DeclarationDetailsPersonDTO ap2 = mapper.map(dec1b, DeclarationDetailsPersonDTO.class);
		ap2.setPerson(new PersonSummaryDTO().id(PERSON_ID_2).username("user2").displayName("name2")
				.externalId("exid2").number(4202).email("2@poo"));

		when(workService.getById(anyLong())).thenReturn(work1);
		when(declarationService.getFilteredDeclarationsForExtraWork(anyLong(), any()))
				.thenReturn(List.of(ap1, ap2));

		int[] stats = { 2, 30, 20 };
		when(declarationService.getStatsForDecs(anyList())).thenReturn(stats);

		mvc.perform(get("/extra-work/{workId}", WORK_ID_1))
				.andExpect(status().isOk())
				.andExpect(view().name("extra_work/view_one"))
				.andExpect(model().attribute("work", work1))
				.andExpect(model().attribute("deadlineClosed", false))
				.andExpect(model().attribute("declarations", List.of(ap1, ap2)))
				.andExpect(model().attribute("stats", stats))
				.andExpect(model().attributeExists("patchExtraWork"));

		verify(workService).getById(WORK_ID_1);
		verify(declarationService).getFilteredDeclarationsForExtraWork(eq(WORK_ID_1), any());
		verify(declarationService).getStatsForDecs(anyList());
	}

	@Test
	@WithUserDetails("admin")
	void createExtraWorkTest() throws Exception {
		when(authService.canManageEdition(anyLong())).thenReturn(true);

		ExtraWorkCreateDTO createDto = ExtraWorkCreateDTO.builder().name("test-name").editionId(EDITION_ID)
				.baanCode("IEP5000").contractName("contract")
				.deadline(LocalDate.now().plusDays(42)).maxHours(100).build();
		long newId = 50099;

		when(workService.newExtraWork(any())).thenReturn(newId);

		mvc.perform(post("/extra-work").with(csrf())
				.contentType(MediaType.APPLICATION_FORM_URLENCODED)
				.flashAttr("extrawork", createDto))
				.andExpect(status().is3xxRedirection())
				.andExpect(redirectedUrl("/extra-work/" + newId));

		verify(authService).canManageEdition(EDITION_ID);
		verify(workService).newExtraWork(createDto);
	}

	@Test
	@WithUserDetails("admin")
	void patchExtraWorkTest() throws Exception {
		when(authService.canManageWork(anyLong())).thenReturn(true);

		ExtraWorkPatchDTO patchDto = ExtraWorkPatchDTO.builder().name("test-name")
				.deadline(LocalDate.now().plusDays(42)).maxHours(100).build();

		doNothing().when(workService).patchExtraWork(anyLong(), any(ExtraWorkPatchDTO.class));

		mvc.perform(patch("/extra-work/{id}", WORK_ID_1).with(csrf())
				.contentType(MediaType.APPLICATION_FORM_URLENCODED)
				.flashAttr("patchExtraWork", patchDto))
				.andExpect(status().is3xxRedirection())
				.andExpect(redirectedUrl("/extra-work/" + WORK_ID_1));

		verify(authService).canManageWork(WORK_ID_1);
		verify(workService).patchExtraWork(WORK_ID_1, patchDto);

	}

	@Test
	@WithUserDetails("admin")
	void closeExtraWorkTest() throws Exception {
		when(authService.canManageWork(anyLong())).thenReturn(true);

		doNothing().when(workService).closeExtraWork(anyLong());

		mvc.perform(post("/extra-work/close/{id}", WORK_ID_1).with(csrf()))
				.andExpect(status().is3xxRedirection())
				.andExpect(redirectedUrl("/extra-work/" + WORK_ID_1));

		verify(authService).canManageWork(WORK_ID_1);
		verify(workService).closeExtraWork(WORK_ID_1);
	}

	@Test
	@WithUserDetails("admin")
	void rejectAllOpenInExtraWorkTest() throws Exception {
		when(authService.canManageWork(anyLong())).thenReturn(true);

		doNothing().when(declarationService).rejectAllOpenInExtraWork(anyLong(), anyLong());

		mvc.perform(post("/extra-work/reject-all-open/{id}", WORK_ID_1).with(csrf()))
				.andExpect(status().is3xxRedirection())
				.andExpect(redirectedUrl("/extra-work/" + WORK_ID_1));

		verify(authService).canManageWork(WORK_ID_1);
		verify(declarationService).rejectAllOpenInExtraWork(WORK_ID_1, ADMIN_ID);
	}

	@Test
	@WithUserDetails("admin")
	void approveAllOpenInExtraWorkTest() throws Exception {
		when(authService.canManageWork(anyLong())).thenReturn(true);

		doNothing().when(declarationService).approveAllOpenInExtraWork(anyLong(), anyLong());

		mvc.perform(post("/extra-work/approve-all-open/{id}", WORK_ID_1).with(csrf()))
				.andExpect(status().is3xxRedirection())
				.andExpect(redirectedUrl("/extra-work/" + WORK_ID_1));

		verify(authService).canManageWork(WORK_ID_1);
		verify(declarationService).approveAllOpenInExtraWork(WORK_ID_1, ADMIN_ID);
	}

	@Test
	@WithUserDetails("admin")
	void exportExtraWorkTest() throws Exception {
		when(authService.canManageWork(anyLong())).thenReturn(true);
		if (!Files.exists(Paths.get(exportDir))) {
			exportDir = Files.createTempDirectory("test-dir").toString();
		}
		when(workService.exportExtraWork(anyLong())).thenReturn(Path.of(exportDir));

		mvc.perform(get("/extra-work/{id}/export", WORK_ID_1))
				.andExpect(status().isOk());

		verify(authService).canManageWork(WORK_ID_1);
		verify(workService).exportExtraWork(WORK_ID_1);
	}

	@ParameterizedTest
	@MethodSource(value = "protectedEndpoints")
	void testWithoutUserDetailsIsForbidden(MockHttpServletRequestBuilder request) throws Exception {
		mvc.perform(request.with(csrf()))
				.andExpect(status().is3xxRedirection())
				.andExpect(redirectedUrl("http://localhost/login"));
	}

	private static List<MockHttpServletRequestBuilder> protectedEndpoints() {
		return List.of(
				get("/extra-work/open"),
				get("/extra-work/manage"),
				get("/extra-work/123"),
				post("/extra-work"),
				post("/extra-work/123"),
				post("/extra-work/close/123"),
				post("/extra-work/reject-all/123"));
	}
}
