/*
 * TAM
 * Copyright (C) 2021 - Delft University of Technology
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package nl.tudelft.tam.controller;

import static org.mockito.Mockito.*;
import static org.springframework.security.test.web.servlet.request.SecurityMockMvcRequestPostProcessors.csrf;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.get;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.post;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.*;

import java.util.List;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.MethodSource;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.autoconfigure.web.servlet.AutoConfigureMockMvc;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.boot.test.mock.mockito.MockBean;
import org.springframework.http.MediaType;
import org.springframework.security.test.context.support.WithUserDetails;
import org.springframework.test.web.servlet.MockMvc;
import org.springframework.test.web.servlet.request.MockHttpServletRequestBuilder;
import org.springframework.test.web.servlet.request.MockMvcRequestBuilders;
import org.springframework.transaction.annotation.Transactional;

import application.test.TestTAMApplication;
import nl.tudelft.labracore.api.dto.*;
import nl.tudelft.tam.cache.EditionCacheManager;
import nl.tudelft.tam.dto.patch.NotificationPatchDTO;
import nl.tudelft.tam.dto.patch.ShirtPatchDTO;
import nl.tudelft.tam.enums.NotificationFrequency;
import nl.tudelft.tam.enums.ShirtGender;
import nl.tudelft.tam.enums.ShirtSize;
import nl.tudelft.tam.model.Profile;
import nl.tudelft.tam.security.AuthorisationService;
import nl.tudelft.tam.service.PersonService;
import nl.tudelft.tam.service.ProfileService;
import nl.tudelft.tam.service.ProgramService;
import nl.tudelft.tam.service.RoleService;
import nl.tudelft.tam.test.TestUserDetailsService;

@AutoConfigureMockMvc
@SpringBootTest(classes = TestTAMApplication.class)
@Transactional
class TAMProfileControllerTest {

	@Autowired
	MockMvc mvc;

	@MockBean
	ProfileService profileService;

	@MockBean
	AuthorisationService authService;

	@MockBean
	ProgramService programService;

	@MockBean
	EditionCacheManager editionCache;

	@MockBean
	PersonService personService;

	@MockBean
	RoleService roleService;

	Profile profile;

	@BeforeEach
	void setup() {
		profile = Profile.builder()
				.id(TestUserDetailsService.id)
				.build();
	}

	@Test
	@WithUserDetails("username")
	void getProfile() throws Exception {
		when(authService.isTeacherOfAny()).thenReturn(false);

		PersonSummaryDTO testPerson = new PersonSummaryDTO().displayName("Test");

		when(profileService.findByIdOrThrow(anyLong())).thenReturn(profile);
		when(personService.getOrThrow(anyLong())).thenReturn(testPerson);

		mvc.perform(get("/profile"))
				.andExpect(status().isOk())
				.andExpect(view().name("profile/view"))
				.andExpect(model().attribute("profile", profile))
				.andExpect(model().attribute("person", testPerson));
	}

	@Test
	@WithUserDetails("username")
	void getPersonProfile() throws Exception {
		Long personId = 1L;

		Profile profile = Profile.builder().id(personId).build();
		PersonSummaryDTO person = new PersonSummaryDTO().id(personId).displayName("Test Person");
		List<ProgramDetailsDTO> coordinatingPrograms = List.of(new ProgramDetailsDTO());
		List<RoleEditionDetailsDTO> roles = List.of(new RoleEditionDetailsDTO()
				.type(RoleEditionDetailsDTO.TypeEnum.TA).edition(new EditionSummaryDTO().id(4L)));

		when(authService.isStaff()).thenReturn(true);
		when(profileService.findByIdOrThrow(anyLong())).thenReturn(profile);
		when(personService.getOrThrow(anyLong())).thenReturn(person);
		when(programService.getCoordinatingPrograms(anyLong())).thenReturn(coordinatingPrograms);
		when(roleService.getStaffRolesForPerson(anyLong())).thenReturn(roles);
		when(editionCache.getOrThrow(anyLong()))
				.thenReturn(new EditionDetailsDTO().course(new CourseSummaryDTO().name("course name")));

		mvc.perform(get("/profile/{personId}", personId))
				.andExpect(status().isOk())
				.andExpect(view().name("profile/view"))
				.andExpect(model().attribute("profile", profile))
				.andExpect(model().attribute("person", person))
				.andExpect(model().attribute("teacherView", true))
				.andExpect(model().attribute("coordinatingPrograms", coordinatingPrograms))
				.andExpect(model().attribute("roles", roles));

		verify(profileService).findByIdOrThrow(personId);
		verify(personService).getOrThrow(personId);
		verify(programService).getCoordinatingPrograms(personId);
		verify(roleService).getStaffRolesForPerson(personId);
	}

	@Test
	@Transactional
	@WithUserDetails("username")
	void updatePoloData() throws Exception {
		ShirtSize size = ShirtSize.M;
		ShirtGender gender = ShirtGender.FEMALE;
		ShirtPatchDTO patch = new ShirtPatchDTO(size, gender);

		doNothing().when(profileService).updatePoloData(anyLong(), any());

		mvc.perform(MockMvcRequestBuilders.patch("/profile/polo").with(csrf())
				.contentType(MediaType.APPLICATION_FORM_URLENCODED)
				.flashAttr("poloPatch", patch))
				.andExpect(status().is3xxRedirection())
				.andExpect(redirectedUrl("/profile"));

		verify(profileService).updatePoloData(anyLong(), eq(patch));
	}

	@Test
	@Transactional
	@WithUserDetails("username")
	void updateNotificationPreferences() throws Exception {
		NotificationFrequency frequency = NotificationFrequency.Daily;
		NotificationPatchDTO patch = new NotificationPatchDTO(false, frequency);

		doNothing().when(profileService).updateNotificationPreferences(anyLong(), any());

		mvc.perform(MockMvcRequestBuilders.patch("/profile/notifications").with(csrf())
				.contentType(MediaType.APPLICATION_FORM_URLENCODED)
				.flashAttr("notificationPatch", patch))
				.andExpect(status().is3xxRedirection())
				.andExpect(redirectedUrl("/profile"));

		verify(profileService).updateNotificationPreferences(anyLong(), eq(patch));
	}

	@ParameterizedTest
	@MethodSource(value = "protectedEndpoints")
	void testWithoutUserDetailsIsForbidden(MockHttpServletRequestBuilder request) throws Exception {
		mvc.perform(request.with(csrf()))
				.andExpect(status().is3xxRedirection())
				.andExpect(redirectedUrl("http://localhost/login"));
	}

	private static List<MockHttpServletRequestBuilder> protectedEndpoints() {
		return List.of(
				get("/profile"),
				post("/profile/polo"));
	}

}
