/*
 * TAM
 * Copyright (C) 2021 - Delft University of Technology
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package nl.tudelft.tam.io;

import static org.assertj.core.api.Assertions.assertThat;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.when;

import java.time.LocalDateTime;
import java.util.Arrays;
import java.util.List;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.boot.test.mock.mockito.MockBean;
import org.springframework.transaction.annotation.Transactional;

import application.test.TestTAMApplication;
import nl.tudelft.labracore.api.dto.*;
import nl.tudelft.tam.model.JobOffer;
import nl.tudelft.tam.service.CohortService;
import nl.tudelft.tam.service.CourseService;
import nl.tudelft.tam.service.EditionService;

@SpringBootTest(classes = TestTAMApplication.class)
@Transactional
public class JobOfferImportHandlerTest {

	@Autowired
	private JobOfferImportHandler importHandler;

	@MockBean
	private EditionService editionService;

	@MockBean
	private CourseService courseService;

	@MockBean
	private CohortService cohortService;

	@BeforeEach
	void setUp() {
		when(editionService.getEditionsById(any())).thenReturn(List.of(
				new EditionDetailsDTO().id(1L).name("edition").course(new CourseSummaryDTO().code("CSE1100")),
				new EditionDetailsDTO().id(2L).name("edition 2")
						.course(new CourseSummaryDTO().code("CSE1200"))));
		when(courseService.getAllCourses()).thenReturn(List.of(
				new CourseSummaryDTO().id(1L).code("CSE1100")));
		when(cohortService.getCohortsById(any())).thenReturn(List.of(
				new CohortDetailsDTO().id(1L).name("cohort")
						.program(new ProgramSummaryDTO().name("programme"))));
	}

	//	 * [0] edition name, [1] course code, [2] job name (TAM), [3] contract name (Flex)
	//				* [4] contract start date, [5] contract end date,
	//	 * [6] baancode, [7] maximum hours, [9] job information, [10] hiring message,
	//	 * [11] reject message, [12] deadline
	//				* (, [13] cohort name, [14] program name, [15] edition start date, [16] edition end date)

	@Test
	void importRegular() {
		String content = """
				edition,CSE1100,TA,TA,01-01-2023,01-01-2024,IIT EIP404,10,,no,,,yes,01-01-2023
				edition,CSE1100,Mentor,Mentor,01-01-2023,01-01-2024,IIT EIP404,10,,no,,,no,01-01-2023
				edition 2,CSE1200,TA,TA,01-01-2023,01-01-2024,IIT EIP404,10,,yes,,,no,01-01-2023
				""".stripIndent();

		ImportHandler.Result<JobOffer> res = importHandler
				.importFile(Arrays.stream(content.split("\n")).map(l -> l.split(",")).toList());

		assertThat(res.errors()).isEmpty();
		assertThat(res.warnings()).isEmpty();
		assertThat(res.result().size()).isEqualTo(3);
		assertThat(res.result().get(0).getName()).isEqualTo("TA");
		assertThat(res.result().get(1).getName()).isEqualTo("Mentor");
		assertThat(res.result().get(2).getName()).isEqualTo("TA");
	}

	@Test
	void importCreateEdition() {
		String content = """
				new edition,CSE1100,TA,TA,01-01-2023,01-01-2024,IIT EIP404,10,,no,,,no,01-01-2023,cohort,programme,01-01-2023 00:00,01-01-2024 00:00
				"""
				.stripIndent();

		ImportHandler.Result<JobOffer> res = importHandler
				.importFile(Arrays.stream(content.split("\n")).map(l -> l.split(",")).toList());

		assertThat(res.errors()).containsKey(1);
		ImportError actual = res.errors().get(1).get(0);
		assertThat(actual).isInstanceOf(JobOfferImportHandler.EditionDoesNotExistError.class);
		assertThat(((JobOfferImportHandler.EditionDoesNotExistError) actual).edition())
				.isEqualTo(new EditionCreateDTO()
						.name("new edition")
						.course(new CourseIdDTO().id(1L))
						.cohort(new CohortIdDTO().id(1L))
						.startDate(LocalDateTime.of(2023, 1, 1, 0, 0))
						.endDate(LocalDateTime.of(2024, 1, 1, 0, 0))
						.enrollability(EditionCreateDTO.EnrollabilityEnum.OPEN));
	}

	@Test
	void importWrongStartDate() {
		String content = """
				edition,CSE1100,TA,TA,2023-01-2023,01-01-2024,IIT EIP404,10,,no,,,no,01-01-2023
				""".stripIndent();

		ImportHandler.Result<JobOffer> res = importHandler
				.importFile(Arrays.stream(content.split("\n")).map(l -> l.split(",")).toList());

		assertThat(res.errors()).containsKey(1);
		assertThat(res.errors().get(1)).contains(
				new SimpleImportError(
						"Error on line 0: Column 'Start date' contained invalid content. Values allowed are expected in the format: dd-MM-yyyy"));
	}

	@Test
	void importWrongEndDate() {
		String content = """
				edition,CSE1100,TA,TA,01-01-2023,01-16-2024,IIT EIP404,10,,no,,,no,01-01-2023
				""".stripIndent();

		ImportHandler.Result<JobOffer> res = importHandler
				.importFile(Arrays.stream(content.split("\n")).map(l -> l.split(",")).toList());

		assertThat(res.errors()).containsKey(1);
		assertThat(res.errors().get(1)).contains(
				new SimpleImportError(
						"Error on line 0: Column 'End date' contained invalid content. Values allowed are expected in the format: dd-MM-yyyy"));
	}

	@Test
	void importWrongAmountOfHours() {
		String content = """
				edition,CSE1100,TA,TA,01-01-2023,01-01-2024,IIT EIP404,20.5,,no,,,no,01-01-2023
				""".stripIndent();

		ImportHandler.Result<JobOffer> res = importHandler
				.importFile(Arrays.stream(content.split("\n")).map(l -> l.split(",")).toList());

		assertThat(res.errors()).containsKey(1);
		assertThat(res.errors().get(1))
				.contains(new SimpleImportError("Error on line 0: Max hours is not a whole number"));
	}

	@Test
	void importWrongRequireResponse() {
		String content = """
				edition,CSE1100,TA,TA,01-01-2023,01-01-2024,IIT EIP404,20,,aaaaa,,,no,01-01-2023
				""".stripIndent();

		ImportHandler.Result<JobOffer> res = importHandler
				.importFile(Arrays.stream(content.split("\n")).map(l -> l.split(",")).toList());

		assertThat(res.errors()).containsKey(1);
		assertThat(res.errors().get(1)).contains(
				new SimpleImportError(
						"Error on line 0: Column 'Require response' contained invalid content. Values allowed are 'yes' or 'no'"));
	}

	@Test
	void importWrongDraft() {
		String content = """
				edition,CSE1100,TA,TA,01-01-2023,01-01-2024,IIT EIP404,20,,yes,,,eeeeeeee,01-01-2023
				""".stripIndent();

		ImportHandler.Result<JobOffer> res = importHandler
				.importFile(Arrays.stream(content.split("\n")).map(l -> l.split(",")).toList());

		assertThat(res.errors()).containsKey(1);
		assertThat(res.errors().get(1)).contains(
				new SimpleImportError(
						"Error on line 0: Column 'draft' contained invalid content. Values allowed are 'yes' or 'no'"));
	}
}
