/*
 * TAM
 * Copyright (C) 2021 - Delft University of Technology
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package nl.tudelft.tam.repository;

import static org.assertj.core.api.Assertions.assertThat;
import static org.junit.jupiter.api.Assertions.assertThrows;

import java.time.LocalDate;
import java.util.List;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.data.rest.webmvc.ResourceNotFoundException;
import org.springframework.transaction.annotation.Transactional;

import application.test.TestTAMApplication;
import nl.tudelft.tam.model.JobOffer;

@Transactional
@SpringBootTest(classes = TestTAMApplication.class)
class JobOfferRepositoryTest {

	@Autowired
	private JobOfferRepository repository;

	private JobOffer jobOffer1;
	private JobOffer jobOffer2;
	private JobOffer jobOffer3;
	private final Long id1 = 789765L;
	private final Long id2 = 984654L;
	private final Long id3 = 321498L;

	@BeforeEach
	void setup() {
		repository.deleteAll();
		jobOffer1 = JobOffer.builder()
				.id(id1)
				.editionId(41L)
				.deadline(LocalDate.now().minusDays(1))
				.contractName("Test")
				.draft(true)
				.contractStartDate(LocalDate.now())
				.contractEndDate(LocalDate.now().plusWeeks(2))
				.baanCode("XX1 XXX999")
				.maxHours(100)
				.build();

		jobOffer2 = JobOffer.builder()
				.id(id2)
				.editionId(42L)
				.deadline(LocalDate.now())
				.draft(true)
				.contractName("Test 3")
				.contractStartDate(LocalDate.now())
				.contractEndDate(LocalDate.now().plusWeeks(2))
				.baanCode("XX3 XXX999")
				.maxHours(100)
				.build();

		jobOffer3 = JobOffer.builder()
				.id(id3)
				.editionId(43L)
				.deadline(LocalDate.now().plusDays(1))
				.contractName("Test 3")
				.draft(true)
				.contractStartDate(LocalDate.now())
				.contractEndDate(LocalDate.now().plusWeeks(2))
				.baanCode("XX3 XXX999")
				.maxHours(100)
				.build();
	}

	@Test
	void findByIdOrThrow() {
		assertThrows(ResourceNotFoundException.class, () -> repository.findByIdOrThrow(id1));
		JobOffer offer_1 = repository.save(jobOffer1);
		assertThat(repository.findByIdOrThrow(offer_1.getId()))
				.isEqualTo(offer_1);
	}

	@Test
	void findAllOpenAnddraft() {
		assert (repository.findAllOpen()).isEmpty();
		repository.save(jobOffer1);
		repository.save(jobOffer2);
		repository.save(jobOffer3);
		List<JobOffer> res = repository.findAllOpen();
		assertThat(res).isEmpty();
	}

	@Test
	void findAllOpenAndSomeNotdraft() {
		assert (repository.findAllOpen()).isEmpty();
		jobOffer1.setDraft(false);
		jobOffer2.setDraft(false);
		JobOffer offer_1 = repository.save(jobOffer1);
		JobOffer offer_2 = repository.save(jobOffer2);
		JobOffer offer_3 = repository.save(jobOffer3);
		List<JobOffer> res = repository.findAllOpen();

		assertThat(res.size()).isEqualTo(1);
		assertThat(res).contains(offer_2);
	}

	@Test
	void findAllClosed() {
		assert (repository.findAllOpen()).isEmpty();
		JobOffer offer_1 = repository.save(jobOffer1);
		JobOffer offer_2 = repository.save(jobOffer2);
		JobOffer offer_3 = repository.save(jobOffer3);

		List<JobOffer> res = repository.findAllClosed();

		assertThat(res.size()).isEqualTo(1);
		assertThat(res).contains(offer_1);
	}

	@Test
	void findAllByEditions() {
		assert (repository.findAllOpen()).isEmpty();

		JobOffer offer_1 = repository.save(jobOffer1);
		JobOffer offer_2 = repository.save(jobOffer2);
		JobOffer offer_3 = repository.save(jobOffer3);
		List<JobOffer> res = repository.findAllByEditions(List.of(41L, 43L));

		assertThat(res.size()).isEqualTo(2);
		assertThat(res).contains(offer_1, offer_3);
	}

}
