/*
 * TAM
 * Copyright (C) 2021 - Delft University of Technology
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package nl.tudelft.tam.service;

import static org.assertj.core.api.Assertions.assertThat;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.time.LocalDate;
import java.time.LocalDateTime;
import java.util.*;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.CsvSource;
import org.modelmapper.ModelMapper;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.boot.test.mock.mockito.MockBean;
import org.springframework.transaction.annotation.Transactional;

import application.test.TestTAMApplication;
import nl.tudelft.labracore.api.CourseControllerApi;
import nl.tudelft.labracore.api.EditionControllerApi;
import nl.tudelft.labracore.api.PersonControllerApi;
import nl.tudelft.labracore.api.dto.*;
import nl.tudelft.tam.dto.create.ApplicationCreateDTO;
import nl.tudelft.tam.dto.id.JobOfferIdDTO;
import nl.tudelft.tam.dto.view.details.JobOfferDetailsDTO;
import nl.tudelft.tam.dto.view.summary.ApplicationStatusSummaryDTO;
import nl.tudelft.tam.dto.view.summary.ApplicationSummaryDTO;
import nl.tudelft.tam.dto.view.summary.JobOfferSummaryDTO;
import nl.tudelft.tam.enums.NotificationFrequency;
import nl.tudelft.tam.enums.PayScale;
import nl.tudelft.tam.enums.Status;
import nl.tudelft.tam.exception.ApplicationStatusException;
import nl.tudelft.tam.model.*;
import nl.tudelft.tam.repository.*;
import reactor.core.publisher.Flux;
import reactor.core.publisher.Mono;

@SpringBootTest(classes = TestTAMApplication.class)
@Transactional
class ApplicationServiceTest {

	@Autowired
	ApplicationService service;

	@Autowired
	ApplicationRepository repository;

	@MockBean
	RoleService roleService;

	@Autowired
	PersonControllerApi personControllerApi;

	@Autowired
	JobOfferRepository jobOfferRepository;

	@Autowired
	ProfileRepository profileRepository;

	@Autowired
	ApplicationChangeEventRepository applicationChangeEventRepository;

	@Autowired
	ModelMapper mapper;

	@Autowired
	EditionControllerApi editionApi;

	@Autowired
	CourseControllerApi courseApi;

	@Autowired
	CoordinatorRepository coordinatorRepository;

	@MockBean
	EmailService emailService;

	@MockBean
	JobOfferService jobOfferService;

	JobOffer offer1, offer2, offer3;

	Application app1a, app1b, app2a;

	PersonSummaryDTO personSummary1, personSummary2;

	long JOFFER_ID_1;
	long JOFFER_ID_2;
	long JOFFER_ID_3;
	final long PERSON_ID_1 = 5005;
	final long PERSON_ID_2 = 5007;
	final long EDITION_ID = 7007;
	final long COURSE_ID = 8007;
	final long PROGRAM_ID = 9007;
	final long ADMIN_ID = 329476L;

	@BeforeEach
	void setUp() {
		repository.deleteAll();
		jobOfferRepository.deleteAll();
		profileRepository.deleteAll();

		offer1 = jobOfferRepository.save(JobOffer.builder().name("offer-1").editionId(EDITION_ID)
				.deadline(LocalDate.now().plusDays(1)).contractName("Test")
				.contractStartDate(LocalDate.now())
				.contractEndDate(LocalDate.now().plusWeeks(2))
				.baanCode("TE1 XXX999").maxHours(100).build());
		offer2 = jobOfferRepository.save(JobOffer.builder().name("offer-2").editionId(EDITION_ID)
				.deadline(LocalDate.now().plusDays(2)).contractName("Test 2")
				.contractStartDate(LocalDate.now())
				.contractEndDate(LocalDate.now().plusWeeks(2))
				.baanCode("TE2 XXX999").maxHours(100).build());
		offer3 = jobOfferRepository.save(JobOffer.builder().name("offer-3").editionId(EDITION_ID)
				.deadline(LocalDate.now().plusDays(3)).contractName("Test 3")
				.contractStartDate(LocalDate.now())
				.contractEndDate(LocalDate.now().plusWeeks(2))
				.baanCode("TE3 XXX999").maxHours(100).build());

		JOFFER_ID_1 = offer1.getId();
		JOFFER_ID_2 = offer2.getId();
		JOFFER_ID_3 = offer3.getId();

		app1a = repository.save(Application.builder().id(new Application.AppId(PERSON_ID_1, JOFFER_ID_1))
				.jobOffer(offer1).status(Status.SUBMITTED).build());
		app1b = repository.save(Application.builder().id(new Application.AppId(PERSON_ID_2, JOFFER_ID_1))
				.jobOffer(offer1).status(Status.OFFERED).build());
		app2a = repository.save(Application.builder().id(new Application.AppId(PERSON_ID_1, JOFFER_ID_2))
				.jobOffer(offer2).status(Status.REJECTED_BY_STUDENT).build());

		personSummary1 = new PersonSummaryDTO().id(PERSON_ID_1).username("Person1").displayName("Person 1")
				.email("Person1@email.com");
		personSummary2 = new PersonSummaryDTO().id(PERSON_ID_2).username("Person2").displayName("Person 2")
				.email("Person2@email.com");

		profileRepository.save(Profile.builder()
				.id(PERSON_ID_1)
				.emailNotifications(true)
				.emailNotificationFrequency(NotificationFrequency.Daily)
				.payScale(PayScale.SA2)
				.createdDate(LocalDate.now())
				.build());
		profileRepository.save(Profile.builder()
				.id(PERSON_ID_2)
				.emailNotifications(true)
				.emailNotificationFrequency(NotificationFrequency.Daily)
				.payScale(PayScale.SA2)
				.createdDate(LocalDate.now())
				.build());

		profileRepository.save(Profile.builder().id(ADMIN_ID).build());
	}

	@Test
	void submitVerifiesNew() {
		EditionDetailsDTO edition = new EditionDetailsDTO().id(EDITION_ID).name("Test").course(
				new CourseSummaryDTO().name("Test Course").code("TC"));

		assertThat(repository.findById(new Application.AppId(PERSON_ID_2, JOFFER_ID_2))).isNotPresent();

		when(personControllerApi.getPeopleById(anyList())).thenReturn(Flux.just(personSummary2));
		when(editionApi.getEditionsById(anyList())).thenReturn(Flux.just(edition));
		// Should be a Future, but result is never used
		when(emailService.sendEmail(anyList(), any())).thenReturn(null);

		LocalDateTime now = LocalDateTime.now();
		service.submit(PERSON_ID_2,
				ApplicationCreateDTO.builder().jobOffer(new JobOfferIdDTO(JOFFER_ID_2)).build());

		assertThat(repository.findById(new Application.AppId(PERSON_ID_2, JOFFER_ID_2))).isPresent();
		Application application = repository.findById(new Application.AppId(PERSON_ID_2, JOFFER_ID_2)).get();
		assertThat(application.getStatus()).isEqualTo(Status.SUBMITTED);

		List<ApplicationChangeEvent> events = applicationChangeEventRepository
				.findByApplicationIdAndChangedAtAfter(new Application.AppId(PERSON_ID_2, JOFFER_ID_2), now);
		assertThat(events).isEmpty();
	}

	// Flakey Test?
	//	@Test
	//	void submitVerifiesResubmit() {
	//		assertThat(repository.findById(new Application.AppId(PERSON_ID_1, JOFFER_ID_2))).isPresent();
	//		assertThat(repository.findById(new Application.AppId(PERSON_ID_1, JOFFER_ID_2)).get().getStatus())
	//				.isEqualTo(Status.REJECTED_BY_STUDENT);
	//
	//		LocalDateTime now = LocalDateTime.now();
	//		service.submit(PERSON_ID_1, JOFFER_ID_2);
	//
	//		assertThat(repository.findById(new Application.AppId(PERSON_ID_1, JOFFER_ID_2))).isPresent();
	//		assertThat(repository.findById(new Application.AppId(PERSON_ID_1, JOFFER_ID_2)).get().getStatus())
	//				.isEqualTo(Status.SUBMITTED);
	//
	//		List<ApplicationChangeEvent> events = applicationChangeEventRepository
	//				.findByApplicationIdAndChangedAtAfter(new Application.AppId(PERSON_ID_1, JOFFER_ID_2), now);
	//		assertThat(events.size()).isEqualTo(1);
	//		assertThat(events.get(0).getApplication()).isEqualTo(app2a);
	//		assertThat(events.get(0).getStatus()).isEqualTo(Status.SUBMITTED);
	//	}

	@Test
	void submitVerifiesDuplicate() {
		assertThat(repository.findById(new Application.AppId(PERSON_ID_1, JOFFER_ID_1))).isPresent();
		assertThat(repository.findById(new Application.AppId(PERSON_ID_1, JOFFER_ID_1)).get().getStatus())
				.isEqualTo(Status.SUBMITTED);

		assertThrows(ApplicationStatusException.class, () -> service.submit(PERSON_ID_1,
				ApplicationCreateDTO.builder().jobOffer(new JobOfferIdDTO(JOFFER_ID_1)).build()));

		assertThat(repository.findById(new Application.AppId(PERSON_ID_1, JOFFER_ID_1))).isPresent();
		assertThat(repository.findById(new Application.AppId(PERSON_ID_1, JOFFER_ID_1)).get().getStatus())
				.isEqualTo(Status.SUBMITTED);
	}

	@Test
	void retract() {
		assertThat(repository.findById(new Application.AppId(PERSON_ID_1, JOFFER_ID_1))).isPresent();
		assertThat(repository.findById(new Application.AppId(PERSON_ID_1, JOFFER_ID_1)).get().getStatus())
				.isEqualTo(Status.SUBMITTED);

		service.retract(PERSON_ID_1, JOFFER_ID_1);

		assertThat(repository.findById(new Application.AppId(PERSON_ID_1, JOFFER_ID_1))).isEmpty();
	}

	@Test
	void retractOffered() {
		assertThat(repository.findById(new Application.AppId(PERSON_ID_2, JOFFER_ID_1))).isPresent();
		assertThat(repository.findById(new Application.AppId(PERSON_ID_2, JOFFER_ID_1)).get().getStatus())
				.isEqualTo(Status.OFFERED);

		service.retract(PERSON_ID_2, JOFFER_ID_1);

		assertThat(repository.findById(new Application.AppId(PERSON_ID_2, JOFFER_ID_1))).isPresent();
		assertThat(repository.findById(new Application.AppId(PERSON_ID_2, JOFFER_ID_1)).get().getStatus())
				.isEqualTo(Status.REJECTED_BY_STUDENT);
	}

	@Test
	void offerForSubmitted() {
		assertThat(repository.findById(new Application.AppId(PERSON_ID_1, JOFFER_ID_1))).isPresent();
		assertThat(repository.findById(new Application.AppId(PERSON_ID_1, JOFFER_ID_1)).get().getStatus())
				.isEqualTo(Status.SUBMITTED);

		EditionDetailsDTO edition = new EditionDetailsDTO().id(EDITION_ID).name("Test").course(
				new CourseSummaryDTO().name("Test Course").code("TC"));
		when(editionApi.getEditionsById(anyList())).thenReturn(Flux.just(edition));
		when(jobOfferService.getById(anyLong())).thenReturn(new JobOfferDetailsDTO());
		// Should be a Future, but result is never used
		when(emailService.sendEmail(anyList(), any())).thenReturn(null);

		service.offer(PERSON_ID_1, JOFFER_ID_1, ADMIN_ID);

		assertThat(repository.findById(new Application.AppId(PERSON_ID_1, JOFFER_ID_1))).isPresent();
		assertThat(repository.findById(new Application.AppId(PERSON_ID_1, JOFFER_ID_1)).get().getStatus())
				.isEqualTo(Status.OFFERED);
		assertThat(repository.findById(new Application.AppId(PERSON_ID_1, JOFFER_ID_1)).get().getHandler()
				.getId()).isEqualTo(ADMIN_ID);
	}

	@Test
	void offerForNew() {
		EditionDetailsDTO edition = new EditionDetailsDTO().id(EDITION_ID).name("Test").course(
				new CourseSummaryDTO().name("Test Course").code("TC"));

		assertThat(repository.findById(new Application.AppId(PERSON_ID_2, JOFFER_ID_2))).isNotPresent();

		when(personControllerApi.getPeopleById(anyList())).thenReturn(Flux.just(personSummary2));
		when(editionApi.getEditionsById(anyList())).thenReturn(Flux.just(edition));
		when(jobOfferService.getById(anyLong())).thenReturn(new JobOfferDetailsDTO());
		// Should be a Future, but result is never used
		when(emailService.sendEmail(anyList(), any())).thenReturn(null);

		service.offer(PERSON_ID_2, JOFFER_ID_2, ADMIN_ID);

		assertThat(repository.findById(new Application.AppId(PERSON_ID_2, JOFFER_ID_2))).isPresent();
		assertThat(repository.findById(new Application.AppId(PERSON_ID_2, JOFFER_ID_2)).get().getStatus())
				.isEqualTo(Status.OFFERED);
		assertThat(repository.findById(new Application.AppId(PERSON_ID_2, JOFFER_ID_2)).get().getHandler()
				.getId()).isEqualTo(ADMIN_ID);
	}

	@ParameterizedTest
	@CsvSource({ "OFFERED", "ACCEPTED" })
	void offerNotValid(Status status) {
		Optional<Application> app = repository.findById(new Application.AppId(PERSON_ID_1, JOFFER_ID_1));
		assertThat(app).isPresent();
		app.get().setStatus(status);
		assertThat(repository.findById(new Application.AppId(PERSON_ID_1, JOFFER_ID_1)).get().getStatus())
				.isEqualTo(status);

		service.offer(PERSON_ID_1, JOFFER_ID_1, ADMIN_ID);

		assertThat(repository.findById(new Application.AppId(PERSON_ID_1, JOFFER_ID_1))).isPresent();
		assertThat(repository.findById(new Application.AppId(PERSON_ID_1, JOFFER_ID_1)).get().getStatus())
				.isEqualTo(status);
		assertThat(repository.findById(new Application.AppId(PERSON_ID_1, JOFFER_ID_1)).get().getHandler())
				.isNull();
	}

	@ParameterizedTest
	@CsvSource({ "REJECTED_BY_STUDENT", "REJECTED_BY_TEACHER", "SUBMITTED", "ACCEPTED" })
	void acceptVerifiesOnlyOffered(Status status) {
		Optional<Application> app = repository.findById(new Application.AppId(PERSON_ID_1, JOFFER_ID_1));
		assertThat(app).isPresent();
		app.get().setStatus(status);

		assertThrows(ApplicationStatusException.class, () -> service.accept(PERSON_ID_1, JOFFER_ID_1));

		assertThat(repository.findById(new Application.AppId(PERSON_ID_1, JOFFER_ID_1))).isPresent();
		assertThat(repository.findById(new Application.AppId(PERSON_ID_1, JOFFER_ID_1)).get().getStatus())
				.isEqualTo(status);
	}

	@Test
	void appExistsFor() {
		assertThat(repository.findById(new Application.AppId(PERSON_ID_2, JOFFER_ID_1))).isPresent();
		assertThat(service.appExistsFor(PERSON_ID_2, JOFFER_ID_1)).isTrue();
	}

	@Test
	void hasOfferValidatesTrue() {
		assertThat(repository.findById(new Application.AppId(PERSON_ID_2, JOFFER_ID_1))).isPresent();
		assertThat(repository.findById(new Application.AppId(PERSON_ID_2, JOFFER_ID_1)).get().getStatus())
				.isEqualTo(Status.OFFERED);
		assertThat(service.hasOffer(PERSON_ID_2, JOFFER_ID_1)).isTrue();
		assertThat(repository.findById(new Application.AppId(PERSON_ID_2, JOFFER_ID_1))).isPresent();
		assertThat(repository.findById(new Application.AppId(PERSON_ID_2, JOFFER_ID_1)).get().getStatus())
				.isEqualTo(Status.OFFERED);
	}

	@Test
	void hasOfferValidatesFalse() {
		assertThat(repository.findById(new Application.AppId(PERSON_ID_1, JOFFER_ID_1))).isPresent();
		assertThat(repository.findById(new Application.AppId(PERSON_ID_1, JOFFER_ID_1)).get().getStatus())
				.isEqualTo(Status.SUBMITTED);
		assertThat(service.hasOffer(PERSON_ID_1, JOFFER_ID_1)).isFalse();
		assertThat(repository.findById(new Application.AppId(PERSON_ID_1, JOFFER_ID_1))).isPresent();
		assertThat(repository.findById(new Application.AppId(PERSON_ID_1, JOFFER_ID_1)).get().getStatus())
				.isEqualTo(Status.SUBMITTED);
	}

	@Test
	void getStatsForApps() {
		List<ApplicationSummaryDTO> list = List.of(
				ApplicationSummaryDTO.builder().status(Status.SUBMITTED).build(),
				ApplicationSummaryDTO.builder().status(Status.SUBMITTED).build(),
				ApplicationSummaryDTO.builder().status(Status.REJECTED_BY_STUDENT).build(),
				ApplicationSummaryDTO.builder().status(Status.REJECTED_BY_TEACHER).build(),
				ApplicationSummaryDTO.builder().status(Status.OFFERED).build(),
				ApplicationSummaryDTO.builder().status(Status.OFFERED).build(),
				ApplicationSummaryDTO.builder().status(Status.OFFERED).build(),
				ApplicationSummaryDTO.builder().status(Status.ACCEPTED).build(),
				ApplicationSummaryDTO.builder().status(Status.ACCEPTED).build());

		assertThat(service.getStatsForApps(list)).isEqualTo(new int[] { 8, 3, 2, 1 });
	}

	@Test
	void getMapAppOpenByIdForPerson() {
		Map<Long, Boolean> res = service.getMapAppOpenByIdForPerson(PERSON_ID_2,
				List.of(JOFFER_ID_1, JOFFER_ID_2));
		assertThat(res.size()).isEqualTo(2);
		assertThat(res).containsKeys(JOFFER_ID_1, JOFFER_ID_2);
		assertThat(res.get(JOFFER_ID_1)).isTrue();
		assertThat(res.get(JOFFER_ID_2)).isFalse();
	}

	@Test
	void getMapWithStatusAppOpenByIdForPerson() {
		Map<Long, ApplicationStatusSummaryDTO> res = service.getMapWithStatusAppOpenByIdForPerson(PERSON_ID_2,
				List.of(JOFFER_ID_1, JOFFER_ID_2));

		ApplicationStatusSummaryDTO jobOffer1Summary = new ApplicationStatusSummaryDTO(true, Status.OFFERED);
		ApplicationStatusSummaryDTO jobOffer2Summary = new ApplicationStatusSummaryDTO(false, null);

		assertThat(res.size()).isEqualTo(2);
		assertThat(res).containsKeys(JOFFER_ID_1, JOFFER_ID_2);
		assertThat(res.get(JOFFER_ID_1)).isEqualTo(jobOffer1Summary);
		assertThat(res.get(JOFFER_ID_2)).isEqualTo(jobOffer2Summary);
	}

	@Test
	void getApplicationsForPersonByStatus() {
		List<Application> testList = new ArrayList<>();
		testList.add(app1a);
		List<Application> emptyList = new ArrayList<>();
		assertThat(service.getApplicationsForPersonByStatus(PERSON_ID_1, Status.SUBMITTED))
				.isEqualTo(testList);
		assertThat(service.getApplicationsForPersonByStatus(PERSON_ID_1, Status.OFFERED))
				.isEqualTo(emptyList);
	}

	@Test
	void bulkOfferWithUniformIdentifiers() {
		List<String> ids = List.of("student1", "student2");
		String idString = "student1\n" + "student2";
		PersonSummaryDTO s1 = new PersonSummaryDTO().id(PERSON_ID_1);
		PersonSummaryDTO s2 = new PersonSummaryDTO().id(PERSON_ID_2);
		PersonSearchDTO search = new PersonSearchDTO().people(Set.of(s1, s2)).notFound(List.of(""));
		EditionDetailsDTO edition = new EditionDetailsDTO().id(EDITION_ID).name("Test").course(
				new CourseSummaryDTO().name("Test Course").code("TC"));

		when(personControllerApi.searchForPeople(ids)).thenReturn(Mono.just(search));
		when(personControllerApi.searchPeopleByUsernameContaining(anyString())).thenReturn(Flux.just(s1));
		when(personControllerApi.searchPeopleByDisplayNameContaining(anyString())).thenReturn(Flux.just(s2));
		when(personControllerApi.getPeopleById(anyList()))
				.thenReturn(Flux.just(personSummary1, personSummary2));
		when(editionApi.getEditionsById(anyList())).thenReturn(Flux.just(edition));
		when(jobOfferService.getById(anyLong())).thenReturn(new JobOfferDetailsDTO());
		// Should be a Future, but result is never used
		when(emailService.sendEmail(anyList(), any())).thenReturn(null);

		service.bulkOffer(JOFFER_ID_3, idString, ADMIN_ID);

		assertThat(repository.findById(new Application.AppId(PERSON_ID_1, JOFFER_ID_3)).get().getStatus())
				.isEqualTo(Status.OFFERED);
		assertThat(repository.findById(new Application.AppId(PERSON_ID_2, JOFFER_ID_3)).get().getStatus())
				.isEqualTo(Status.OFFERED);
		assertThat(repository.findById(new Application.AppId(PERSON_ID_2, JOFFER_ID_3)).get().getHandler()
				.getId()).isEqualTo(ADMIN_ID);
	}

	@Test
	void findAllByOfferIdTest() {
		assertThat(service.findAllByOfferId(JOFFER_ID_1)).isEqualTo(List.of(app1a, app1b));
	}

	@Test
	void getCoordinatingApplicationsForProgram() {
		when(editionApi.getEditionsById(anyList())).thenReturn(Flux
				.just(new EditionDetailsDTO().id(EDITION_ID).course(new CourseSummaryDTO().id(COURSE_ID))));
		when(courseApi.getAllCoursesById(anyList())).thenReturn(Flux
				.just(new CourseDetailsDTO().id(COURSE_ID).program(new ProgramSummaryDTO().id(PROGRAM_ID))));
		coordinatorRepository.save(Coordinator.builder().programId(PROGRAM_ID).personId(ADMIN_ID).build());
		assertThat(service.getCoordinatingApplicationsForProgram(ADMIN_ID, PROGRAM_ID))
				.containsExactlyInAnyOrder(app1a, app1b, app2a);
	}

	@Test
	void hasApplicationInEditionNoJobOffers() {
		when(jobOfferService.getAllInEditions(anyList())).thenReturn(new ArrayList<>());
		assertFalse(service.hasApplicationInEdition(3L, 5L));
	}

	@Test
	void hasApplicationInEditionNoJobOfferForPerson() {
		when(jobOfferService.getAllInEditions(anyList()))
				.thenReturn(List.of(JobOfferSummaryDTO.builder().id(JOFFER_ID_2).build()));
		assertFalse(service.hasApplicationInEdition(PERSON_ID_2, 5L));
		assertTrue(service.hasApplicationInEdition(PERSON_ID_1, 5L));
	}

}
