/*
 * TAM
 * Copyright (C) 2021 - Delft University of Technology
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package nl.tudelft.tam.service;

import static org.assertj.core.api.Assertions.assertThat;

import java.io.File;
import java.io.StringReader;
import java.net.MalformedURLException;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.List;

import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.core.io.Resource;
import org.springframework.http.HttpStatus;
import org.springframework.http.ResponseEntity;
import org.springframework.transaction.annotation.Transactional;

import application.test.TestTAMApplication;

@Transactional
@SpringBootTest(classes = TestTAMApplication.class)
class CSVServiceTest {

	@Autowired
	private CSVService service;

	private final String fileName = "test.csv";

	@Test
	void detectSeparator() throws Exception {
		List<String[]> rows = service
				.readCSV(new StringReader("this, is a file;that has, both, commas, and semicolons;1\n" +
						"a semicolon;should be detected, as separator;2"));
		assertThat(rows.size()).isEqualTo(2);
		assertThat(rows.get(0))
				.isEqualTo(new String[] { "this, is a file", "that has, both, commas, and semicolons", "1" });
		assertThat(rows.get(1))
				.isEqualTo(new String[] { "a semicolon", "should be detected, as separator", "2" });
	}

	@Test
	void writeReadCSV() throws Exception {
		// prepare the data
		List<String[]> data = new ArrayList<>();
		data.add(new String[] { "name", "age" });
		data.add(new String[] { "John", "13" });
		data.add(new String[] { "Jane", "65" });
		data.add(new String[] { "Jack, with comma", "3" });

		// write to file
		service.writeToCSV(data, Path.of(""), fileName);

		Path path = Path.of(fileName);

		// check that the file exists
		File file = new File(path.toString());
		assertThat(file).exists();
		assertThat(file.isDirectory()).isFalse();

		// retrieve with default config
		List<String[]> retrieved = service.readCSV(path);
		assertThat(retrieved.size()).isEqualTo(data.size());
		for (int i = 0; i < retrieved.size(); i++) {
			assertThat(retrieved.get(i)).isEqualTo(data.get(i));
		}

		// retrieve with ignoring first line
		data.remove(0);
		retrieved = service.readCSV(path, false, 1);
		assertThat(retrieved.size()).isEqualTo(data.size());
		for (int i = 0; i < retrieved.size(); i++) {
			assertThat(retrieved.get(i)).isEqualTo(data.get(i));
		}

		// delete the file
		file.delete();
	}

	@Test
	void setHeaders() throws MalformedURLException {
		ResponseEntity<Resource> response = service.getResponse(Paths.get("hello"));
		assertThat(response.getStatusCode()).isEqualTo(HttpStatus.OK);
	}

}
