/*
 * TAM
 * Copyright (C) 2021 - Delft University of Technology
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package nl.tudelft.tam.service;

import static org.assertj.core.api.Assertions.assertThat;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.io.IOException;
import java.time.LocalDate;
import java.util.ArrayList;
import java.util.List;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.modelmapper.ModelMapper;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.boot.test.mock.mockito.MockBean;
import org.springframework.boot.test.mock.mockito.SpyBean;
import org.springframework.transaction.annotation.Transactional;

import application.test.TestTAMApplication;
import nl.tudelft.labracore.api.dto.*;
import nl.tudelft.labracore.lib.security.user.Person;
import nl.tudelft.tam.dto.view.details.ExtraWorkDetailsDTO;
import nl.tudelft.tam.enums.PayScale;
import nl.tudelft.tam.enums.Status;
import nl.tudelft.tam.model.*;
import nl.tudelft.tam.repository.DeclarationRepository;
import nl.tudelft.tam.repository.ProfileRepository;

@SpringBootTest(classes = TestTAMApplication.class)
@Transactional
public class DeclarationServiceMockTest {

	@SpyBean
	DeclarationService service;

	@MockBean
	DeclarationRepository repository;

	@MockBean
	CoordinatorService coordinatorService;

	@MockBean
	EditionService editionService;

	@MockBean
	CourseService courseService;

	@MockBean
	ExtraWorkService extraWorkService;

	@MockBean
	PersonService personService;

	@MockBean
	CohortService cohortService;

	@MockBean
	CSVService csvService;

	@MockBean
	JobOfferService jobOfferService;

	@MockBean
	CoordinatorDefaultService coordinatorDefaultService;

	@MockBean
	ProfileRepository profileRepository;

	@Autowired
	ModelMapper mapper;

	Profile person1, person2;
	EditionDetailsDTO edition;
	ExtraWork work1, work2;
	Declaration dec1a, dec1b, dec2;
	PersonSummaryDTO personSummary1, personSummary2;
	EditionDetailsDTO editionDetails;
	CourseDetailsDTO courseDetails;
	CourseSummaryDTO courseSummary;
	CohortDetailsDTO cohortDetails;
	ProgramSummaryDTO programSummary;
	CohortSummaryDTO cohortSummary;

	final long PERSON_ID_1 = 100;
	final long PERSON_ID_2 = 101;
	final long EDITION_ID = 7007;
	final long WORK_ID_1 = 1;
	final long WORK_ID_2 = 2;
	final long DEC_ID_1A = 201;
	final long DEC_ID_1B = 202;
	final long DEC_ID_2 = 203;
	final long ADMIN_ID = 329476;
	final long COURSE_ID = 10932;
	final long PROGRAM_ID = 1;
	final long COHORT_ID = 219;

	@BeforeEach
	void setUp() {
		edition = new EditionDetailsDTO().name("Edition 1").id(EDITION_ID)
				.course(new CourseSummaryDTO().name("course-1").code("code-1"))
				.cohort(new CohortSummaryDTO().name("COHOrt-1"));

		work1 = ExtraWork.builder().id(WORK_ID_1).name("work 1").editionId(EDITION_ID)
				.deadline(LocalDate.now().minusDays(7)).maxHours(20).build();
		work2 = ExtraWork.builder().id(WORK_ID_2).name("work 2").editionId(EDITION_ID)
				.deadline(LocalDate.now().plusDays(7)).build();

		person1 = Profile.builder().id(PERSON_ID_1).build();
		person2 = Profile.builder().id(PERSON_ID_2).build();

		personSummary1 = new PersonSummaryDTO().id(PERSON_ID_1).username("Person1").displayName("Person 1")
				.email("Person1@email.com").number(1);
		personSummary2 = new PersonSummaryDTO().id(PERSON_ID_2).username("Person2").displayName("Person 2")
				.email("Person2@email.com").number(2);

		courseDetails = new CourseDetailsDTO().id(COURSE_ID).name("Course");
		courseSummary = new CourseSummaryDTO().id(COURSE_ID).name("Course");
		programSummary = new ProgramSummaryDTO().id(PROGRAM_ID).name("Program");
		cohortDetails = new CohortDetailsDTO().id(COHORT_ID).name("Cohort").program(programSummary);
		cohortSummary = new CohortSummaryDTO().id(COHORT_ID).name("Cohort");
		editionDetails = new EditionDetailsDTO().id(EDITION_ID).name("Edition").course(courseSummary)
				.cohort(cohortSummary);

		dec1a = Declaration.builder().id(DEC_ID_1A).person(person1).extraWork(work1)
				.status(Status.SUBMITTED).amount(10).build();
		dec1b = Declaration.builder().id(DEC_ID_1B).person(person2).extraWork(work1)
				.status(Status.ACCEPTED).amount(20).build();
		dec2 = Declaration.builder().id(DEC_ID_2).person(person1).extraWork(work2)
				.status(Status.SUBMITTED).amount(30).build();
	}

	@Test
	void getCoordinatingDeclarationsTest() {
		Long programId = 1L;
		Long courseId = 100L;

		Coordinator coordinator = new Coordinator();
		coordinator.setProgramId(programId);

		when(coordinatorService.findAllByPersonId(PERSON_ID_1)).thenReturn(List.of(coordinator));

		when(repository.findAll()).thenReturn(List.of(dec1a));

		when(editionService.getOrThrow(eq(EDITION_ID)))
				.thenReturn(
						new EditionDetailsDTO().id(EDITION_ID).course(new CourseSummaryDTO().id(courseId)));
		when(courseService.getOrThrow(eq(courseId)))
				.thenReturn(
						new CourseDetailsDTO().id(courseId).program(new ProgramSummaryDTO().id(programId)));

		List<Declaration> result = service.getCoordinatingDeclarations(PERSON_ID_1);

		assertThat(result).isEqualTo(List.of(dec1a));

		verify(coordinatorService).findAllByPersonId(anyLong());
		verify(repository).findAll();
	}

	@Test
	void getNumberOfUnhandledHoursTest() {
		when(extraWorkService.getById(eq(WORK_ID_1)))
				.thenReturn(mapper.map(work1, ExtraWorkDetailsDTO.class));
		when(repository.findAllByPersonId(PERSON_ID_1)).thenReturn(List.of(dec1a, dec2));

		assertThat(service.numberOfHoursRemainingForExtraWork(PERSON_ID_1, WORK_ID_1)).isEqualTo(17.5);
	}

	@Test
	void exportDeclarationsFlexDelft() throws IOException {
		List<Declaration> declarations = new ArrayList<>();
		declarations.add(dec1a);
		declarations.add(dec1b);

		String name = "test-application-flex-delft";

		when(personService.get(any()))
				.thenReturn(List.of(personSummary1, personSummary2));
		when(personService.getOrThrow(PERSON_ID_1)).thenReturn(personSummary1);
		when(personService.getOrThrow(PERSON_ID_2)).thenReturn(personSummary2);

		when(editionService.getOrThrow(EDITION_ID)).thenReturn(editionDetails);
		when(cohortService.get(any())).thenReturn(List.of(cohortDetails));
		when(cohortService.getOrThrowById(anyLong())).thenReturn(cohortDetails);

		when(jobOfferService.getApprover(anyLong())).thenReturn(personSummary2);
		when(profileRepository.findByIdOrThrow(anyLong())).thenReturn(Profile.builder()
				.id(PERSON_ID_1)
				.payScale(PayScale.SA2)
				.build());
		when(coordinatorDefaultService.getOrCreate(anyLong())).thenReturn(CoordinatorDefault.builder()
				.programId(PROGRAM_ID)
				.coordinatorContractEmail("email@coordinator.com")
				.coordinatorContractName("Contract Name Default")
				.build());

		CourseDetailsDTO course = new CourseDetailsDTO();
		ProgramSummaryDTO program = new ProgramSummaryDTO();
		program.name("idk");
		course.program(program);
		when(courseService.getCourseById(anyLong())).thenReturn(course);

		@SuppressWarnings("unchecked") ArgumentCaptor<List<String[]>> captor = ArgumentCaptor
				.forClass(List.class);
		service.exportDeclarationsForFlexDelft(Person.builder().id(1L).build(), List.of(1L), 1, name,
				declarations,
				null);
		verify(csvService).writeToCSV(captor.capture(), any(), eq(name + ".csv"));

		assertThat(captor.getValue())
				.contains(
						new String[] { "Name",
								"Email",
								"Course Code",
								"Contract Name",
								"Baan Code",
								"Approver Name",
								"Approver Email",
								"Coordinator",
								"Max Hours",
								"Payscale",
								"Program",
								"Batch",
								"Remarks" },

						new String[] { "Person 1",
								"Person1@email.com",
								null,
								null,
								null,
								"Contract Name Default",
								"email@coordinator.com",
								"Contract Name Default",
								"2.5",
								"SA2",
								"idk",
								"1",
								"" },

						new String[] { "Person 2",
								"Person2@email.com",
								null,
								null,
								null,
								"Contract Name Default",
								"email@coordinator.com",
								"Contract Name Default",
								"5.0",
								"SA2",
								"idk",
								"1",
								"" });
	}

}
