/*
 * TAM
 * Copyright (C) 2021 - Delft University of Technology
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package nl.tudelft.tam.service;

import static org.assertj.core.api.Assertions.assertThat;
import static org.assertj.core.api.Assertions.assertThatThrownBy;
import static org.mockito.Mockito.*;

import java.time.LocalDate;
import java.util.List;
import java.util.stream.Collectors;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.CsvSource;
import org.modelmapper.ModelMapper;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.boot.test.mock.mockito.MockBean;

import application.test.TestTAMApplication;
import nl.tudelft.labracore.api.dto.*;
import nl.tudelft.librador.dto.DTOConverter;
import nl.tudelft.tam.dto.create.DeclarationCreateDTO;
import nl.tudelft.tam.dto.view.details.DeclarationDetailsExtraWorkDTO;
import nl.tudelft.tam.dto.view.details.DeclarationDetailsPersonDTO;
import nl.tudelft.tam.dto.view.details.ExtraWorkDetailsDTO;
import nl.tudelft.tam.dto.view.summary.DeclarationSummaryDTO;
import nl.tudelft.tam.enums.Status;
import nl.tudelft.tam.model.*;
import nl.tudelft.tam.repository.*;

@SpringBootTest(classes = TestTAMApplication.class)
class DeclarationServiceTest {

	@Autowired
	DeclarationService declarationService;

	@Autowired
	DeclarationRepository declarationRepository;

	@Autowired
	ExtraWorkService workService;

	@MockBean
	PersonService personService;

	@MockBean
	EditionService editionService;

	@Autowired
	ExtraWorkRepository workRepository;

	@Autowired
	ProfileRepository profileRepository;

	@Autowired
	ModelMapper mapper;

	@Autowired
	DeclarationChangeEventRepository declarationChangeEventRepository;

	@MockBean
	EmailService emailService;

	@Autowired
	DTOConverter converter;

	Profile person1, person2;

	EditionDetailsDTO edition;

	ExtraWork work1, work2;

	Declaration dec1a, dec1b, dec2, dec2b;

	final long PERSON_ID_1 = 100;
	final long PERSON_ID_2 = 101;
	final long EDITION_ID = 7007;
	long WORK_ID_1, WORK_ID_2;
	long DEC_ID_1A, DEC_ID_1B, DEC_ID_2, DEC_ID_2B;
	final long ADMIN_ID = 329476L;

	@BeforeEach
	void setUp() {
		declarationChangeEventRepository.deleteAll();
		declarationRepository.deleteAll();
		workRepository.deleteAll();
		profileRepository.deleteAll();

		edition = new EditionDetailsDTO().name("Edition 1").id(EDITION_ID)
				.course(new CourseSummaryDTO().name("course-1").code("code-1"))
				.cohort(new CohortSummaryDTO().name("COHOrt-1"));

		work1 = workRepository.save(ExtraWork.builder().name("work 1").editionId(EDITION_ID)
				.contractName("Extra-Work-1-Contract")
				.baanCode("XX1 XXX999")
				.deadline(LocalDate.now().minusDays(7)).maxHours(100).build());
		work2 = workRepository.save(ExtraWork.builder().name("work 2").editionId(EDITION_ID)
				.contractName("Extra-Work-2-Contract")
				.baanCode("XX2 XXX999")
				.deadline(LocalDate.now().plusDays(7)).maxHours(100).build());

		WORK_ID_1 = work1.getId();
		WORK_ID_2 = work2.getId();

		person1 = profileRepository.save(Profile.builder().id(PERSON_ID_1).build());
		person2 = profileRepository.save(Profile.builder().id(PERSON_ID_2).build());
		profileRepository.save(Profile.builder().id(ADMIN_ID).build());

		dec1a = declarationRepository
				.save(Declaration.builder().id(DEC_ID_1A).person(person1).extraWork(work1)
						.status(Status.SUBMITTED).amount(10).reason("Test Reason 1a").build());
		dec1b = declarationRepository
				.save(Declaration.builder().id(DEC_ID_1B).person(person2).extraWork(work1)
						.status(Status.ACCEPTED).amount(20).reason("Test Reason 1b").build());
		dec2 = declarationRepository.save(Declaration.builder().id(DEC_ID_2).person(person1).extraWork(work2)
				.status(Status.SUBMITTED).amount(30).reason("Test Reason 2").build());
		dec2b = declarationRepository.save(Declaration.builder().id(DEC_ID_2B).person(person1)
				.extraWork(work2).status(Status.REJECTED_BY_TEACHER).amount(1000).reason("Haha").build());

		DEC_ID_1A = dec1a.getId();
		DEC_ID_1B = dec1b.getId();
		DEC_ID_2 = dec2.getId();
		DEC_ID_2B = dec2b.getId();
	}

	@Test
	void submitTest() {
		DeclarationCreateDTO dto = DeclarationCreateDTO.builder().personId(PERSON_ID_2).workId(WORK_ID_2)
				.amount(1).reason("Test Reason").build();
		declarationService.submit(dto);
		List<Declaration> res = declarationRepository.findAllByPersonId(PERSON_ID_2).stream().distinct()
				.filter(declarationRepository.findAllByWorkId(WORK_ID_2)::contains)
				.collect(Collectors.toList());
		assertThat(res.size()).isEqualTo(1);
		Declaration comparison = converter.apply(dto);
		comparison.setId(res.get(0).getId()); // ignore ID
		comparison.setPerson(profileRepository.findByIdOrThrow(PERSON_ID_2));
		comparison.setExtraWork(workRepository.findByIdOrThrow(WORK_ID_2));
		assertThat(res.get(0)).usingDefaultComparator().isEqualTo(comparison);

		DeclarationCreateDTO dtoFailing = DeclarationCreateDTO.builder().personId(PERSON_ID_2)
				.workId(WORK_ID_2)
				.amount(9999999).reason("Fail bc exceeds max hours.").build();
		assertThatThrownBy(() -> declarationService.submit(dtoFailing))
				.isInstanceOf(IllegalArgumentException.class);
	}

	@Test
	void retract() {
		assertThat(declarationRepository.findByIdOrThrow(DEC_ID_1A).getStatus()).isEqualTo(Status.SUBMITTED);
		declarationService.retract(DEC_ID_1A);
		assertThat(declarationRepository.findById(DEC_ID_1A)).isEmpty();
	}

	@Test
	void reject() {
		EditionDetailsDTO edition = new EditionDetailsDTO().id(EDITION_ID).name("Test").course(
				new CourseSummaryDTO().name("Test Course").code("TC"));
		when(editionService.getEditionById(anyLong())).thenReturn(edition);
		when(personService.getPersonById(anyLong()))
				.thenReturn(new PersonDetailsDTO().id(PERSON_ID_1).email("test@test.com")
						.displayName("Test Name"));
		// Should be a Future, but result is never used
		when(emailService.sendEmail(anyList(), any())).thenReturn(null);

		assertThat(declarationRepository.findByIdOrThrow(DEC_ID_1A).getStatus()).isEqualTo(Status.SUBMITTED);
		declarationService.reject(DEC_ID_1A, ADMIN_ID);
		assertThat(declarationRepository.findByIdOrThrow(DEC_ID_1A).getStatus())
				.isEqualTo(Status.REJECTED_BY_TEACHER);
		assertThat(declarationRepository.findByIdOrThrow(DEC_ID_1A).getHandler().getId())
				.isEqualTo(ADMIN_ID);
	}

	@Test
	void approveTest() {
		EditionDetailsDTO edition = new EditionDetailsDTO().id(EDITION_ID).name("Test").course(
				new CourseSummaryDTO().name("Test Course").code("TC"));
		when(editionService.getEditionById(anyLong())).thenReturn(edition);
		when(personService.getPersonById(anyLong()))
				.thenReturn(new PersonDetailsDTO().id(PERSON_ID_1).email("test@test.com")
						.displayName("Test Name"));
		// Should be a Future, but result is never used
		when(emailService.sendEmail(anyList(), any())).thenReturn(null);

		assertThat(declarationRepository.findByIdOrThrow(DEC_ID_1A).getStatus()).isEqualTo(Status.SUBMITTED);
		declarationService.approve(DEC_ID_1A, ADMIN_ID);
		assertThat(declarationRepository.findByIdOrThrow(DEC_ID_1A).getStatus()).isEqualTo(Status.ACCEPTED);
		assertThat(declarationRepository.findByIdOrThrow(DEC_ID_1A).getHandler().getId())
				.isEqualTo(ADMIN_ID);
	}

	@Test
	void rejectAllOpenInExtraWorkTest() {
		EditionDetailsDTO edition = new EditionDetailsDTO().id(EDITION_ID).name("Test").course(
				new CourseSummaryDTO().name("Test Course").code("TC"));
		when(editionService.getEditionById(anyLong())).thenReturn(edition);
		when(personService.getPersonById(anyLong()))
				.thenReturn(new PersonDetailsDTO().id(PERSON_ID_1).email("test@test.com")
						.displayName("Test Name"));
		// Should be a Future, but result is never used
		when(emailService.sendEmail(anyList(), any())).thenReturn(null);

		assertThat(declarationRepository.findByIdOrThrow(DEC_ID_1A).getStatus()).isEqualTo(Status.SUBMITTED);
		assertThat(declarationRepository.findByIdOrThrow(DEC_ID_1B).getStatus()).isEqualTo(Status.ACCEPTED);

		declarationService.rejectAllOpenInExtraWork(WORK_ID_1, ADMIN_ID);

		assertThat(declarationRepository.findByIdOrThrow(DEC_ID_1A).getStatus())
				.isEqualTo(Status.REJECTED_BY_TEACHER);
		assertThat(declarationRepository.findByIdOrThrow(DEC_ID_1A).getHandler().getId())
				.isEqualTo(ADMIN_ID);
		assertThat(declarationRepository.findByIdOrThrow(DEC_ID_1B).getStatus()).isEqualTo(Status.ACCEPTED);
		assertThat(declarationRepository.findByIdOrThrow(DEC_ID_1B).getHandler()).isNull();
	}

	@Test
	void approveAllOpenInExtraWorkTest() {
		EditionDetailsDTO edition = new EditionDetailsDTO().id(EDITION_ID).name("Test").course(
				new CourseSummaryDTO().name("Test Course").code("TC"));
		when(editionService.getEditionById(anyLong())).thenReturn(edition);
		when(personService.getPersonById(anyLong()))
				.thenReturn(new PersonDetailsDTO().id(PERSON_ID_1).email("test@test.com")
						.displayName("Test Name"));
		// Should be a Future, but result is never used
		when(emailService.sendEmail(anyList(), any())).thenReturn(null);

		assertThat(declarationRepository.findByIdOrThrow(DEC_ID_1A).getStatus()).isEqualTo(Status.SUBMITTED);
		assertThat(declarationRepository.findByIdOrThrow(DEC_ID_1B).getStatus()).isEqualTo(Status.ACCEPTED);

		declarationService.approveAllOpenInExtraWork(WORK_ID_1, ADMIN_ID);

		assertThat(declarationRepository.findByIdOrThrow(DEC_ID_1A).getStatus())
				.isEqualTo(Status.ACCEPTED);
		assertThat(declarationRepository.findByIdOrThrow(DEC_ID_1A).getHandler().getId())
				.isEqualTo(ADMIN_ID);
		assertThat(declarationRepository.findByIdOrThrow(DEC_ID_1B).getStatus()).isEqualTo(Status.ACCEPTED);
		assertThat(declarationRepository.findByIdOrThrow(DEC_ID_1B).getHandler()).isNull();
	}

	@Test
	void personOwnsDeclarationTest() {
		assertThat(declarationService.personOwnsDeclaration(PERSON_ID_1, DEC_ID_1A)).isTrue();
		assertThat(declarationService.personOwnsDeclaration(PERSON_ID_1, DEC_ID_1B)).isFalse();
		assertThat(declarationService.personOwnsDeclaration(PERSON_ID_1, DEC_ID_2)).isTrue();
	}

	@Test
	void getExtraWorkIdFromDeclarationTest() {
		assertThat(declarationService.getExtraWorkIdFromDeclaration(DEC_ID_1A)).isEqualTo(WORK_ID_1);
		assertThat(declarationService.getExtraWorkIdFromDeclaration(DEC_ID_1B)).isEqualTo(WORK_ID_1);
		assertThat(declarationService.getExtraWorkIdFromDeclaration(DEC_ID_2)).isEqualTo(WORK_ID_2);
	}

	@Test
	void getDeclarationsForPersonTest() {
		when(editionService.getEditionsById(anyList())).thenReturn(List.of(edition));

		DeclarationDetailsExtraWorkDTO dto1 = mapper.map(dec1a, DeclarationDetailsExtraWorkDTO.class);
		DeclarationDetailsExtraWorkDTO dto2 = mapper.map(dec1b, DeclarationDetailsExtraWorkDTO.class);
		ExtraWorkDetailsDTO work = mapper.map(work1, ExtraWorkDetailsDTO.class);
		work.setEdition(edition);
		dto1.setExtraWork(work);
		dto2.setExtraWork(work);

		List<DeclarationDetailsExtraWorkDTO> res = declarationService.getDeclarationsForPerson(PERSON_ID_1);
		assertThat(res.size()).isEqualTo(3);
		assertThat(res).contains(dto1, dto2);

		verify(editionService).getEditionsById(List.of(EDITION_ID, EDITION_ID, EDITION_ID));
	}

	@Test
	void getDeclarationsForPersonTestEmpty() {
		when(editionService.getEditionsById(anyList())).thenReturn(List.of());

		declarationRepository.deleteAll();
		assertThat(declarationService.getDeclarationsForPerson(PERSON_ID_1)).isEmpty();

		verify(editionService, never()).getEditionsById(anyList());
	}

	@Test
	void getDeclarationsForExtraWorkTest() {
		PersonSummaryDTO p1 = new PersonSummaryDTO().id(PERSON_ID_1);
		PersonSummaryDTO p2 = new PersonSummaryDTO().id(PERSON_ID_2);
		when(personService.getPeopleById(anyList())).thenReturn(List.of(p1, p2));

		DeclarationDetailsPersonDTO dto1 = mapper.map(dec1a, DeclarationDetailsPersonDTO.class);
		DeclarationDetailsPersonDTO dto2 = mapper.map(dec2, DeclarationDetailsPersonDTO.class);
		dto1.setPerson(p1);
		dto2.setPerson(p2);

		List<DeclarationDetailsPersonDTO> res = declarationService.getDeclarationsForExtraWork(WORK_ID_1);
		assertThat(res.size()).isEqualTo(2);
		assertThat(res).contains(dto1, dto2);

		verify(personService).getPeopleById(List.of(PERSON_ID_1, PERSON_ID_2));
	}

	@Test
	void getDeclarationsForExtraWorkTestEmpty() {
		assertThat(declarationService.getDeclarationsForExtraWork(56484651L)).isEmpty();

		verify(personService, never()).getPeopleById(anyList());
	}

	@ParameterizedTest
	@CsvSource({ ", 3",
			"abadfsasdxx, 0",
			"work, 3",
			"submitted, 1",
			"accepted, 1",
			"course-1, 3",
			"code-1, 3",
			"CoDe-1, 3" })
	void getFilteredDeclarationsForPersonTest(String filter, int count) {
		dec2.setStatus(Status.ACCEPTED);
		declarationRepository.save(dec2);

		when(editionService.getEditionsById(anyList())).thenReturn(List.of(edition));

		List<DeclarationDetailsExtraWorkDTO> res = declarationService.getFilteredDeclarationsForPerson(
				PERSON_ID_1,
				filter);
		assertThat(res.size()).isEqualTo(count);
	}

	@ParameterizedTest
	@CsvSource({ ", 2",
			"sadfasdfasd, 0",
			"dis-, 2",
			"dis-1, 1",
			"uSEr-2, 1",
			"email-1, 1",
			"ext-2, 1",
			"11111, 1" })
	void getFilteredDeclarationsForExtraWorkTest(String filter, int count) {
		PersonSummaryDTO p1 = new PersonSummaryDTO().id(PERSON_ID_1).displayName("dis-1").username("user-1")
				.email("email-1").externalId("ext-1").number(11111);
		PersonSummaryDTO p2 = new PersonSummaryDTO().id(PERSON_ID_2).displayName("dis-2").username("user-2")
				.email("email-2").externalId("ext-2").number(22222);
		when(personService.getPeopleById(anyList())).thenReturn(List.of(p1, p2));

		List<DeclarationDetailsPersonDTO> res = declarationService.getFilteredDeclarationsForExtraWork(
				WORK_ID_1,
				filter);
		assertThat(res.size()).isEqualTo(count);
	}

	@Test
	void getStatsForDecsTest() {

		dec2.setStatus(Status.REJECTED_BY_STUDENT);

		DeclarationSummaryDTO d1 = mapper.map(dec1a, DeclarationSummaryDTO.class);
		DeclarationSummaryDTO d2 = mapper.map(dec1b, DeclarationSummaryDTO.class);
		DeclarationSummaryDTO d3 = mapper.map(dec2, DeclarationSummaryDTO.class);
		DeclarationSummaryDTO d4 = mapper.map(dec2b, DeclarationSummaryDTO.class);

		int[] res = declarationService.getStatsForDecs(List.of(d1, d2, d3, d4));

		assertThat(res.length).isEqualTo(4);
		assertThat(res[0]).isEqualTo(4);
		assertThat(res[1]).isEqualTo(1060);
		assertThat(res[2]).isEqualTo(20);
		assertThat(res[3]).isEqualTo(1000);
	}

	@Test
	void getNumberOfUnhandledHours() {
		float numberOfHours = declarationService.getNumberOfUnhandledHours(WORK_ID_1);
		assertThat(numberOfHours).isEqualTo(2.5f);
	}

	@Test
	void findAllByWorkIdTest() {
		assertThat(declarationService.findAllByWorkId(WORK_ID_1)).isEqualTo(List.of(dec1a, dec1b));
	}
}
