/*
 * TAM
 * Copyright (C) 2021 - Delft University of Technology
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package nl.tudelft.tam.service;

import static org.assertj.core.api.Assertions.assertThat;
import static org.junit.jupiter.api.Assertions.*;

import java.io.File;
import java.io.IOException;
import java.nio.file.Path;
import java.nio.file.Paths;

import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.core.io.Resource;
import org.springframework.core.io.UrlResource;
import org.springframework.mock.web.MockMultipartFile;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.util.StringUtils;

import application.test.TestTAMApplication;

@SpringBootTest(classes = TestTAMApplication.class)
@Transactional
class FileServiceTest {

	@Value("${tam.filesys.block-size}")
	private Integer SIZE_OF_BLOCKS;

	@Value("${tam.filesys.level-depth}")
	private Integer NUM_0F_LEVELS;

	final String STORAGE_DIR = "storage/test";

	@Autowired
	FileService fileService;

	private final static Long ID_1 = 1234L;
	private final static Long ID_2 = 56789L;

	@AfterEach
	void cleanUp() {
		new File(STORAGE_DIR).delete();
	}

	@Test
	void getDirectoryTest() {
		String code = fileService.hash(ID_1);

		if (!(SIZE_OF_BLOCKS == 2 && NUM_0F_LEVELS == 3))
			fail();

		assertThat(fileService.getDirectory(ID_1)).isEqualTo(code.substring(0, 2) + File.separator
				+ code.substring(2, 4) + File.separator + code.substring(4, 6) + File.separator + ID_1
				+ File.separator);
	}

	@Test
	void hashTest() {
		assertThat(fileService.hash(ID_1))
				.isEqualTo(fileService.hash(ID_1));
		assertThat(fileService.hash(ID_2))
				.isEqualTo(fileService.hash(ID_2));
		assertThat(fileService.hash(ID_1))
				.isNotEqualTo(fileService.hash(ID_2));
	}

	@Test
	void uploadFileTest() throws IOException {
		String submissionsDirectory = STORAGE_DIR + "/submissions";
		MockMultipartFile file = new MockMultipartFile("file", "filename.pdf", "application/pdf",
				"submission".getBytes());

		fileService.uploadFile(file, ID_1, submissionsDirectory);

		String path = submissionsDirectory + File.separator + fileService.getDirectory(ID_1);

		assertThat(new File(path)).exists();

		new File(path + File.separator + "filename.pdf").delete();
	}

	@Test
	void downloadFileTest() throws IOException {
		String submissionsDirectory = STORAGE_DIR + "/submissions";

		String submissionPath = fileService.getDirectory(ID_1);
		Path path = Paths
				.get(submissionsDirectory + File.separator + StringUtils.cleanPath(submissionPath));
		File dir = new File(String.valueOf(path));
		dir.mkdirs();
		File file = new File(dir, "file.txt");
		file.createNewFile();

		Resource resource = fileService.downloadFile(ID_1, submissionsDirectory);

		assertThat(resource).isEqualTo(new UrlResource(
				"file:" + submissionsDirectory + File.separator + StringUtils.cleanPath(submissionPath)
						+ "file.txt"));

		new File(submissionsDirectory + File.separator + StringUtils.cleanPath(submissionPath)
				+ "file.txt").delete();
	}

	@Test
	void downloadFileTestFailNoFile() throws IOException {
		String submissionsDirectory = STORAGE_DIR + "/submissions";

		String submissionPath = fileService.getDirectory(ID_1);
		Path path = Paths
				.get(submissionsDirectory + File.separator + StringUtils.cleanPath(submissionPath));
		File dir = new File(String.valueOf(path));
		dir.mkdirs();
		File file = new File(dir, "file.txt");
		file.createNewFile();

		new File(submissionsDirectory + File.separator + StringUtils.cleanPath(submissionPath)
				+ "file.txt").delete();

		assertThrows(IOException.class, () -> fileService.downloadFile(ID_1, submissionsDirectory));
	}

	@Test
	void downloadFileTestFailMoreFiles() throws IOException {
		String submissionsDirectory = STORAGE_DIR + "/submissions";

		String submissionPath = fileService.getDirectory(ID_1);
		Path path = Paths
				.get(submissionsDirectory + File.separator + StringUtils.cleanPath(submissionPath));
		File dir = new File(String.valueOf(path));
		dir.mkdirs();
		File file1 = new File(dir, "file1.txt");
		file1.createNewFile();
		File file2 = new File(dir, "file2.txt");
		file2.createNewFile();

		assertThrows(IOException.class, () -> fileService.downloadFile(ID_1, submissionsDirectory));

		new File(submissionsDirectory + File.separator + StringUtils.cleanPath(submissionPath)
				+ "file1.txt").delete();
		new File(submissionsDirectory + File.separator + StringUtils.cleanPath(submissionPath)
				+ "file2.txt").delete();
	}

	@Test
	void deleteDirectoryTest() throws IOException {
		File dir1 = new File("./test");
		File dir2 = new File("./test/test2");
		dir2.mkdirs();
		File file1 = new File(dir2, "file1.txt");
		File file2 = new File(dir2, "file2.txt");
		file1.createNewFile();
		file2.createNewFile();

		fileService.deleteDirectory(dir1);

		assertThat(new File("./test")).doesNotExist();
	}

	@Test
	void toFileNameTest() {
		assertThat(fileService.toFileName(
				"As̸̴̨̧̼̝̠̹̮̜̝͕͚̰̫̺̣̹͉̣̜̫̤̗̜̝̣̬̤͎͈̮̮̥͍͋̍͆̇͊̾͑̕͠͠͠ͅsignm̸̧̡͓̰̮̮̯͔̿eǹ̨̢̛̙̺̭͇̠̣͓͖͉͖̣̯̥̘͇͕̣́̍͝͝t 1: A̡̧̱͙͚͓̭̮̫͉͇͎͉̯̖͈̣̙̝̳͉̾̐̓̃͌̍͆̂̂͊̄̓͛͐̒͜͝ͅͅss̀̒̋̇̾̈́̅̕̚e̢̡͇̺̼͊͌͋̋̍̀̒̈́͋̕͝mblÿ̴̛́̄̚͝/̷̼̼̲̹̞͙͎̬͓͓̻͓̱̝̤̎̾"))
				.isEqualTo("assignment_1_assembly");
	}

}
