/*
 * TAM
 * Copyright (C) 2021 - Delft University of Technology
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package nl.tudelft.tam.service;

import static org.assertj.core.api.Assertions.assertThat;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.List;
import java.util.Optional;
import java.util.Set;
import java.util.stream.Stream;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.modelmapper.ModelMapper;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.boot.test.mock.mockito.MockBean;
import org.springframework.transaction.annotation.Transactional;

import application.test.TestTAMApplication;
import nl.tudelft.labracore.api.PersonControllerApi;
import nl.tudelft.labracore.api.dto.PersonDetailsDTO;
import nl.tudelft.labracore.api.dto.PersonSearchDTO;
import nl.tudelft.labracore.api.dto.PersonSummaryDTO;
import nl.tudelft.labracore.api.dto.RoleEditionDetailsDTO;
import nl.tudelft.labracore.api.dto.RoleId;
import nl.tudelft.tam.cache.PersonCacheManager;
import reactor.core.publisher.Flux;
import reactor.core.publisher.Mono;

@SpringBootTest(classes = TestTAMApplication.class)
@Transactional
public class PersonServiceTest {

	@Autowired
	PersonService service;

	@Autowired
	PersonControllerApi personControllerApi;

	@MockBean
	PersonCacheManager personCache;

	PersonSummaryDTO personSummary;
	PersonDetailsDTO personDetails;

	@Autowired
	ModelMapper mapper;

	@BeforeEach
	void setup() {
		personDetails = new PersonDetailsDTO().id(1L).displayName("Jake");
		personSummary = new PersonSummaryDTO().displayName("Jake").id(1L).username("jake123");
	}

	@Test
	void getRolesForPersonTest() {
		RoleEditionDetailsDTO roleEditionDetailsDTO = new RoleEditionDetailsDTO()
				.id(new RoleId().personId(1L));

		when(personControllerApi.getRolesForPerson(anyLong())).thenReturn(Flux.just(roleEditionDetailsDTO));

		assertThat(service.getRolesForPerson(1L)).isEqualTo(List.of(roleEditionDetailsDTO));

		verify(personControllerApi).getRolesForPerson(1L);
	}

	@Test
	void searchPeopleByDisplayNameContainingTest() {
		when(personControllerApi.searchPeopleByDisplayNameContaining(eq("Jake")))
				.thenReturn(Flux.just(personSummary));

		assertThat(service.searchPeopleByDisplayNameContaining("Jake")).isEqualTo(List.of(personSummary));

		verify(personControllerApi).searchPeopleByDisplayNameContaining(eq("Jake"));
	}

	@Test
	void searchPeopleByUsernameContainingTest() {
		when(personControllerApi.searchPeopleByUsernameContaining(eq("jake123")))
				.thenReturn(Flux.just(personSummary));

		assertThat(service.searchPeopleByUsernameContaining("jake123")).isEqualTo(List.of(personSummary));

		verify(personControllerApi).searchPeopleByUsernameContaining(eq("jake123"));
	}

	@Test
	void searchPeopleByNumberContainingTest() {
		when(personControllerApi.searchPeopleByNumberContaining(eq(1)))
				.thenReturn(Flux.just(personSummary));

		assertThat(service.searchPeopleByNumberContaining(1)).isEqualTo(List.of(personSummary));

		verify(personControllerApi).searchPeopleByNumberContaining(eq(1));
	}

	@Test
	void getPeopleByIdTest() {
		when(personCache.get(anyList())).thenReturn(List.of(personSummary));

		assertThat(service.getPeopleById(List.of(1L))).isEqualTo(List.of(personSummary));

		verify(personCache).get(List.of(1L));
	}

	@Test
	void searchForPeopleTest() {
		PersonSearchDTO psSearch = new PersonSearchDTO().people(Set.of(personSummary));
		when(personControllerApi.searchForPeople(List.of("jake123", "Jake"))).thenReturn(Mono.just(psSearch));
		when(personControllerApi.searchPeopleByUsernameContaining(eq("jake123")))
				.thenReturn(Flux.just(personSummary));
		when(personControllerApi.searchPeopleByUsernameContaining(eq("Jake")))
				.thenReturn(Flux.just(personSummary));
		when(personControllerApi.searchPeopleByDisplayNameContaining(eq("jake123")))
				.thenReturn(Flux.just(personSummary));
		when(personControllerApi.searchPeopleByDisplayNameContaining(eq("Jake")))
				.thenReturn(Flux.just(personSummary));

		String identifiers = "jake123\nJake";
		assertThat(service.searchForPeople(identifiers)).isEqualTo(List.of(personSummary));

		verify(personControllerApi).searchForPeople(List.of("jake123", "Jake"));
	}

	@Test
	void searchForPeopleNullTest() {
		assertThat(service.searchForPeople(null)).isEqualTo(List.of());
		verify(personControllerApi, times(0)).searchForPeople(List.of());
	}

	@Test
	void searchForPeopleEmptyTest() {
		assertThat(service.searchForPeople("")).isEqualTo(List.of());
		verify(personControllerApi, times(0)).searchForPeople(List.of());
	}

	@Test
	void getOrThrowTest() {
		when(personCache.getOrThrow(anyLong())).thenReturn(personSummary);

		assertThat(service.getOrThrow(1L)).isEqualTo(personSummary);

		verify(personCache).getOrThrow(eq(1L));
	}

	@Test
	void getPersonByUsernameTest() {
		when(personCache.get(anyString())).thenReturn(mapper.map(personDetails, PersonSummaryDTO.class));

		assertThat(service.getPersonByUsername("Jake")).isEqualTo(personDetails);

		verify(personCache).get(eq("Jake"));
	}

	@Test
	void getPersonByIdTest() {
		when(personCache.get(anyLong()))
				.thenReturn(Optional.of(mapper.map(personDetails, PersonSummaryDTO.class)));

		assertThat(service.getPersonById(1L)).isEqualTo(personDetails);

		verify(personCache).get(eq(1L));
	}

	@Test
	@SuppressWarnings("unchecked")
	void getStreamTest() {
		when(personCache.get(any(Stream.class))).thenReturn(List.of(personSummary));

		assertThat(service.get(Stream.of(1L))).isEqualTo(List.of(personSummary));

		verify(personCache).get(any(Stream.class));
	}
}
