/*
 * TAM
 * Copyright (C) 2021 - Delft University of Technology
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package nl.tudelft.tam.service;

import static org.assertj.core.api.Assertions.assertThat;
import static org.junit.jupiter.api.Assertions.*;

import java.io.IOException;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.core.io.Resource;
import org.springframework.core.io.UrlResource;
import org.springframework.mock.web.MockMultipartFile;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.web.multipart.MultipartFile;

import application.test.TestTAMApplication;
import nl.tudelft.labracore.lib.security.user.Person;
import nl.tudelft.tam.dto.create.RaiseRequestCreateDTO;
import nl.tudelft.tam.dto.id.ProfileIdDTO;
import nl.tudelft.tam.dto.view.summary.RaiseRequestSummaryDTO;
import nl.tudelft.tam.enums.PayScale;
import nl.tudelft.tam.enums.Status;
import nl.tudelft.tam.exception.FileHandlingException;
import nl.tudelft.tam.exception.InvalidDatabaseEntryException;
import nl.tudelft.tam.model.Profile;
import nl.tudelft.tam.model.RaiseRequest;
import nl.tudelft.tam.repository.ProfileRepository;
import nl.tudelft.tam.repository.RaiseRequestRepository;
import nl.tudelft.tam.test.TestUserDetailsService;

@SpringBootTest(classes = TestTAMApplication.class)
@Transactional
class RaiseRequestServiceTest {

	@Autowired
	RaiseRequestService raiseRequestService;

	@Autowired
	RaiseRequestRepository raiseRequestRepository;

	@Autowired
	ProfileRepository profileRepository;

	@Autowired
	FileService fileService;

	final PayScale SCALE_TO = PayScale.SA4;
	String PROOF = "proof-68431238496";

	Profile profile;
	RaiseRequestCreateDTO createDTO;
	RaiseRequest request;
	MultipartFile filea, fileb, filec;

	@Value("${tam.filesys.payscale-proofs-dir}")
	private String PROOF_DIRECTORY;

	@BeforeEach
	void setup() {
		profile = profileRepository.save(Profile.builder()
				.id(TestUserDetailsService.id)
				.build());

		createDTO = RaiseRequestCreateDTO.builder()
				.payScale(SCALE_TO)
				.requestedBy(new ProfileIdDTO(profile.getId()))
				.build();

		request = raiseRequestRepository.save(RaiseRequest.builder()
				.requestedBy(profile)
				.payScale(SCALE_TO)
				.build());

		filea = new MockMultipartFile("file-a.pdf", "file-a.pdf", "application/pdf", "file-a.pdf".getBytes());
		fileb = new MockMultipartFile("file-b.png", "file-b.png", "image/png", "file-b.png".getBytes());
		filec = new MockMultipartFile("file-c.txt", "file-c.txt", "text/plain", "file-c.txt".getBytes());

		try {
			fileService.uploadFile(filea, request.getId(), PROOF_DIRECTORY);
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	@Test
	void validateFileSubmissionVerifiesOneFile() {
		Map<String, MultipartFile> fileMap = new HashMap<>();
		fileMap.put("file-a.pdf", filea);
		fileMap.put("file-b.pdf", fileb);

		assertThrows(FileHandlingException.class, () -> raiseRequestService.validateFileSubmission(fileMap));
	}

	@Test
	void validateFileSubmissionVerifiesType() {
		Map<String, MultipartFile> fileMap = new HashMap<>();
		fileMap.put("c", filec);

		assertThrows(FileHandlingException.class, () -> raiseRequestService.validateFileSubmission(fileMap));
	}

	@Test
	void validateFileSubmission() {
		Map<String, MultipartFile> fileMap = new HashMap<>();
		fileMap.put("file-a.pdf", filea);

		assertEquals(raiseRequestService.validateFileSubmission(fileMap), filea);
	}

	@Test
	@Transactional
	void newRaiseRequestDuplicate() {
		assertThat(raiseRequestRepository.findByRequesterAndSubmitted(profile.getId())).isPresent();

		boolean res = raiseRequestService.newRaiseRequest(createDTO, fileb);

		assertThat(res).isFalse();
	}

	@Test
	@Transactional
	void newRaiseRequest() {
		raiseRequestRepository.deleteAll();
		assertThat(raiseRequestRepository.findByRequesterAndSubmitted(profile.getId())).isEmpty();

		boolean res = raiseRequestService.newRaiseRequest(createDTO, fileb);

		assertThat(res).isTrue();
		assertThat(raiseRequestRepository.findByRequesterAndSubmitted(profile.getId())).isPresent();
	}

	@Test
	@Transactional
	void newRaiseRequestFileSaveFailed() {
		raiseRequestRepository.deleteAll();
		assertThat(raiseRequestRepository.findByRequesterAndSubmitted(profile.getId())).isEmpty();

		boolean res = raiseRequestService.newRaiseRequest(createDTO, fileb);

		assertThat(res).isTrue();
		assertThat(raiseRequestRepository.findByRequesterAndSubmitted(profile.getId())).isPresent();
	}

	@Test
	@Transactional
	void getBy() {
		assertThat(raiseRequestService.getBy(request.getId())).isEqualTo(request);
	}

	@Test
	@Transactional
	void retract() {
		assertThat(raiseRequestRepository.findByIdOrThrow(request.getId()).getStatus())
				.isEqualTo(Status.SUBMITTED);

		raiseRequestService.retract(Person.builder().id(profile.getId()).build());

		assertThat(raiseRequestRepository.findById(request.getId())).isEmpty();
	}

	@Test
	void getAllSubmittedEmpty() throws InvalidDatabaseEntryException {
		raiseRequestRepository.deleteAll();

		assertThat(raiseRequestService.getAllSubmitted(List.of()).size()).isEqualTo(0);
	}

	@Test
	@Transactional
	void getOpenForPersonNull() {
		raiseRequestRepository.deleteAll();
		assertThat(raiseRequestRepository.findByRequesterAndSubmitted(profile.getId())).isEmpty();

		RaiseRequestSummaryDTO res = raiseRequestService
				.getOpenForPerson(Person.builder().id(profile.getId()).build());

		assertThat(res).isNull();
	}

	@Test
	@Transactional
	void getOpenForPerson() {
		assertThat(raiseRequestRepository.findByRequesterAndSubmitted(profile.getId())).isPresent();

		RaiseRequestSummaryDTO res = raiseRequestService
				.getOpenForPerson(Person.builder().id(profile.getId()).build());

		assertThat(res.getId()).isEqualTo(request.getId());
		assertThat(res.getPayScale()).isEqualTo(request.getPayScale());
		assertThat(res.getStatus()).isEqualTo(request.getStatus());
	}

	@Test
	@Transactional
	void downloadFile() throws IOException {
		assertThat(raiseRequestRepository.findById(request.getId())).isPresent();

		Resource res = raiseRequestService.downloadFile(request.getId());

		UrlResource shouldBe = new UrlResource("file:" + PROOF_DIRECTORY + "/"
				+ fileService.getDirectory(request.getId()) + filea.getOriginalFilename());

		assertThat(res).isEqualTo(shouldBe);
	}

}
