/*
 * TAM
 * Copyright (C) 2021 - Delft University of Technology
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package nl.tudelft.tam.service;

import static org.assertj.core.api.AssertionsForClassTypes.assertThat;
import static org.junit.jupiter.api.Assertions.assertTrue;
import static org.mockito.Mockito.*;

import java.util.List;
import java.util.Set;

import org.assertj.core.api.Assertions;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.boot.test.mock.mockito.MockBean;
import org.springframework.transaction.annotation.Transactional;

import application.test.TestTAMApplication;
import nl.tudelft.labracore.api.EditionControllerApi;
import nl.tudelft.labracore.api.PersonControllerApi;
import nl.tudelft.labracore.api.RoleControllerApi;
import nl.tudelft.labracore.api.dto.*;
import nl.tudelft.tam.cache.EditionRolesCacheManager;
import reactor.core.publisher.Flux;
import reactor.core.publisher.Mono;

@SpringBootTest(classes = TestTAMApplication.class)
@Transactional
public class RoleServiceTest {

	@Autowired
	RoleService service;

	@MockBean
	JobOfferService jobOfferService;

	@MockBean
	PersonService personService;

	@MockBean
	EditionRolesCacheManager editionRolesCache;

	@Autowired
	PersonControllerApi personApi;

	@MockBean
	EditionService editionService;

	@Autowired
	RoleControllerApi roleControllerApi;

	@Autowired
	EditionControllerApi editionControllerApi;

	final long JOFFER_ID = 1001;
	final long PERSON_ID = 5001;
	final long EDITION_ID = 7000;

	RoleEditionDetailsDTO roleCorrectEdition, roleCorrectEditionHeadTA;
	RoleEditionDetailsDTO roleIncorrectEdition;

	private RolePersonDetailsDTO roleTeacher1;
	private RolePersonDetailsDTO roleTA1;
	private RolePersonDetailsDTO roleTA2;
	private RolePersonDetailsDTO roleTeacher2;

	private RolePersonDetailsDTO roleStudent1;
	private RolePersonDetailsDTO roleHeadTA;
	private RolePersonDetailsDTO roleTeacherRO;

	private PersonSummaryDTO person;
	private PersonSummaryDTO person2;
	private PersonSummaryDTO person3;

	private PersonSummaryDTO person4;
	private PersonSummaryDTO person5;
	private PersonSummaryDTO person6;
	private PersonSummaryDTO person7;

	@BeforeEach
	void setup() {
		roleCorrectEdition = new RoleEditionDetailsDTO()
				.type(RoleEditionDetailsDTO.TypeEnum.STUDENT)
				.edition(new EditionSummaryDTO().id(EDITION_ID));

		roleCorrectEditionHeadTA = new RoleEditionDetailsDTO()
				.type(RoleEditionDetailsDTO.TypeEnum.HEAD_TA)
				.edition(new EditionSummaryDTO().id(EDITION_ID));

		roleIncorrectEdition = new RoleEditionDetailsDTO()
				.type(RoleEditionDetailsDTO.TypeEnum.STUDENT)
				.edition(new EditionSummaryDTO().id(0L));

		this.person = new PersonSummaryDTO().id(1L).displayName("person 1");
		this.person2 = new PersonSummaryDTO().id(2L).displayName("person 2");
		this.person3 = new PersonSummaryDTO().id(3L).displayName("person 3");
		this.person4 = new PersonSummaryDTO().id(4L).displayName("person 4");
		this.person5 = new PersonSummaryDTO().id(5L).displayName("person 5");
		this.person6 = new PersonSummaryDTO().id(6L).displayName("person 6");
		this.person7 = new PersonSummaryDTO().id(7L).displayName("person 6");

		this.roleTeacher1 = new RolePersonDetailsDTO().person(person)
				.type(RolePersonDetailsDTO.TypeEnum.TEACHER);
		this.roleTeacher2 = new RolePersonDetailsDTO().person(person3)
				.type(RolePersonDetailsDTO.TypeEnum.TEACHER);
		this.roleTeacherRO = new RolePersonDetailsDTO().person(person7)
				.type(RolePersonDetailsDTO.TypeEnum.TEACHER_RO);
		this.roleHeadTA = new RolePersonDetailsDTO().person(person5)
				.type(RolePersonDetailsDTO.TypeEnum.HEAD_TA);
		this.roleTA1 = new RolePersonDetailsDTO().person(person2).type(RolePersonDetailsDTO.TypeEnum.TA);
		this.roleTA2 = new RolePersonDetailsDTO().person(person6).type(RolePersonDetailsDTO.TypeEnum.TA);
		this.roleStudent1 = new RolePersonDetailsDTO().person(person4)
				.type(RolePersonDetailsDTO.TypeEnum.STUDENT);
	}

	@Test
	void hasRoleNonMatching() {
		when(editionRolesCache.getRolesInEdition(anyLong(), anyLong())).thenReturn(new EditionRolesCacheManager.RoleHolder(34L, List.of()));
		Boolean resultMatching = service.hasAnyRoleInEdition(PERSON_ID, EDITION_ID);
		assertThat(resultMatching).isFalse();
	}

	@Test
	void hasRoleMatching() {
		when(editionRolesCache.getRolesInEdition(anyLong(), anyLong())).thenReturn(new EditionRolesCacheManager.RoleHolder(34L, List.of(new RolePersonDetailsDTO())));
		Boolean resultMatching = service.hasAnyRoleInEdition(PERSON_ID, EDITION_ID);
		assertThat(resultMatching).isTrue();
	}

	@Test
	void removeRole() {
		when(jobOfferService.getEditionFromJobOffer(JOFFER_ID)).thenReturn(EDITION_ID);

		when(editionRolesCache.getRolesInEdition(anyLong(), anyLong())).thenReturn(new EditionRolesCacheManager.RoleHolder(34L, List.of(roleStudent1)));

		doNothing().when(editionService).removePersonFromEdition(anyLong(), anyLong());

		service.removeRole(PERSON_ID, JOFFER_ID);

		verify(editionService).removePersonFromEdition(EDITION_ID, PERSON_ID);
	}

	@Test
	void changeToTaPatchRole() {
		when(jobOfferService.getEditionFromJobOffer(JOFFER_ID)).thenReturn(EDITION_ID);

		when(editionRolesCache.getRolesInEdition(anyLong(), anyLong())).thenReturn(new EditionRolesCacheManager.RoleHolder(34L, List.of(roleStudent1)));

		when(roleControllerApi.patchRole(any(), anyLong(), anyLong()))
				.thenReturn(Mono.just(new RoleId().personId(PERSON_ID).editionId(EDITION_ID)));

		when(editionControllerApi.getEditionParticipants(anyLong())).thenReturn(
				Flux.just(new RolePersonDetailsDTO().person(new PersonSummaryDTO().id(PERSON_ID)))
		);

		service.changeToTa(PERSON_ID, JOFFER_ID, false);

		verify(roleControllerApi).patchRole(new RolePatchDTO().type(RolePatchDTO.TypeEnum.TA), PERSON_ID,
				EDITION_ID);
	}

	@Test
	void changeToTaPatchRole_HeadTaAlready() {
		when(jobOfferService.getEditionFromJobOffer(JOFFER_ID)).thenReturn(EDITION_ID);

		when(editionRolesCache.getRolesInEdition(anyLong(), anyLong())).thenReturn(new EditionRolesCacheManager.RoleHolder(34L, List.of(roleHeadTA)));

		when(roleControllerApi.patchRole(any(), anyLong(), anyLong()))
				.thenReturn(Mono.just(new RoleId().personId(PERSON_ID).editionId(EDITION_ID)));

		when(editionControllerApi.getEditionParticipants(anyLong())).thenReturn(
				Flux.just(new RolePersonDetailsDTO()
						.person(new PersonSummaryDTO().id(PERSON_ID))
						.type(RolePersonDetailsDTO.TypeEnum.HEAD_TA)
				)
		);

		service.changeToTa(PERSON_ID, JOFFER_ID, false);

		verify(roleControllerApi, times(0)).patchRole(new RolePatchDTO().type(RolePatchDTO.TypeEnum.TA), PERSON_ID,
				EDITION_ID);
	}

	@Test
	void changeToTaPatchRole_HeadTaAlready_Demote() {
		when(jobOfferService.getEditionFromJobOffer(JOFFER_ID)).thenReturn(EDITION_ID);

		when(editionRolesCache.getRolesInEdition(anyLong(), anyLong())).thenReturn(new EditionRolesCacheManager.RoleHolder(34L, List.of(roleHeadTA)));

		when(roleControllerApi.patchRole(any(), anyLong(), anyLong()))
				.thenReturn(Mono.just(new RoleId().personId(PERSON_ID).editionId(EDITION_ID)));

		when(editionControllerApi.getEditionParticipants(anyLong())).thenReturn(
				Flux.just(new RolePersonDetailsDTO()
						.person(new PersonSummaryDTO().id(PERSON_ID))
						.type(RolePersonDetailsDTO.TypeEnum.HEAD_TA)
				)
		);

		service.changeToTa(PERSON_ID, JOFFER_ID, true);

		verify(roleControllerApi).patchRole(new RolePatchDTO().type(RolePatchDTO.TypeEnum.TA), PERSON_ID,
				EDITION_ID);
	}

	@Test
	void changeToTaAddRole() {
		when(jobOfferService.getEditionFromJobOffer(JOFFER_ID)).thenReturn(EDITION_ID);

		when(editionRolesCache.getRolesInEdition(anyLong(), anyLong())).thenReturn(new EditionRolesCacheManager.RoleHolder(34L, List.of()));

		when(roleControllerApi.addRole(any()))
				.thenReturn(Mono.just(new RoleId().personId(PERSON_ID).editionId(EDITION_ID)));

		service.changeToTa(PERSON_ID, JOFFER_ID, false);

		verify(roleControllerApi).addRole(new RoleCreateDTO()
				.edition(new EditionIdDTO().id(EDITION_ID))
				.person(new PersonIdDTO().id(PERSON_ID))
				.type(RoleCreateDTO.TypeEnum.TA));
	}

	@Test
	void changeToHeadTaPatchRole() {
		when(jobOfferService.getEditionFromJobOffer(JOFFER_ID)).thenReturn(EDITION_ID);

		when(editionRolesCache.getRolesInEdition(anyLong(), anyLong())).thenReturn(new EditionRolesCacheManager.RoleHolder(34L, List.of(roleTA1)));

		when(roleControllerApi.patchRole(any(), anyLong(), anyLong()))
				.thenReturn(Mono.just(new RoleId().personId(PERSON_ID).editionId(EDITION_ID)));

		service.changeToHeadTa(PERSON_ID, JOFFER_ID);

		verify(roleControllerApi).patchRole(new RolePatchDTO().type(RolePatchDTO.TypeEnum.HEAD_TA), PERSON_ID,
				EDITION_ID);
	}

	@Test
	void changeToHeadTaAddRole() {
		when(jobOfferService.getEditionFromJobOffer(JOFFER_ID)).thenReturn(EDITION_ID);

		when(editionRolesCache.getRolesInEdition(anyLong(), anyLong())).thenReturn(new EditionRolesCacheManager.RoleHolder(34L, List.of()));

		when(roleControllerApi.addRole(any()))
				.thenReturn(Mono.just(new RoleId().personId(PERSON_ID).editionId(EDITION_ID)));

		service.changeToHeadTa(PERSON_ID, JOFFER_ID);

		verify(roleControllerApi).addRole(new RoleCreateDTO()
				.edition(new EditionIdDTO().id(EDITION_ID))
				.person(new PersonIdDTO().id(PERSON_ID))
				.type(RoleCreateDTO.TypeEnum.HEAD_TA));
	}

	@Test
	void changeToTeacherPatchRole() {
		when(editionRolesCache.getRolesInEdition(anyLong(), anyLong())).thenReturn(new EditionRolesCacheManager.RoleHolder(34L, List.of(roleTeacher1)));

		when(roleControllerApi.patchRole(any(), anyLong(), anyLong()))
				.thenReturn(Mono.just(new RoleId().personId(PERSON_ID).editionId(EDITION_ID)));

		service.changeToTeacher(PERSON_ID, EDITION_ID);

		verify(roleControllerApi).patchRole(new RolePatchDTO().type(RolePatchDTO.TypeEnum.TEACHER), PERSON_ID,
				EDITION_ID);
	}

	@Test
	void changeToTeacherAddRole() {
		when(editionRolesCache.getRolesInEdition(anyLong(), anyLong())).thenReturn(new EditionRolesCacheManager.RoleHolder(34L, List.of()));

		when(roleControllerApi.addRole(any()))
				.thenReturn(Mono.just(new RoleId().personId(PERSON_ID).editionId(EDITION_ID)));

		service.changeToTeacher(PERSON_ID, EDITION_ID);

		verify(roleControllerApi).addRole(new RoleCreateDTO()
				.edition(new EditionIdDTO().id(EDITION_ID))
				.person(new PersonIdDTO().id(PERSON_ID))
				.type(RoleCreateDTO.TypeEnum.TEACHER));
	}

	@Test
	void getRolesByIdTest() {
		RoleDetailsDTO role = new RoleDetailsDTO().id(new RoleId().personId(PERSON_ID).editionId(EDITION_ID));

		when(roleControllerApi.getRolesById(anySet(), anySet())).thenReturn(Flux.just(role));

		assertThat(service.getRolesById(List.of(EDITION_ID), List.of(PERSON_ID))).isEqualTo(List.of(role));

		verify(roleControllerApi).getRolesById(Set.of(EDITION_ID), Set.of(PERSON_ID));
	}

	@Test
	void getStaffRolesForPerson_withValidRoles_returnsFilteredRoles() {
		RoleEditionDetailsDTO roleTA = new RoleEditionDetailsDTO().type(RoleEditionDetailsDTO.TypeEnum.TA);
		RoleEditionDetailsDTO roleHeadTA = new RoleEditionDetailsDTO()
				.type(RoleEditionDetailsDTO.TypeEnum.HEAD_TA);
		RoleEditionDetailsDTO roleOther = new RoleEditionDetailsDTO()
				.type(RoleEditionDetailsDTO.TypeEnum.STUDENT);

		when(personApi.getRolesForPerson(PERSON_ID))
				.thenReturn(Flux.just(roleTA, roleHeadTA, roleOther));

		List<RoleEditionDetailsDTO> result = service.getStaffRolesForPerson(PERSON_ID);

		assertThat(result.size()).isEqualTo(2);
		assertTrue(result.contains(roleTA));
		assertTrue(result.contains(roleHeadTA));
	}

	@Test
	void getStaffRolesForPerson_withNoMatchingRoles_returnsEmptyList() {
		RoleEditionDetailsDTO roleOther = new RoleEditionDetailsDTO()
				.type(RoleEditionDetailsDTO.TypeEnum.STUDENT);

		when(personApi.getRolesForPerson(PERSON_ID))
				.thenReturn(Flux.just(roleOther));

		List<RoleEditionDetailsDTO> result = service.getStaffRolesForPerson(PERSON_ID);

		assertThat(result.isEmpty()).isTrue();
	}

	@Test
	void getTAsAndHeadTAsById_withMatchingRoles_returnsFilteredRoles() {
		RolePersonDetailsDTO roleTA = new RolePersonDetailsDTO().type(RolePersonDetailsDTO.TypeEnum.TA);
		RolePersonDetailsDTO roleHeadTA = new RolePersonDetailsDTO()
				.type(RolePersonDetailsDTO.TypeEnum.HEAD_TA);
		RolePersonDetailsDTO roleOther = new RolePersonDetailsDTO()
				.type(RolePersonDetailsDTO.TypeEnum.STUDENT);

		when(roleControllerApi.getRolesByEditions(Set.of(EDITION_ID)))
				.thenReturn(Flux.just(roleTA, roleHeadTA, roleOther));

		List<RolePersonDetailsDTO> result = service.getTAsAndHeadTAsById(EDITION_ID);

		assertThat(result.size()).isEqualTo(2);
		assertTrue(result.contains(roleTA));
		assertTrue(result.contains(roleHeadTA));
	}

	@Test
	void getTAsAndHeadTAsById_withNoMatchingRoles_returnsEmptyList() {
		RolePersonDetailsDTO roleOther = new RolePersonDetailsDTO()
				.type(RolePersonDetailsDTO.TypeEnum.STUDENT);

		when(roleControllerApi.getRolesByEditions(Set.of(EDITION_ID)))
				.thenReturn(Flux.just(roleOther));

		List<RolePersonDetailsDTO> result = service.getTAsAndHeadTAsById(EDITION_ID);

		assertTrue(result.isEmpty());
	}

	@Test
	void isStaffRole_teacher_returnsTrue() {
		boolean result = service.isStaffRole(PersonSummaryDTO.DefaultRoleEnum.TEACHER);

		assertThat(result).isTrue();
	}

	@Test
	void isStaffRole_admin_returnsTrue() {
		boolean result = service.isStaffRole(PersonSummaryDTO.DefaultRoleEnum.ADMIN);

		assertThat(result).isTrue();
	}

	@Test
	void isStaffRole_nonStaff_returnsFalse() {
		boolean result = service.isStaffRole(PersonSummaryDTO.DefaultRoleEnum.STUDENT);

		assertThat(result).isFalse();
	}

	@Test
	void isOnlyTeacherWorks() {
		Flux<RolePersonDetailsDTO> roleDetailsList = Flux.just(roleTeacher1, roleTA1, roleTeacher2,
				roleStudent1, roleHeadTA,
				roleTA2);
		when(roleControllerApi.getRolesByEditions(anySet())).thenReturn(roleDetailsList);

		Assertions.assertThat(service.isTheOnlyTeacherInEdition(person.getId(), 2L)).isFalse();
		Assertions.assertThat(service.isTheOnlyTeacherInEdition(person5.getId(), 2L)).isFalse();

		Flux<RolePersonDetailsDTO> rolesOneTeacher = Flux.just(roleTeacher1, roleTA1, roleStudent1,
				roleHeadTA, roleTA2);
		when(roleControllerApi.getRolesByEditions(anySet()))
				.thenReturn(rolesOneTeacher);
		Assertions.assertThat(service.isTheOnlyTeacherInEdition(person.getId(), 2L)).isTrue();
	}

}
